//Copyright (c) 2006, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// Tiger's Java versions (both 1.4.2 and 1.5.0) have a new "threading model".
// Apple doesn't say much about what this is, and it doesn't seem to be
// related to what Sun calls Java 1.5.0's "new threading model" -- i.e.
// its new synchronization primitives.  From our point of view, though, all
// Tiger's new "threading model" means is that there are _many_ more calls
// from Java drawing code to Cocoa's [NSObject performSelectorOnMainThread]
// (or more precisely to [ThreadUtilities performOnMainThread]) -- which in
// turn means that we must _never_ run Java drawing code on the main thread,
// or even block the main thread while waiting for Java drawing code to
// finish.  Otherwise you will almost certainly get hangs.  (Java 1.4.1 and
// 1.4.2 "plain", though, react badly to taking _all_ drawing code off the
// main thread -- they hang much more often.  So we temper our enthusiasm
// in those older versions.)

package jep;

import java.applet.*;
import java.awt.*;
import java.util.*;
import java.io.*;
import java.net.*;
import java.security.*;
import java.lang.ref.*;
import sun.awt.AppContext;
import sun.awt.PeerEvent;
import sun.awt.SunToolkit;
import sun.plugin.AppletViewer;
import apple.awt.GraphicsClipper;

public class AppletFrame extends Frame {
	protected static boolean showDebugInfo = false;
	protected long cocoaParentView = 0;
	protected long prevCocoaParentView = 0;
	protected int pluginInstance = 0; // MRJPluginInstance *
	protected URL url = null;
	protected AppletViewer panel = null;
	protected HashSet clipperTable = new HashSet();
	protected HashSet childWindows = new HashSet();
	protected static InheritableThreadLocal
		currentAppletFrame = new InheritableThreadLocal();
	public boolean isCarbonApp = true;
	public boolean updatesOff = false;
	public boolean asyncUpdatesOff = false;
	public boolean isInvisible = false;
	public boolean destroying = false;
	public Rectangle currentClip = null;
	public final TimedLock clipLock = new TimedLock();
	protected Rectangle initialAppletBounds = null;
	protected Rectangle savedAppletBounds = null;
	public int imposeClipCount = 0;
	public int removeClipCount = 0;

	public AppletFrame() {
	}

	public AppletFrame(long cocoaParentView, int pluginInstance, URL docbase,
		HashMap attributes, boolean isCarbonApp, boolean createdInvisible)
	{
		this.cocoaParentView = cocoaParentView;
		this.pluginInstance = pluginInstance;
		url = docbase;
		panel = new AppletFramePanel(docbase, attributes, this);
		add(panel);
		initialAppletBounds = new Rectangle(0, 0, panel.getWidth(),
			panel.getHeight());
		setBounds(initialAppletBounds);
		panel.setBounds(initialAppletBounds);
		panel.setAppletContext(new AppletFrameContext(this));
		currentClip = new Rectangle(initialAppletBounds);
		this.isCarbonApp = isCarbonApp;
		this.isInvisible = createdInvisible;
	}

	protected void displayAllComponents(Component components[], String indent) {
		for (int i = 1; i <= components.length; ++i) {
			System.err.println(indent + components[i-1].toString());
			try {
				displayAllComponents(((Container)components[i-1]).getComponents(),
					new String(indent) + "--");
			} catch (ClassCastException e) {
			}
		}
	}

	public void showAppletInfo() {
		Component components[] = getComponents();
		System.err.println("AppletFrame showAppletInfo(): Self and all components:");
		System.err.println("--" + toString());
		displayAllComponents(components, "----");
		Permission perm = new AllPermission();
		try {
			AccessController.checkPermission(perm);
			System.err.println("AppletFrame: showAppletInfo(): We've got AllPermission");
		} catch (AccessControlException e) {
			System.err.println("AppletFrame: showAppletInfo(): We don't have AllPermission");
		}
	}

	public void addClipper(GraphicsClipper clipper) {
		if (clipperTable == null)
			return;
		synchronized(clipperTable) {
			cleanClipperTable();
			clipperTable.add(clipper);
		}
	}

	public void removeClipper(GraphicsClipper clipper) {
		if (clipperTable == null)
			return;
		synchronized(clipperTable) {
			Iterator enumer = clipperTable.iterator();
			while (enumer.hasNext()) {
				GraphicsClipper item = (GraphicsClipper) enumer.next();
				if (item.getGraphics() == null) {
					enumer.remove();
				} else if (item.equals(clipper)) {
					enumer.remove();
				}
			}
		}
	}

	public void cleanClipperTable() {
		if (clipperTable == null)
			return;
		synchronized(clipperTable) {
			Iterator enumer = clipperTable.iterator();
			while (enumer.hasNext()) {
				GraphicsClipper item = (GraphicsClipper) enumer.next();
				if (item.getGraphics() == null)
					enumer.remove();
			}
		}
	}

	public void emptyClipperTable() {
		if (clipperTable == null)
			return;
		synchronized(clipperTable) {
			clipperTable.clear();
		}
	}

	public void validateClips() {
		if (destroying || (clipperTable == null))
			return;
		synchronized(clipperTable) {
			Iterator enumer = clipperTable.iterator();
			while (enumer.hasNext()) {
				GraphicsClipper item = (GraphicsClipper) enumer.next();
				if (item.getGraphics() == null) {
					enumer.remove();
				} else {
					item.validateDevClip();
				}
			}
		}
	}

	public void zeroClips() {
		if (destroying || (clipperTable == null))
			return;
		synchronized(clipperTable) {
			Iterator enumer = clipperTable.iterator();
			while (enumer.hasNext()) {
				GraphicsClipper item = (GraphicsClipper) enumer.next();
				if (item.getGraphics() == null) {
					enumer.remove();
				} else {
					item.zeroDevClip();
				}
			}
		}
	}

	// getCurrentAppletFrame() and setCurrentAppletFrame() use thread local
	// storage to allow us to find out which AppletFrame we're running under
	// at any given time.  This is needed in MyCToolkit.createFrame() and
	// MyCToolkit.createWindow() (from which addChildWindow() is called).
	public static AppletFrame getCurrentAppletFrame() {
		return (AppletFrame) currentAppletFrame.get();
	}

	public static void setCurrentAppletFrame(AppletFrame aFrame) {
		currentAppletFrame.set(aFrame);
	}

	public void addChildWindow(Window aWindow) {
		if ((childWindows == null) || (aWindow == null))
			return;
		synchronized(childWindows) {
			if (childWindows == null)
				return;
			childWindows.add(new WeakReference(aWindow));
		}
	}

	public void disableChildWindows() {
		if (childWindows == null)
			return;
		synchronized(childWindows) {
			if (childWindows == null)
				return;
			Iterator enumer = childWindows.iterator();
			while (enumer.hasNext()) {
				Window aWindow = (Window)
					((WeakReference) enumer.next()).get();
				if (aWindow != null) {
					aWindow.setVisible(false);
					Component components[] = aWindow.getComponents();
					if (components != null) {
						setAllComponentsInvisible(components);
						disableAllComponents(components);
					}
					aWindow.setEnabled(false);
				} else {
					enumer.remove();
				}
			}
		}
	}

	public void disposeChildWindows() {
		if (childWindows == null)
			return;
		synchronized(childWindows) {
			if (childWindows == null)
				return;
			Iterator enumer = childWindows.iterator();
			while (enumer.hasNext()) {
				Window aWindow = (Window)
					((WeakReference) enumer.next()).get();
				if (aWindow != null)
					aWindow.dispose();
				enumer.remove();
			}
			childWindows = null;
		}
	}

	public AppletViewer getPanel() {
		return panel;
	}

	public Applet getApplet() {
		return panel.getApplet();
	}

	public int getLoadingStatus() {
		return panel.getLoadingStatus();
	}

	public void validate() {
		if (destroying || AppletHandlerFactory.isMainThread())
			return;
		super.validate();
	}

	public void invalidate() {
		if (destroying || AppletHandlerFactory.isMainThread())
			return;
		super.invalidate();
	}

	public void validateApplet() {
		if (destroying)
			return;
		// Make sure our validate() method runs in the appropriate execution
		// context (i.e. on the the appropriate thread and event queue).  It
		// should never run on the main thread.  If an applet is running, the
		// execution context will be specific to its thread group (which
		// determines its privileges and access).  Since an applet-writer can
		// change a component's layout manager, he/she might otherwise be able
		// to run code with inappropriate privileges.
		AppContext context = null;
		Object target = null;
		Applet applet = getApplet();
		if (applet != null) {
			target = applet;
		} else {
			target = this;
		}
		context = SunToolkit.targetToAppContext(target);
		boolean isOurContext = AppContext.getAppContext().equals(context);
		if (isOurContext && !AppletHandlerFactory.isMainThread()) {
			validate();
		} else {
			SunToolkit.postEvent(context,
				new PeerEvent(target, new Validate(),
				PeerEvent.PRIORITY_EVENT));
		}
	}

	public void invalidateAndValidate() {
		if (destroying)
			return;
		// Make sure our validate() and invalidate() methods run in the
		// appropriate execution context (i.e. on the the appropriate thread
		// and event queue).  They should never run on the main thread.  If
		// an applet is running, the execution context will be specific to
		// its thread group (which determines its privileges and access).
		// Since an applet-writer can change a component's layout manager,
		// he/she might otherwise be able to run code with inappropriate
		// privileges.
		AppContext context = null;
		Object target = null;
		Applet applet = getApplet();
		if (applet != null) {
			target = applet;
		} else {
			target = this;
		}
		context = SunToolkit.targetToAppContext(target);
		boolean isOurContext = AppContext.getAppContext().equals(context);
		if (isOurContext && !AppletHandlerFactory.isMainThread()) {
			if (applet != null) {
				applet.invalidate();
			} else {
				getPanel().invalidate();
			}
			validate();
			return;
		} else {
			SunToolkit.postEvent(context,
				new PeerEvent(target, new InvalidateAndValidate(),
				PeerEvent.PRIORITY_EVENT));
		}
	}

	// If flagOnly is false and wait is true, this method will wait until
	// setAppletVisibleDirectly() is finished.  But that's generally a bad
	// idea (especially in Tiger, whose new "threading model" reacts even
	// worse than previously to running methods on the main thread which draw
	// or trigger drawing -- or to blocking the main thread while waiting for
	// them to finish).  Now it's probably best to set both flagOnly and wait
	// to false -- isInvisible will be changed immediately (so we can return
	// without waiting), but the applet will also (eventually) be properly
	// made invisible.
	// Our AppletFrame is always visible (after the call to activateFrame()
	// below), as far as the JVM is concerned.  Whether or not the JEP thinks
	// it's visible is determined by the value of isInvisible.  But the applet
	// it contains (if any) may be made visible or invisible by calls to
	// getApplet().setVisible() -- if flagOnly is false.
	public void setAppletVisible(boolean visible, boolean flagOnly, boolean wait) {
		if (destroying)
			return;
		Applet applet = getApplet();
		if ((applet == null) || flagOnly || !wait) {
			boolean visibilityChanged = (visible == isInvisible);
			isInvisible = !visible;
			if (visibilityChanged) {
				if (isInvisible) {
					zeroClips();
				} else {
					validateClips();
				}
			}
			if ((applet == null) || flagOnly)
				return;
		}
		// Any code that draws or triggers a draw should run on the applet's
		// thread (if we have an applet).
		AppContext appletContext = SunToolkit.targetToAppContext(applet);
		boolean isAppletContext = AppContext.getAppContext().equals(appletContext);
		if (isAppletContext) {
			setAppletVisibleDirectly(applet, visible);
		} else {
			long jobID = 0;
			if (wait)
				jobID = AppletHandlerFactory.beforeProcessing(new Object());
			SunToolkit.postEvent(appletContext,
				new PeerEvent(applet, new SetVisible(visible, wait, jobID),
				PeerEvent.PRIORITY_EVENT));
			if (wait)
				AppletHandlerFactory.processMainThreadEvents(200, jobID);
		}
	}

	// If the applet exists (was created successfully), this code should
	// run on its thread.  The same is true of all the other "...Directly()"
	// methods below.
	public void setAppletVisibleDirectly(Applet applet, boolean visible) {
		if (destroying)
			return;
		boolean visibilityChanged = (visible == isInvisible);
		if (applet == null) {
			isInvisible = !visible;
			if (visibilityChanged) {
				if (isInvisible) {
					zeroClips();
				} else {
					validateClips();
				}
			}
			return;
		}
		// If our applet is currently clipped, we must (temporarily) remove
		// the clip before making it invisible, and restore it just before
		// making it visible -- otherwise our clipping gets broken.
		if (savedAppletBounds != null) {
			if (visible) {
				imposeClipDirectly(currentClip.x, currentClip.y,
					currentClip.width, currentClip.height);
			} else {
				removeClipTemporarily();
			}
		}
		applet.setVisible(visible);
		isInvisible = !applet.isVisible();
		if (visibilityChanged) {
			if (isInvisible) {
				zeroClips();
			} else {
				validateClips();
			}
		}
	}

	public boolean isAppletVisible() {
		if (destroying)
			return false;
		try {
			return getApplet().isVisible();
		} catch (NullPointerException e) {
			return !isInvisible;
		}
	}

	// We may have messed with the applet's bounds (and those of its parent)
	// to do our clipping, but we don't want this to change what's returned
	// by getAppletBounds() -- so if need be we return the bounds the applet
	// would have if it hadn't been clipped.
	public Rectangle getAppletBounds() {
		// We really should lock this.  But this usually runs on the main
		// thread, so it's unsafe to do so.
		try {
			return new Rectangle(savedAppletBounds);
		} catch (NullPointerException e) {
			return new Rectangle(initialAppletBounds);
		}
	}

	// Should never be called on main thread, except (maybe) on Java 1.4.1 and
	// Java 1.4.2 "plain".
	public void setAppletBounds(int x, int y, int width, int height) {
		Applet applet = getApplet();
		if (applet == null)
			return;
		applet.setBounds(x, y, width, height);
	}

	// Should never be called on main thread, except (maybe) on Java 1.4.1 and
	// Java 1.4.2 "plain".
	public void setAppletBounds(Rectangle bounds) {
		Applet applet = getApplet();
		if (applet == null)
			return;
		applet.setBounds(bounds);
	}

	// Prior to version 0.8.3, the Java Embedding Plugin only supported
	// resizing applets via Cocoa methods.  This works fine with most
	// applets, which don't expect to be resized dynamically.  But it can
	// cause problems with applets that are designed for this purpose.
	// So we now also support resizing applets directly or "natively", via
	// this method.  It only gets used if the html that loads an applet
	// specifies its height and width as percentages of the space available
	// (instead of as absolute numbers of pixels).
	// As of version 0.9.1 the bulk of this method's code has been moved to
	// AppletFramePanel.appletResize(), so that things will work correctly
	// when an applet (or JavaScript code) calls Applet.resize() directly.
	public void resizeApplet(int width, int height) {
		if (destroying)
			return;
		Applet applet = getApplet();
		if ((width == 0) || (height == 0) || (applet == null))
			return;
		// This stuff needs to be done right away (before we return, and
		// before AppletFramePanel.appletResizeDirectly() runs) -- it matches
		// changes to currentFrame and currentBounds that are made in
		// AppletView resizeApplet.  It really should be done under a lock ...
		// but it would be unsafe to use one since we're on the main thread.
		if ((width != panel.getWidth()) || (height != panel.getHeight())) {
			currentClip = new Rectangle(0, 0, width, height);
			initialAppletBounds = new Rectangle(0, 0, width, height);
			savedAppletBounds = null;
		}
		panel.appletResize(width, height);
	}

	// See SimpleLayout.java.
	public void setSimpleLayout() {
		if (destroying)
			return;
		setLayout(new SimpleLayout(panel.getWidth(), panel.getHeight(), this));
		panel.setLayout(new SimpleLayout(panel.getWidth(), panel.getHeight(), this));
	}

	// We don't currently use this (it causes display problems and doesn't
	// seem to do anything useful).  I leave it in for future reference.
	public void setIgnoreRepaint(boolean flag) {
		super.setIgnoreRepaint(flag);
		panel.setIgnoreRepaint(flag);
		Applet applet = getApplet();
		if (applet == null)
			return;
		applet.setIgnoreRepaint(flag);
	}

	public void setUpdatesOff(boolean b) {
		if (showDebugInfo) {
			if (b && !updatesOff)
				System.err.println("AppletFrame: Turning updates off");
			else if (!b && updatesOff)
				System.err.println("AppletFrame: Turning updates on");
		}
		updatesOff = b;
		boolean changed = (updatesOff != b);
		if (updatesOff) {
			if (changed)
				zeroClips();
			setAsyncUpdatesOff(true);
		}
	}

	public void setAsyncUpdatesOff(boolean b) {
		if (showDebugInfo) {
			if (b && !asyncUpdatesOff)
				System.err.println("AppletFrame: Turning async updates off");
			else if (!b && asyncUpdatesOff)
				System.err.println("AppletFrame: Turning async updates on");
		}
		boolean changed = (asyncUpdatesOff != b);
		asyncUpdatesOff = b;
		if (!asyncUpdatesOff && changed)
			validateClips();
	}

	// Previous versions (before 0.8.2) naively assumed that every component
	// to be clipped uses the same coordinate system as the top-level
	// component (this class) ... but that isn't always true.
	// Our 0.8.2 version of this method was (groan) pretty buggy, though it
	// seems now to be fixed in version 0.8.3.
	public Rectangle translateCurrentClip(Component componentToClip) {
		if (destroying)
			return new Rectangle();
		// Using the Cocoa calls to globally disable and enable screen updates
		// does what we really need here (we don't want screen updates to
		// invalidate our clip translation while we're computing it).  Using
		// getClipLock() (as we did in versions previous to 0.8.5) didn't.
		// All the code between disableScreenUpdates() and enableScreenUpdates()
		// needs to go into a try-catch block -- otherwise we might make one
		// or more unmatched calls to disableScreenUpdates().
		AppletHandlerFactory.disableScreenUpdates();
		Rectangle translatedClip;
		try {
			if (componentToClip.equals(panel))
				componentToClip = (Component) this;
			else if (componentToClip.equals(getApplet()))
				componentToClip = (Component) this;
			Component comp = componentToClip;
			Point compLocation = comp.getLocation();
			int xDiff = compLocation.x;
			int yDiff = compLocation.y;
			while (!comp.equals(this)) {
				Component parent = comp.getParent();
				if (parent.equals(panel))
					parent = (Component) this;
				else if (parent.equals(getApplet()))
					parent = (Component) this;
				if (!parent.equals(this)) {
					Point parentLocation = parent.getLocation();
					xDiff += parentLocation.x;
					yDiff += parentLocation.y;
				}
				comp = parent;
			}
			translatedClip = new Rectangle(currentClip);
			if ((translatedClip.width >= 0) && (translatedClip.height >= 0)) {
				translatedClip.x -= xDiff;
				translatedClip.y -= yDiff;
			}
		} catch (Throwable e) {
			translatedClip = new Rectangle();
		}
		AppletHandlerFactory.enableScreenUpdates();
		return translatedClip;
	}

	public void imposeClip(int clipX, int clipY, int clipWidth, int clipHeight) {
		if (destroying)
			return;
		if (!AppletHandlerFactory.isJava142Update1()) {
			imposeClipDirectly(clipX, clipY, clipWidth, clipHeight);
			return;
		}
		AppContext context = null;
		Object target = null;
		Applet applet = getApplet();
		if (applet != null) {
			target = applet;
		} else {
			target = this;
		}
		context = SunToolkit.targetToAppContext(target);
		boolean isOurContext = AppContext.getAppContext().equals(context);
		if (isOurContext && !AppletHandlerFactory.isMainThread()) {
			imposeClipDirectly(clipX, clipY, clipWidth, clipHeight);
		} else {
			if (++imposeClipCount == 1)
				AppletHandlerFactory.suspendAppletUpdates(getCocoaParentViewLong());
			SunToolkit.postEvent(context, new PeerEvent(target,
				new ImposeClip(clipX, clipY, clipWidth, clipHeight),
				PeerEvent.PRIORITY_EVENT));
		}
	}

	// This is the centerpiece of our grand strategy to make sure that our
	// applet gets clipped correctly (other parts appear in drawRect_AWT()
	// of Handlers.m and in the handleEvent() methods of MyCPanel.java,
	// MyCAppletFrame.java and MyCCanvas.java).  It's a gross hack, but it
	// works surprisingly well.  The basic idea is that if we have to clip the
	// right or the bottom of our applet, we reduce the applet's size on the
	// right and/or the bottom by the appropriate amount.  And if we have to
	// clip our applet's top or left, we shift our applet up or to the left by
	// the amount we need to clip, then shift it's parent (the AppletFramePanel
	// object) down or to the right by equal amounts.  Expects the clip to be
	// in our applet view's coordinate system (i.e. to be either its "bounds"
	// or a subset thereof), and to have a top-left origin (like all Java
	// objects).
	public void imposeClipDirectly(int clipX, int clipY, int clipWidth, int clipHeight) {
		if (destroying)
			return;
		if (!clipLock.lock(1000)) {
			System.err.println("AppletFrame.imposeClip(): lock timed out");
			Throwable barf = new Throwable();
			barf.fillInStackTrace();
			barf.printStackTrace();
			return;
		}
		Rectangle currentAppletBounds = savedAppletBounds;
		if (savedAppletBounds == null) {
			currentAppletBounds = (savedAppletBounds = getAppletBounds());
		}
		if ((clipX == currentAppletBounds.x) &&
			(clipY == currentAppletBounds.y) &&
			(clipWidth == currentAppletBounds.width) &&
			(clipHeight == currentAppletBounds.height))
		{
			removeClipDirectly();
		} else {
			if ((clipX == currentClip.x) &&
				(clipY == currentClip.y) &&
				(clipWidth == currentClip.width) &&
				(clipHeight == currentClip.height))
			{
				validateApplet();
				clipLock.unlock();
				return;
			}
			Applet applet = getApplet();
			// The '22' is needed to ensure that the applet view doesn't
			// overwrite the browser window's title bar.  (This hack is no
			// longer needed in Java 1.4.2 Update 1.)
			if (applet != null) {
				//panel.setBounds(clipX, clipY + 22, clipWidth, clipHeight);
				//setAppletBounds(-clipX, -clipY, clipWidth + clipX,
				//	clipHeight + clipY);
				// It's better for applets that can dynamically resize
				// themselves if we _don't_ clip the right or bottom of
				// the applet here -- otherwise such an applet (e.g.
				// the NOAA radar looping applet) might lay itself out
				// incorrectly.  Our other code will do this clipping.
				if (AppletHandlerFactory.isJava142Update1()) {
					panel.setBounds(clipX, clipY, currentAppletBounds.width,
						currentAppletBounds.height);
				} else {
					panel.setBounds(clipX, clipY + 22, currentAppletBounds.width,
						currentAppletBounds.height);
				}
				setAppletBounds(-clipX, -clipY, currentAppletBounds.width,
					currentAppletBounds.height);
			} else {
				// We don't do any clipping here if our applet wasn't
				// created (i.e. if all we have is an empty view) --
				// instead we rely on our other code to do the clipping.
				if (AppletHandlerFactory.isJava142Update1()) {
					panel.setBounds(0, 0, currentAppletBounds.width,
						currentAppletBounds.height);
				} else {
					panel.setBounds(0, 22, currentAppletBounds.width,
						currentAppletBounds.height);
				}
			}
			currentClip = new Rectangle(clipX, clipY, clipWidth, clipHeight);
			invalidateAndValidate();
			validateClips();
			if (showDebugInfo)
				showAppletInfo();
		}
		clipLock.unlock();
	}

	public void removeClip() {
		if (destroying)
			return;
		if (!AppletHandlerFactory.isJava142Update1()) {
			removeClipDirectly();
			return;
		}
		AppContext context = null;
		Object target = null;
		Applet applet = getApplet();
		if (applet != null) {
			target = applet;
		} else {
			target = this;
		}
		context = SunToolkit.targetToAppContext(target);
		boolean isOurContext = AppContext.getAppContext().equals(context);
		if (isOurContext && !AppletHandlerFactory.isMainThread()) {
			removeClipDirectly();
		} else {
			if (++removeClipCount == 1)
				AppletHandlerFactory.suspendAppletUpdates(getCocoaParentViewLong());
			SunToolkit.postEvent(context, new PeerEvent(target,
				new RemoveClip(), PeerEvent.PRIORITY_EVENT));
		}
	}

	public void removeClipDirectly() {
		if (destroying)
			return;
		Rectangle boundsToRestore = savedAppletBounds;
		if (boundsToRestore == null)
			boundsToRestore = initialAppletBounds;
		if (!clipLock.lock(1000)) {
			System.err.println("AppletFrame.removeClip(): lock timed out");
			Throwable barf = new Throwable();
			barf.fillInStackTrace();
			barf.printStackTrace();
			return;
		}
		if ((boundsToRestore.x == currentClip.x) &&
			(boundsToRestore.y == currentClip.y) &&
			(boundsToRestore.width == currentClip.width) &&
			(boundsToRestore.height == currentClip.height))
		{
			validateApplet();
			clipLock.unlock();
			return;
		}
		if (AppletHandlerFactory.isJava142Update1()) {
			panel.setBounds(0, 0, boundsToRestore.width,
				boundsToRestore.height);
		} else {
			panel.setBounds(0, 22, boundsToRestore.width,
				boundsToRestore.height);
		}
		setAppletBounds(boundsToRestore);
		savedAppletBounds = null;
		currentClip = new Rectangle(boundsToRestore);
		invalidateAndValidate();
		validateClips();
		if (showDebugInfo)
			showAppletInfo();
		clipLock.unlock();
		return;
	}

	// Some methods require that we temporarily remove the clip and then
	// restore it.  Never call this on the main thread (except maybe in
	// Java 1.4.1 and Java 1.4.2 "plain").
	protected void removeClipTemporarily() {
		if (!clipLock.lock(1000)) {
			System.err.println("AppletFrame.removeClipTemporarily(): lock timed out");
			Throwable barf = new Throwable();
			barf.fillInStackTrace();
			barf.printStackTrace();
			return;
		}
		Rectangle boundsToRestore = savedAppletBounds;
		if (boundsToRestore == null)
			boundsToRestore = initialAppletBounds;
		if (AppletHandlerFactory.isJava142Update1()) {
			panel.setBounds(0, 0, boundsToRestore.width,
				boundsToRestore.height);
		} else {
			panel.setBounds(0, 22, boundsToRestore.width,
				boundsToRestore.height);
		}
		setAppletBounds(boundsToRestore);
		clipLock.unlock();
	}

	// Makes our applet view visible for the first time (after suitable
	// preparation).  On Tiger this method must not be called on the
	// main thread.  But with earlier Java versions it actually _must_
	// be called on the main thread.  It's probably significant that the
	// applet doesn't exist yet, so there isn't yet an applet thread to
	// draw on :-)
	public void activateFrame() {
		pack();
		setVisible(true);
	}

	// We use startApplet to start the applet for the first time (just after
	// it's been initialized), and we use stopApplet to stop it for the last
	// time just before we destroy it.  But for starting and stopping the
	// applet at other times we use appletStartApplet and appletStopApplet.
	// These last two methods seem to get along better with our clipping hacks.
	public void startApplet() {
		if (destroying)
			return;
		panel.appletStart();
		isInvisible = false;
		if (showDebugInfo)
			showAppletInfo();
	}

	// At least in Apple's J2SE 5.0 Release 3 (DP3) for OS X 10.4, you can get
	// hangs waiting for panel.appletStop() to return (when rapidly switching
	// between pages containing many applets) if you call it on the main
	// thread -- so on OS X 10.4 we now (as of JEP 0.9.5) call it on the main
	// AWT event queue.  (We can't call it on the applet's event queue -- for
	// some reason that makes the Java garbage collector not collect the
	// applet's garbage after it's destroyed.)  We need to wait for
	// panel.appletStop() to return (otherwise we crash) -- so we must also
	// process main thread events while we're waiting.
	public void stopApplet() {
		if (destroying)
			return;
		if (getApplet() == null)
			return;
		if (!AppletHandlerFactory.isJava142Tiger() ||
			!AppletHandlerFactory.isMainThread())
		{
			stopAppletDirectly();
		} else {
			long jobID = AppletHandlerFactory.beforeProcessing(new Object());
			SunToolkit.postEvent(AppletHandlerFactory.mainAppContext,
				new PeerEvent(this, new StopApplet(jobID),
				PeerEvent.PRIORITY_EVENT));
			AppletHandlerFactory.processMainThreadEvents(0, jobID);
		}
	}

	public void stopAppletDirectly() {
		if (destroying)
			return;
		isInvisible = true;
		panel.appletStop();
	}

	public void appletStartApplet() {
		if (destroying)
			return;
		Applet applet = getApplet();
		if (applet == null)
			return;
		AppContext appletContext = SunToolkit.targetToAppContext(applet);
		boolean isAppletContext = AppContext.getAppContext().equals(appletContext);
		if (isAppletContext) {
			appletStartAppletDirectly(applet);
			return;
		}
		SunToolkit.postEvent(appletContext,
			new PeerEvent(applet, new AppletStartApplet(),
			PeerEvent.PRIORITY_EVENT));
	}

	public void appletStartAppletDirectly(Applet applet) {
		if ((applet == null) || destroying)
			return;
		setAppletVisibleDirectly(applet, true);
		applet.start();
	}

	public void appletStopApplet() {
		if (destroying)
			return;
		Applet applet = getApplet();
		if (applet == null)
			return;
		AppContext appletContext = SunToolkit.targetToAppContext(applet);
		boolean isAppletContext = AppContext.getAppContext().equals(appletContext);
		if (isAppletContext) {
			appletStopAppletDirectly(applet);
			return;
		}
		SunToolkit.postEvent(appletContext,
			new PeerEvent(applet, new AppletStopApplet(),
			PeerEvent.PRIORITY_EVENT));
	}

	public void appletStopAppletDirectly(Applet applet) {
		if ((applet == null) || destroying)
			return;
		setAppletVisibleDirectly(applet, false);
		applet.stop();
	}

	protected void setAllComponentsInvisible(Component components[]) {
		for (int i = 1; i <= components.length; ++i) {
			components[i-1].setVisible(false);
			try {
				setAllComponentsInvisible(((Container)components[i-1]).getComponents());
			} catch (ClassCastException e) {
			}
		}
	}

	protected void disableAllComponents(Component components[]) {
		for (int i = 1; i <= components.length; ++i) {
			components[i-1].setEnabled(false);
			try {
				disableAllComponents(((Container)components[i-1]).getComponents());
			} catch (ClassCastException e) {
			}
		}
	}

	// Before we destroy the applet, we must first make the panel object (the
	// applet's parent) and all its components invisible, then disable them
	// (which makes them undisplayable) -- otherwise external code may attempt
	// to draw the applet after it's been destroyed, which leads to crashes.
	// We do this on the applet's event queue thread in destroyAppletDirectly1().
	// From there we post a call to destroyAppletDirectly2() on the "main"
	// event queue (_not_ the main thread), which actually destroys the applet
	// and "disposes" of its resources.  (We used to do this on the applet's
	// event queue -- but that caused some of the applet's threads (including
	// its event loop) to be leaked.)  (As of JEP 0.9.4 we do more cleanup as
	// the applet is destroyed, to make sure we're not somehow preventing the
	// garbage collector from finalizing our objects (for example by not
	// removing all our global references to these objects).  The only object
	// I can't make the garbage collector reclaim immediately is
	// AppletFramePanel -- I currently have no idea why.)
	public void destroyApplet() {
		if (destroying)
			return;
		destroying = true;
		prevCocoaParentView = cocoaParentView;
		cocoaParentView = 0;
		isInvisible = true;
		Applet applet = getApplet();
		if (AppletHandlerFactory.getPropertyBoolean("jep.debug.release", false)) {
			System.out.println(new Date() + " AppletFrame.destroyApplet() for " +
				((applet != null) ? applet.getClass().getName() :
				"0x" + Long.toHexString(prevCocoaParentView)));
		}
		if (applet == null) {
			emptyClipperTable();
			clipperTable = null;
			panel.setLayout(null);
			setLayout(null);
			SunToolkit.postEvent(AppletHandlerFactory.mainAppContext,
				new PeerEvent(this, new Destroy2(), 0));
			return;
		}
		//AppletHandlerFactory.deleteAppletJNIGlobalRefs(prevCocoaParentView);
		AppContext appletContext = SunToolkit.targetToAppContext(applet);
		boolean isAppletContext = AppContext.getAppContext().equals(appletContext);
		if (isAppletContext) {
			destroyAppletDirectly1();
			return;
		}
		SunToolkit.postEvent(appletContext,
			new PeerEvent(applet, new Destroy1(),
			PeerEvent.PRIORITY_EVENT));
	}

	// _Don't_ make the frame object (the AppletFrame object, which is our
	// applet's grandparent) invisible or disable it here -- that messes
	// things up.  Call this on the applet's thread.
	public void destroyAppletDirectly1() {
		destroying = true;
		cocoaParentView = 0;
		isInvisible = true;
		// We need to explicitly dispose of any child frames or windows that
		// may have been created by our applet -- simply reloading the page
		// won't get rid of them.  For an example of this, visit
		// http://www.lohn1.de/lobn.htm, click on the "berechnen" button
		// (which brings up a new frame), and reload the page:  Without this
		// change, the frame won't get closed.  Moreover, we need to disable
		// them first and only dispose of them later -- otherwise we get
		// crashes, particularly on the trunk.  (Unfortunately we can't
		// dispose of them on the main AWT event queue -- this would allow
		// user code (in overridden Window.dispose() methods) to run in a
		// privileged context.)
		disableChildWindows();
		panel.setVisible(false);
		Component components[] = panel.getComponents();
		if (components != null) {
			setAllComponentsInvisible(components);
			disableAllComponents(components);
		}
		panel.setEnabled(false);
		panel.setLayout(null);
		setLayout(null);
		emptyClipperTable();
		clipperTable = null;
		disposeChildWindows();
		SunToolkit.postEvent(AppletHandlerFactory.mainAppContext,
			new PeerEvent(this, new Destroy2(), 0));
	}

	// Call this on the main AWT event queue -- _not_ on the main thread.
	public void destroyAppletDirectly2() {
		if (getApplet() == null) {
			panel.setVisible(false);
			panel.setEnabled(false);
		}
		AppletHandlerFactory.deleteAppletJNIGlobalRefs(prevCocoaParentView);
		((AppletFramePanel) panel).cleanup();
		// Wait indefinitely for our applet to be destroyed (by specifying '0'
		// in our call to panel.appletDestroy()), but do this on another
		// thread (to avoid blocking the main AWT thread).  Since we're
		// launching the thread here, it will run in the main AppContext
		// (which is what we want).
		Runnable croak = new Runnable() {
			public void run() {
				AppletFrame.this.panel.appletDestroy(0);
				AppletFrame.this.panel = null;
				// AppletFrame.dispose() needs to be called on an event
				// dispatch thread (otherwise it may hang).
				SunToolkit.postEvent(AppletHandlerFactory.mainAppContext,
					new PeerEvent(AppletFrame.this, new Destroy3(), 0));
			}
		};
		Thread croaker = new Thread(croak);
		croaker.start();
	}

	// Call this on the main AWT event queue -- _not_ on the main thread.
	public void destroyAppletDirectly3() {
		dispose();
	}

	// Safe to use on either Java 1.4.1 or 1.4.2.
	public int getCocoaParentViewInt() {
		return (int) cocoaParentView;
	}

	// Safe to use on either Java 1.4.1 or 1.4.2.
	public long getCocoaParentViewLong() {
		return (long) cocoaParentView;
	}

	public int getPluginInstance() {
		return pluginInstance;
	}

	// See changeFrameAppContext() in AppletFrame142.java.
	public void changeFrameAppContext(AppContext appcontext) {
		SunToolkit.insertTargetMapping(this, appcontext);
		SunToolkit.insertTargetMapping(panel, appcontext);
	}

	public AppletContext getAppletContext() {
		return panel.getAppletContext();
	}

	public String toString() {
		return super.toString() + " (jep)";
	}

	public static class TimedLock {
		private Stack lockStack = new Stack();

		public TimedLock() {
		}

		public void lock() {
			lock(0);
		}

		public boolean lock(long timeout) {
			Thread newLock = null;
			Thread firstLock = null;
			synchronized(lockStack) {
				try {
					firstLock = (Thread) lockStack.elementAt(0);
				} catch (ArrayIndexOutOfBoundsException e) {
				}
				newLock = Thread.currentThread();
				lockStack.push(newLock);
				if (firstLock == null)
					return true;
				if (newLock.equals(firstLock))
					return true;
			}
			synchronized(newLock) {
				try {
					newLock.wait(timeout);
				} catch (InterruptedException e) {
				}
			}
			synchronized(lockStack) {
				try {
					firstLock = (Thread) lockStack.elementAt(0);
				} catch (ArrayIndexOutOfBoundsException e) {
					firstLock = null;
				}
				if (firstLock == null) {	// Should never happen.
					System.err.println("AppletFrame.TimedLock.lock(): Lock stack is empty!");
					Throwable barf = new Throwable();
					barf.fillInStackTrace();
					barf.printStackTrace();
					return false;
				}
				Thread currentThread = Thread.currentThread();
				if (!currentThread.equals(firstLock)) {
					// Clean out the object representing our call to
					// lock(long), which has just timed out.  There should
					// be only one, but we don't know where in the stack
					// it will be.
					for (int i = lockStack.size(); i >= 2; --i) {
						Thread lock = (Thread) lockStack.elementAt(i-1);
						if (currentThread.equals(lock))
							lockStack.removeElementAt(i-1);
					}
					return false;
				}
				return true;
			}
		}

		public boolean trylock() {
			synchronized(lockStack) {
				Thread firstLock = null;
				try {
					firstLock = (Thread) lockStack.elementAt(0);
				} catch (ArrayIndexOutOfBoundsException e) {
				}
				if (firstLock == null) {
					lock();
					return true;
				}
				if (Thread.currentThread().equals(firstLock)) {
					lock();
					return true;
				}
				return false;
			}
		}

		public void unlock() {
			synchronized(lockStack) {
				int currentThreadLockCount = 0;
				Thread currentThread = Thread.currentThread();
				for (int i = lockStack.size(); i >= 1; --i) {
					Thread lock = (Thread) lockStack.elementAt(i-1);
					if (currentThread.equals(lock)) {
						if (++currentThreadLockCount == 1)
							lockStack.removeElementAt(i-1);
					}
				}
				if (currentThreadLockCount >= 2)
					return;
				if (currentThreadLockCount == 0) {
					System.err.println("AppletFrame.TimedLock.unlock() called without matching call to lock()!");
					Throwable barf = new Throwable();
					barf.fillInStackTrace();
					barf.printStackTrace();
					return;
				}
				if (!lockStack.empty()) {
					Thread firstLock = (Thread) lockStack.elementAt(0);
					synchronized(firstLock) {
						firstLock.notify();
					}
				}
			}
		}

	}

	private class Validate implements Runnable {

		public void run() {
			AppletFrame.this.validate();
			if (showDebugInfo) {
				System.err.println("AppletFrame Validate run() called on " +
					AppletFrame.this.toString());
			}
		}

	}

	private class InvalidateAndValidate implements Runnable {

		public void run() {
			Applet applet = AppletFrame.this.getApplet();
			if (applet != null) {
				applet.invalidate();
			} else {
				AppletFrame.this.getPanel().invalidate();
			}
			AppletFrame.this.validate();
			if (showDebugInfo) {
				System.err.println("AppletFrame InvalidateAndValidate run() called on " +
					AppletFrame.this.toString());
			}
		}

	}

	private class SetVisible implements Runnable {
		private final boolean flag;
		private final boolean wait;
		private final long jobID;

		SetVisible(boolean flag, boolean wait, long jobID) {
			this.flag = flag;
			this.wait = wait;
			this.jobID = jobID;
		}

		public void run() {
			AppletFrame.this.setAppletVisibleDirectly(
				AppletFrame.this.getApplet(), flag);
			if (wait)
				AppletHandlerFactory.stopProcessing(jobID);
			if (showDebugInfo) {
				System.err.println("AppletFrame SetVisible run() called, flag is " +
					(flag ? "true" : "false"));
			}
		}

	}

	private class StopApplet implements Runnable {
		private final long jobID;

		StopApplet(long jobID) {
			this.jobID = jobID;
		}

		public void run() {
			AppletFrame.this.stopAppletDirectly();
			AppletHandlerFactory.stopProcessing(jobID);
			if (showDebugInfo) {
				System.err.println("AppletFrame StopApplet run() called");
			}
		}

	}

	private class AppletStartApplet implements Runnable {

		public void run() {
			AppletFrame.this.appletStartAppletDirectly(
				AppletFrame.this.getApplet());
			if (showDebugInfo) {
				System.err.println("AppletFrame AppletStartApplet run() called");
			}
		}

	}

	private class AppletStopApplet implements Runnable {

		public void run() {
			AppletFrame.this.appletStopAppletDirectly(
				AppletFrame.this.getApplet());
			if (showDebugInfo) {
				System.err.println("AppletFrame AppletStopApplet run() called");
			}
		}

	}

	private class ImposeClip implements Runnable {
		private final int clipX;
		private final int clipY;
		private final int clipWidth;
		private final int clipHeight;

		ImposeClip(int clipX, int clipY, int clipWidth, int clipHeight) {
			this.clipX = clipX;
			this.clipY = clipY;
			this.clipWidth = clipWidth;
			this.clipHeight = clipHeight;
		}

		public void run() {
			if (--imposeClipCount == 0) {
				AppletFrame.this.imposeClipDirectly(clipX, clipY, clipWidth,
					clipHeight);
				AppletHandlerFactory.resumeAppletUpdates(
					AppletFrame.this.getCocoaParentViewLong());
			} else if (imposeClipCount < 0) {
				imposeClipCount = 0;
			}
			if (showDebugInfo) {
				System.err.println("AppletFrame ImposeClip run() called");
			}
		}

	}

	private class RemoveClip implements Runnable {

		public void run() {
			if (--removeClipCount == 0) {
				AppletFrame.this.removeClipDirectly();
				AppletHandlerFactory.resumeAppletUpdates(
					AppletFrame.this.getCocoaParentViewLong());
			} else if (removeClipCount < 0) {
				removeClipCount = 0;
			}
			if (showDebugInfo) {
				System.err.println("AppletFrame RemoveClip run() called");
			}
		}

	}

	private class Destroy1 implements Runnable {

		public void run() {
			AppletFrame.this.destroyAppletDirectly1();
			if (showDebugInfo) {
				System.err.println("AppletFrame Destroy1 run() called");
			}
		}

	}

	private class Destroy2 implements Runnable {

		public void run() {
			AppletFrame.this.destroyAppletDirectly2();
			if (showDebugInfo) {
				System.err.println("AppletFrame Destroy2 run() called");
			}
		}

	}

	private class Destroy3 implements Runnable {

		public void run() {
			AppletFrame.this.destroyAppletDirectly3();
			if (showDebugInfo) {
				System.err.println("AppletFrame Destroy3 run() called");
			}
		}

	}

	static {
		if ("true".equalsIgnoreCase(System.getProperty("jep.debuginfo", "false"))) {
			showDebugInfo = true;
		}
	}

}

