/*
 * Copyright (c) 2004, Steven Michaud, All Rights Reserved
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#import <unistd.h>
#import <Carbon/Carbon.h>
#import <JavaVM/jni.h>
#import <objc/Object.h>
#import <mach-o/dyld.h>
#import <pthread.h>

#import "JavaEmbeddingPlugin.h"
#import "AppletView.h"
#import "CarbonInternals.h"

#ifdef __cplusplus
extern "C" {
#endif

// Undocumented calls in AppKit Framework (present in Jaguar and still there
// in Tiger).
extern BOOL _NSIsCarbonApp(void);
extern MenuRef _NSGetCarbonMenu(NSMenu *);
extern void _NSAppKitLock(void);
extern void _NSAppKitUnlock(void);
// I've found two sources for these two calls -- a program called RXStack
// (http://zohar.devklog.net:47805/projects/rivenx/changeset/19/Engine/RXStack.m)
// and some old WebKit code
// (http://trac.webkit.org/projects/webkit/browser/tags/WebKit-125~4~2/WebKit/Carbon.subproj/CarbonUtils.m?rev=7119),
// RXStack gives them the following definitions:
//   void NSPushAutoreleasePool(void);
//   void NSPopAutoreleasePool(void);
// But I suspect the definitions that can be deduced from the old WebKit code
// are closer to accurate.  And I've used gdb hacking and other detective work
// to fill them out (and slightly alter them).  NSPushAutoreleasePool() is
// called from [NSAutoreleasePool initWithCapacity:], using the same "capacity"
// parameter (though I'm not sure what "capacity" means in this this context).
// The WebKit code implies that NSPushAutoreleasePool()'s return value is
// "void *".  But it casts this value to "unsigned" and considers it to be the
// number of autorelease pools in existence before the call to
// NSPushAutoreleasePool().  So it really only makes sense to say that its
// return value is "unsigned".  NSPopAutoreleasePool()'s "offset" parameter
// seems to indicate the offset of the pool that should be popped (presumably
// the call will only work properly if you ask it to pop the topmost pool).
// NSPopAutoreleasePool() really does return a pointer, but I don't know what
// kind (neither po nor CFShow() work on it in gdb), so I've called it a
// "void *".  Tiger's HIToolbox framework has HIPushAutoreleasePool() and
// HIPopAutoreleasePool(), which call NSPushAutoreleasePool() and
// NSPopAutoreleasePool() (respectively), and which the debug version of this
// framework defines as follows:
//   void *HIPushAutoreleasePool(void);
//   void HIPopAutoreleasePool(HIAutoreleasePool *);
// Like NSPushAutoreleasePool(), HIPushAutoreleasePool() seems to return the
// number of autorelease pools in existence before the call was made.  So its
// return value should also really be "unsigned".
extern unsigned NSPushAutoreleasePool(unsigned capacity);
extern void *NSPopAutoreleasePool(unsigned offset);

// Undocumented calls in HIToolbox.
extern Boolean GetEventPlatformEventRecord(EventRef inEvent, void *eventRec);
// ConvertPlatformEventRecordAndPost() gets called fron NSApplication
// postEvent:atStart:.  When atStart is false 'priority' is '-2'; when it's
// true 'priority' is '-1'.  'signal' seems to default to 'true'.
extern Boolean ConvertPlatformEventRecordAndPost(void *inEventRec,
	int priority, Boolean signal);
extern Boolean TSMProcessMouseEvent(EventRef inEvent);
extern Boolean TSMProcessRawKeyEvent(EventRef inEvent);

// As of JEP 0.9.6 (0.9.5+d) I use Objective C "categories" to make it
// possible to use Objective C syntax to call undocumented methods.  This
// may help make the code easier to read.  But it's also (partly) necessary:
// There's a bug in Apple's Objective C compiler that prevents
// objc_msgSend_stret() and its friends from compiling correctly on the
// Intel platform (Apple doesn't call this a bug, but that's what it is):
// http://developer.apple.com/documentation/MacOSX/Conceptual/universal_binary/universal_binary_tips/chapter_5_section_23.html
// Thanks to Mark Mentovai for pointing this out to me.  And also much
// thanks to him for patching and maintaining the Intel binaries of the
// Java Embedding Plugin (and the MRJ Plugin JEP) while I wait to get an
// Intel Mac from Apple!

@interface NSApplication (Undocumented)

- (void)_setCurrentEvent:(NSEvent *)anEvent;
- (void)_removeWindowFromCache:(NSWindow *)aWindow;
- (void *)_getWindowCache:(NSWindow *)aWindow add:(BOOL)flag;
- (id)_setKeyWindow:(NSWindow *)aWindow;
- (BOOL)_isRunningModal;
// Returns an array of visible windows in their z-order, with the top-most
// window first.  The list may include miniaturized windows, and the z-order
// seems ill-defined when this happens.
- (NSArray *)orderedWindows;

@end

@interface NSWindow (Undocumented)

- (BOOL)_hasWindowRef;
- (void)_setVisible:(BOOL)flag;
- (int)_realWindowNumber;
- (void)_setWindowNumber:(int)aNumber;
// The following is available on OS X 10.2.8, where we use it in place of
// - (NSRect)contentRectForFrameRect:(NSRect)frameRect.
- (NSRect)contentRectForFrameRect:(NSRect)frameRect styleMask:(unsigned int)aStyle;
// The following is only available on OS X 10.3 and higher.
- (NSRect)contentRectForFrameRect:(NSRect)frameRect;
- (void)_installCarbonWindowEventHandlers;

@end

@class NSInputContext;

@interface NSView (Undocumented)

- (NSRect)_dirtyRect;
- (id)_invalidateFocus;
- (NSInputContext *)inputContext;
- (void)_recursiveDisplayAllDirtyWithLockFocus:(BOOL)lock visRect:(NSRect)aRect;

@end

@interface NSEvent (Undocumented)

- (EventRef)_eventRef;
- (void)_setEventRef:(void *)anEventRef;

@end

// For methods of undocumented classes I use categories of NSObject.  No,
// these methods can't be called on any object.  But you also can't refer
// to the class name in a category definition.

@interface NSObject (Whatever)

// cWindow is a method of NSViewAWT on Jaguar and Panther, and of
// AWTDrawableContextManager on Tiger.
- (id)cWindow;

@end

@interface NSObject (NSViewAWT)

- (jobject)peer;	// Also in ControlModel and other ...ViewAWT objects
- (id) model;		// Also in other <AWTPeerControl> protocol objects
- (void)mouseMovedAWT:(NSEvent *)theEvent;
- (jobject)awtComponent:(JNIEnv *)javaEnv;	// Panther and above
- (jobject)awtComponent;					// Jaguar

@end

@interface NSObject (ControlModel)

- (id)getOwningWindow;	// Jaguar and Panther only
- (id)getOwningModel;	// Tiger only
- (id)getParent;
- (NSView *)getView;
- (void)updateDraggingDestination;	// Jaguar and Panther only

@end

@interface NSObject (CWindow)

- (id)getNSViewAWTContext;		// Only on Panther
- (id)getAWTDrawableContext;	// Only on Tiger
- (BOOL)windowLockOnThirdThread:(NSView *)aView;	// Only on Jaguar
- (void)windowUnlock;			// Only on Jaguar
- (NSWindow *)getNSWindow;

@end

// Methods of CAppletWindow on Jaguar and Panther, and of CEmbedded on Tiger.
@interface NSObject (AppletWindow)

- (NSView *)hybridView;

@end

// Methods of NSViewAWTContext on Panther and AWTDrawableContext on Tiger.
@interface NSObject (AWTContext)

- (void)enable;
- (void)disable;

@end

// Tiger and Leopard only.
@interface NSObject (CPerformer)

// Java 1.4.2/1.5 only (except on Leopard)
- (id)initWithTarget:(id)target selector:(SEL)aSelector arg:(id)anObject;
// Java 1.6 only (except on Leopard)
- (id)initWithTarget:(id)target selector:(SEL)aSelector arg:(id)anObject wait:(BOOL)wait;

@end

@interface NSObject (CMenu)

- (NSMenu *)menu;

@end

// Methods of ThreadUtilities on Jaguar (OS X 10.2.8), or of
// AWTModelUpdater on Panther (OS X 10.3) and Tiger (OS X 10.4).
@interface NSObject (AWTModelUpdaterEtc)

+ (void)registerModelForUpdate:(id)aModel;

@end

Boolean EqualRectFuzzy(Rect rectA, Rect rectB, int fuzzFactor);
Boolean ContainsRectFuzzy(Rect rectA, Rect rectB, int fuzzFactor);
Boolean SectRectFuzzy(Rect rectA, Rect rectB, Rect *intersection,
	int fuzzFactor);
BOOL EqualNSRectFuzzy(NSRect rectA, NSRect rectB, int fuzzFactor);
BOOL ContainsNSRectFuzzy(NSRect rectA, NSRect rectB, int fuzzFactor);
BOOL SectNSRectFuzzy(NSRect rectA, NSRect rectB, NSRect *intersection,
	int fuzzFactor);
Boolean DiffRect(Rect rectA, Rect rectB, unsigned *numResults, Rect *results,
	int fuzzFactor);
BOOL DiffNSRect(NSRect rectA, NSRect rectB, unsigned *numResults,
	NSRect *results, int fuzzFactor);
Boolean DiffRgnRect(RgnHandle region, Rect rect, unsigned *numResults,
	Rect **results, int fuzzFactor);
Boolean GetRgnRects(RgnHandle region, unsigned *numResults, Rect **results);

@class JEPController;

jint JNICALL java_vfprintf(FILE *fp, const char *format, va_list args);

OSStatus InstallCarbonEventHandlers(WindowRef parentWindow,
	JEPController *controller);
OSStatus InstallTSMEventHandler(WindowRef parentWindow,
	JEPController *controller);
OSStatus MaybeInstallTSMAEEventHandler(void);
void RemoveTSMAEEventHandler(void);
OSStatus MaybeInstallAppEventHandlers(void);
void RemoveAppEventHandlers(void);

BOOL PoseAsNSCarbonWindow(void);
BOOL PoseAsNSCarbonWindowContentView(void);
BOOL PoseAsNSCarbonWindowFrame(void);
BOOL PoseAsEmbeddedAppletView(void);
BOOL PoseAsNSViewAWT(void);
BOOL PoseAsNSTSMInputContext(void);
BOOL PoseAsCTextField(void);
BOOL PoseAsCWindow(void);
BOOL PoseAsCAppletWindow(void);
BOOL PoseAsCEmbedded(void);
BOOL PoseAsEventFactory(void);
BOOL PoseAsFocusManager(void);
BOOL PoseAsThreadUtilities(void);
BOOL PoseAsAWTModelUpdater(void);
BOOL PoseAsControlModel(void);
BOOL PoseAsNSButtonAWT(void);
BOOL PoseAsNSScroller(void);
BOOL PoseAsNSScrollViewAWT(void);
BOOL PoseAsNSChoiceAWT(void);
BOOL PoseAsNSTableViewAWT(void);
BOOL PoseAsNSTextFieldAWT(void);
BOOL PoseAsNSTextViewAWT(void);
BOOL PoseAsNSTextFieldFocusView(void);
BOOL PoseAsNSWindowViewAWT(void);
BOOL PoseAsAWTCombinedTextField(void);
BOOL PoseAsAWTCombinedTextFieldCell(void);
BOOL PoseAsAWTCombinedTextView(void);
BOOL PoseAsNSViewAWTContext(void);
BOOL PoseAsAWTDrawableContextManager(void);
BOOL PoseAsAWTStarter(void);
BOOL PoseAsStartupUtility(void);
BOOL PoseAsAwtException(void);
BOOL PoseAsCPopupMenu(void);
BOOL PoseAsSyntheticEvent(void);

@class NSAppShadow;

extern BOOL nsAppTerminating;
extern BOOL isCarbonApp;
extern unsigned long osxVersion;
extern BOOL isJava142;
extern BOOL isJava142Update1;
extern BOOL isJava142Update2;
extern BOOL isJava142Tiger;
extern BOOL isJava150;
extern BOOL isJava160;
extern BOOL awtLibrariesLoaded;
extern BOOL asyncInitializeApplet;
extern BOOL debugVisibility;
extern BOOL debugRelease;
extern BOOL debugUpdates;
extern BOOL wasInEventTrackingMode;
extern NSAppShadow *nsAppShadow;
extern void (*afterInitAWT) (void *);
extern void *afterInitAWTInfo;
extern NSThread *mainThread;
extern CFRunLoopRef mainRunLoop;
extern CFRunLoopSourceRef jepRunLoopSource;
extern CFRunLoopSourceRef jepCRunLoopSource;
extern BOOL everCreatedJVM;

extern void (*ProcessBrowserMainThreadEvents) (void);

extern NSMutableArray *controllerArray;
extern NSRecursiveLock *controllersLock;

@class PopupInfo;

extern NSMutableArray *popupsArray;
extern NSRecursiveLock *popupsLock;

extern NSString *installedPath;
extern NSString *homeDirectory;
extern NSString *javaPluginCocoaPath;

extern Class carbonWindowClass;
extern Class awtViewContextClass;
extern Class cPerformerClass;
extern Class cScrollPaneClass;
extern Class appletWindowClass;

extern NSLock *screenUpdatesLock;
extern NSLock *modelUpdaterLock;
extern NSLock *allocCAWLock;
extern NSMutableArray *deletedCAWs;

extern NSMutableArray *Java_runningJobs;
extern NSLock *Java_runningJobs_lock;

//#define SUPPORT_64BITS 1

//#define DEBUG_EVENTS 1
//#define DEBUG_JVM_EVENTS 1	// VERY noisy.  Beware.
//#define DEBUG_APPLET_HOLDER 1

#define LOCK_TIMEOUT .1
#define SHORT_LOCK_TIMEOUT .01

typedef enum {
	lockTimeoutNone = 0,
	lockTimeoutShort,
	lockTimeoutNormal
} LockTimeout;

#ifndef NSNumericSearch
#define NSNumericSearch 64
#endif

enum {
	kJEPCreatorCode = 'jepb',
	kPropertyNSWindow = 'nswn',
	kPropertyBrowserTSMDoc = 'jept',
	kPropertyCocoaTSMDoc = 'jepc',			// typeBoolean
	kPropertyMouseActivated = 'msac',		// typeBoolean
	kEventParamJEPRedispatched = 'jepr',	// typeBoolean
	kEventParamJEPFixedUp = 'jepf',			// typeBoolean
	kEventParamJEPNotForCocoa = 'jepn',		// typeBoolean
	kEventParamJEPNotForPosting = 'jepp',	// typeBoolean
	kEventParamJEPAlreadyPosted = 'jepa',	// typeBoolean
	kEventParamAlreadyPrinted = 'alpr',		// typeBoolean
	kEventParamOrigUpdate = 'orup'			// typeQDRectangle
};

typedef struct {
	BOOL handlersUpdated;
	AEEventHandlerUPP textServicesHandler;
	AEEventHandlerUPP updateActiveInputAreaHandler;
	AEEventHandlerUPP showHideInputWindowHandler;
	AEEventHandlerUPP pos2OffsetHandler;
	AEEventHandlerUPP offset2PosHandler;
	AEEventHandlerUPP unicodeNotFromInputMethodHandler;
	AEEventHandlerUPP getSelectedTextHandler;
	SInt32 textServicesRefcon;
	SInt32 updateActiveInputAreaRefcon;
	SInt32 showHideInputWindowRefcon;
	SInt32 pos2OffsetRefcon;
	SInt32 offset2PosRefcon;
	SInt32 unicodeNotFromInputMethodRefcon;
	SInt32 getSelectedTextRefcon;
} tsmAEHandlerInfo;

typedef struct {
	BOOL stateSaved;
	BOOL carbonStateSet;
	BOOL cocoaStateSet;
	TSMDocumentID parentActiveDoc;	// parentWindow's TSMDocument, if any
} tsmStateInfo;

@class ActivateInputContext;
@class AppletView;

@interface JEPController : NSObject
{
@public
	NSWindow *myWindow;
	WindowRef myWindowCarbon;
	BOOL myWindowClosing;
	AppletView *myAppletView;
	id appletWindow;
	id appletAWTContext;
	JavaVM *javaVM;
	void *pluginInstance;
	BOOL createdInvisible;
	BOOL erasePreviousFrame;
	BOOL resizeAppletDirectly;
	int updatesOffCounter;
	BOOL updatesOff;
	BOOL asyncUpdatesOff;
	BOOL waitingForScreenUpdatesEnabled;
	BOOL skipDisplayOnce;
	BOOL skipUpdateOnce;
	BOOL suspendMouseDown;
	BOOL hasKeyWindow;
	BOOL everMouseFocused;
	BOOL drawAppletPending;
	int suspendDrawAppletCounter;
	BOOL skipDrawAppletOnce;
	BOOL allowDrawAppletOnce;
	BOOL displayNeedsUpdate;
	NSRecursiveLock *myLock;
	NSPoint initialPosition;
	unsigned initialHeight;
	unsigned initialWidth;
	NSRect initialFrame;	// In parent window's coordinate system, whether Carbon or Cocoa
	NSRect currentFrame;	// In parent window's coordinate system, whether Carbon or Cocoa
	NSRect previousFrame;	// In parent window's coordinate system, whether Carbon or Cocoa
	EventHandlerUPP windowHandlerUPP;
	EventHandlerUPP tsmHandlerUPP;
	EventHandlerUPP dispatcherHandlerUPP;
	EventHandlerRef windowHandlerRef;
	EventHandlerRef tsmHandlerRef;
	EventHandlerRef dispatcherHandlerRef;
	tsmStateInfo tsmInfo;
	UInt32 lastKeyDownControlModifiers;
	JEPShowStatusCallback showStatusCallback;
	JEPShowDocumentCallback showURLCallback;
	JEPAppletCreatedCallback appletCreatedCallback;
	JEPAppletInitializedCallback appletInitializedCallback;
	JEPAppletStartedCallback appletStartedCallback;
	void *showStatusUserData;
	void *showURLUserData;
	void *appletCreatedUserData;
	void *appletInitializedUserData;
	void *appletStartedUserData;
@protected
	BOOL fNeedsCleanup;
	BOOL fAppletDestroyed;
	BOOL fPurgeOnAppletWindowCleanup;
}

+ (BOOL)checkBundle;
+ (BOOL)checkBundleWithVersion:(NSString *)versionString;
+ (void)hookExistingObjects;
+ (JEPController *)myController:(JEPController *)handle;
+ (BOOL)isControllerValid:(JEPController *)controller;
+ (void)purgeControllerArray;
+ (void)destroyControllerArray;
+ (void)purgeIfAppletDestroyedInWindow:(WindowRef)carbonWindow;
+ (void)disableScreenUpdates;
+ (void)enableScreenUpdates;
+ (BOOL)screenUpdatesDisabled;
+ (NSPoint)flipScreenPoint:(NSPoint)pointToFlip;
+ (NSPoint)flipWindowPoint:(NSPoint)pointToFlip window:(NSWindow *)window;
+ (NSPoint)flipViewPoint:(NSPoint)pointToFlip view:(NSView *)view;
+ (NSURL *)fixDocbase:(NSURL *)docbaseToFix;
+ (NSWindow *)nsWindowFromCarbonWindow:(WindowRef)carbonWindow;
+ (JEPController *)controllerFromNSWindow:(NSWindow *)window
	timeoutType:(LockTimeout)timeoutType;
+ (JEPController *)controllerFromAppletView:(AppletView *)appletView
	timeoutType:(LockTimeout)timeoutType;
+ (JEPController *)controllerFromAWTContext:(id)awtContext
	timeoutType:(LockTimeout)timeoutType;
+ (JEPController *)controllerFromCWindow:(id)cWindow
	timeoutType:(LockTimeout)timeoutType;
+ (JEPController *)controllerFromPluginInstance:(void *)instance
	timeoutType:(LockTimeout)timeoutType;
+ (NSMutableArray *)controllersInWindow:(WindowRef)carbonWindow
	timeoutType:(LockTimeout)timeoutType;
+ (void)allSubviews:(NSView *)view container:(NSMutableArray *)container;
+ (JEPController *)controllerFromNSView:(NSView *)view
	timeoutType:(LockTimeout)timeoutType;
+ (NSMutableArray *)controllersInNSView:(NSView *)view
	timeoutType:(LockTimeout)timeoutType;
+ (JEPController *)controllerContainingNSView:(NSView *)view
	timeoutType:(LockTimeout)timeoutType;
+ (void)sendToAllControllers:(NSInvocation *)message
	targetIsMyWindow:(BOOL)targetIsMyWindow
	timeoutType:(LockTimeout)timeoutType;
+ (void)sendToControllersInWindow:(NSInvocation *)message
	carbonWindow:(WindowRef)carbonWindow
	targetIsMyWindow:(BOOL)targetIsMyWindow
	timeoutType:(LockTimeout)timeoutType;
+ (void)sendToOtherControllersInWindow:(NSInvocation *)message
	sender:(JEPController *)sender carbonWindow:(WindowRef)carbonWindow
	targetIsMyWindow:(BOOL)targetIsMyWindow timeoutType:(LockTimeout)timeoutType;
+ (BOOL)windowHasKeyWindow:(WindowRef)windowRef
	timeoutType:(LockTimeout)timeoutType;
+ (BOOL)appletViewInWindowHasPoint:(WindowRef)windowRef point:(Point)point
	timeoutType:(LockTimeout)timeoutType;
+ (JEPController *)focusedControllerInWindow:(WindowRef)windowRef
	timeoutType:(LockTimeout)timeoutType;
+ (TSMDocumentID)getWindowTSMDocIfShouldBeActive:(WindowRef)windowRef
	timeoutType:(LockTimeout)timeoutType;
+ (void)routeEvent:(NSEvent *)theEvent origTarget:(NSWindow *)origTarget;
+ (void)routeMouseMovedEvent:(NSEvent *)theEvent
	targetController:(JEPController *)targetController
	targetWindow:(NSWindow *)targetWindow;
- (id)init;
- (void)dealloc;
- (BOOL)destroying;
- (OSStatus)handleCarbonUpdate:(EventRef)theEvent;
- (BOOL)purgeOnAppletWindowCleanup;
- (void)saveParentTSMState:(WindowRef)parentWindow;
- (void)copyParentTSMState:(JEPController *)controller;
- (void)setCarbonTSMState;
- (void)setCocoaTSMState;
- (void)activateCarbonInputContext;
- (void)activateCocoaInputContext;
- (NSDictionary *)fixAttributes:(NSDictionary *)attributesToFix
	parentWindow:(WindowRef)parentWindow;
- (AppletView *)startAppletInit:(NSURL *)docbase
	appletAttributes:(NSDictionary *)attributes;
- (void)setClip:(NSRect)clip;
- (void)refreshBounds;
- (OSStatus)setAppletState:(AppletStateJEP)newState;
- (void)drawApplet;
- (void)maybeDrawApplet;
- (void)suspendAppletUpdates;
- (void)resumeAppletUpdates;
- (void)sizeApplet:(NSSize)size;
- (void)positionApplet:(NSPoint)point;
- (void)showHideApplet:(BOOL)visible;
- (void)setAppletCreatedCallback:(JEPAppletCreatedCallback)infoFunction
	userData:(void *)userData;
- (void)setAppletInitializedCallback:(JEPAppletInitializedCallback)infoFunction
	userData:(void *)userData;
- (void)setAppletStartedCallback:(JEPAppletStartedCallback)infoFunction
	userData:(void *)userData;
- (void)setShowStatusCallback:(JEPShowStatusCallback)showStatusFunction
	userData:(void *)userData;
- (void)setShowURLCallback:(JEPShowDocumentCallback)showURLFunction
	userData:(void *)userData;
/* The following are all "private" methods (they're made private by not
 * being listed in this header file).
 */
//- (void)afterInitializeOnMainThread;
//- (void)maybeStartAppletOnMainThread
//- (void)showURL:(NSURL *)url inFrame:(NSString *)windowName;
//- (void)showStatus:(NSString *)statusMessage;
//- (void)windowWillClose:(NSNotification *)notification;
//- (void)destroyApplet;
//- (void)cleanupSelf;
//- (void)cleanupWindow;
//- (void)cleanupHandlers;
//- (void)cleanupTSMHandler;
//- (void)cleanupApplet;
//- (void)cleanupAppletView;
//- (void)unsetHasKeyWindow;
//- (void)setWindowClosing;
//- (void)unsetWindowClosing;
//- (void)setNeedsCleanup;
//- (void)setAppletDestroyed;
//- (unsigned)hash;
//- (BOOL)isEqual:(id)anObject;

@end

@interface JavaRunner : NSObject
{
@public
	jobject jrunnable;
	jthrowable exception;
	Boolean abort;
	Boolean done;
}

- (void)runJRunnableOnMainThread;

@end

@class ShowURLParams;
@class ShowStatusParams;
@class CreateJavaVMParams;
@class CreateCocoaWindowParams;
@class PerformWithNumberParams;
@class SizeParams;
@class PositionParams;
@class ClipParams;
@class MoveAndClipParams;
@class CarbonUpdateParams;
@class CocoaUpdateParams;

@interface NSAppShadow : NSObject
{
}

- (void)applicationWillTerminate:(NSNotification *)notification;
- (void)setMainThreadVariable;
- (void)setMainJavaThread;
- (void)cancelMenuTracking;
- (void)dismissJavaPopupMenu;
- (void)resetHybridView:(JEPController *)controller;
- (void)showURLOnMainThread:(ShowURLParams *)params;
- (void)showStatusOnMainThread:(ShowStatusParams *)params;
- (void)drawAppletOnMainThread:(JEPController *)controller;
- (void)suspendDrawAppletAfterDelay:(JEPController *)controller;
- (void)scheduleSuspendDrawApplet:(JEPController *)controller;
- (void)resumeDrawAppletAfterDelay:(JEPController *)controller;
- (void)scheduleResumeDrawApplet:(JEPController *)controller;
- (void)createJavaVMOnMainThread:(CreateJavaVMParams *)params;
- (void)createCocoaWindowOnMainThread:(CreateCocoaWindowParams *)params;
- (void)performWithNumberOnMainThread:(PerformWithNumberParams *)params;
- (void)sizeOnMainThread:(SizeParams *)params;
- (void)positionOnMainThread:(PositionParams *)params;
- (void)setClipOnMainThread:(ClipParams *)params;
- (void)moveAndClipOnMainThread:(MoveAndClipParams *)params;
- (void)postCarbonUpdateOnMainThread:(CarbonUpdateParams *)params;
- (void)doCarbonUpdateOnMainThread:(CarbonUpdateParams *)params;
- (void)postCocoaUpdateOnMainThread:(CocoaUpdateParams *)params;
- (void)doCocoaUpdateOnMainThread:(CocoaUpdateParams *)params;

@end

@interface ShowURLParams : NSObject
{
	NSURL *myURL;
	NSString *myWindowName;
	JEPController *myController;
}

- (id)initWithParams:(NSURL *)url windowName:(NSString *)windowName
	controller:(JEPController *)controller;
- (NSURL *)getURL;
- (NSString *)getWindowName;
- (JEPController *)getController;

@end

@interface ShowStatusParams : NSObject
{
	NSString *myMessage;
	JEPController *myController;
}

- (id)initWithParams:(NSString *)message controller:(JEPController *)controller;
- (NSString *)getMessage;
- (JEPController *)getController;

@end

@interface CreateJavaVMParams : NSObject
{
	JavaVMOption *myExtraOpts;
	int myExtraOptsNum;
	JavaVM **myVMPtr;
	JNIEnv *myEnvPtr;
	BOOL myInitAWT;
}

- (id)initWithParams:(JavaVMOption *)jvmExtraOpts
	extraOptsNum:(int)jvmExtraOptsNum vmPtr:(JavaVM **)vmPtr
	envPtr:(void **)envPtr initAWT:(BOOL)initAWT;
- (JavaVMOption *)getExtraOpts;
- (int)getExtraOptsNum;
- (JavaVM **)getVMPtr;
- (JNIEnv *)getEnvPtr;
- (BOOL)getInitAWT;

@end

@interface CreateCocoaWindowParams : NSObject
{
	JEPController *myController;
	NSDictionary *myAttributes;
	WindowRef myParentWindow;
	Rect myAppletClip;
	int myAppletPosX;
	int myAppletPosY;
	unsigned myAppletWidth;
	unsigned myAppletHeight;
}

- (id)initWithParams:(JEPController *)controller
	attributes:(NSDictionary *)attributes parentWindow:(WindowRef)parentWindow
	appletClip:(Rect)appletClip appletPosX:(int)appletPosX
	appletPosY:(int)appletPosY appletWidth:(unsigned)appletWidth
	appletHeight:(unsigned) appletHeight;
- (JEPController *)getController;
- (NSDictionary *)getAttributes;
- (WindowRef)getParentWindow;
- (Rect)getAppletClip;
- (int)getAppletPosX;
- (int)getAppletPosY;
- (unsigned)getAppletWidth;
- (unsigned)getAppletHeight;

@end

@interface PerformWithNumberParams : NSObject
{
	JEPController *myController;
	SEL mySelector;
	int myNumber;
}

- (id)initWithParams:(JEPController *)controller selector:(SEL)selector
	number:(int)number;
- (JEPController *)getController;
- (SEL)getSelector;
- (int)getNumber;

@end

@interface SizeParams : NSObject
{
	JEPController *myController;
	NSSize mySize;
}

- (id)initWithParams:(JEPController *)controller size:(NSSize)size;
- (JEPController *)getController;
- (NSSize)getSize;

@end

@interface PositionParams : NSObject
{
	JEPController *myController;
	NSPoint myPosition;
}

- (id)initWithParams:(JEPController *)controller position:(NSPoint)position;
- (JEPController *)getController;
- (NSPoint)getPosition;

@end

@interface ClipParams : NSObject
{
	JEPController *myController;
	NSRect myClip;
}

- (id)initWithParams:(JEPController *)controller clip:(NSRect)clip;
- (JEPController *)getController;
- (NSRect)getClip;

@end

@interface MoveAndClipParams : NSObject
{
	JEPController *myController;
	NSPoint myLocation;
	NSRect myClip;
}

- (id)initWithParams:(JEPController *)controller location:(NSPoint)location
	clip:(NSRect)clip;
- (JEPController *)getController;
- (NSPoint)getLocation;
- (NSRect)getClip;

@end

@interface CarbonUpdateParams : NSObject
{
	WindowRef myWindow;
	Rect *myUpdateRects;
	unsigned myNumUpdateRects;
}

- (id)initWithParams:(WindowRef)window updateRects:(Rect *)updateRects
	numUpdateRects:(unsigned)numUpdateRects;
- (WindowRef)getWindow;
- (Rect *)getUpdateRects;
- (unsigned)getNumUpdateRects;

@end

@interface CocoaUpdateParams : NSObject
{
	NSView *myView;
	NSRect myUpdateNSRects[4];
	unsigned myNumUpdateNSRects;
}

- (id)initWithParams:(NSView *)aView updateNSRects:(NSRect *)updateNSRects
	numUpdateNSRects:(unsigned)numUpdateNSRects;
- (NSView *)getView;
- (NSRect *)getUpdateNSRects;
- (unsigned)getNumUpdateNSRects;

@end

@interface PopupInfo : NSObject
{
	id myModel;
	NSMenu *myMenu;
	JEPController *myController;
	BOOL mVisible;
}

+ (void)addInfo:(PopupInfo *)info;
+ (void)removeInfoForController:(JEPController *)controller;
+ (PopupInfo *)infoForModel:(id)model;
+ (PopupInfo *)getVisible;
+ (void)setVisible:(id)model;
- (id)initWithParams:(id)model menu:(NSMenu *)menu controller:(JEPController *)controller;
- (id)getModel;
- (NSMenu *)getMenu;
- (JEPController *)getController;
- (BOOL)isVisible;
- (void)setVisible:(BOOL)visible;

@end

@interface ActivateInputContext : NSObject
{
	EventRef myEvent;
	EventQueueRef myEventQueue;
	JEPController *myController;
}

- (id)initWithParams:(EventQueueRef)eventQueue event:(EventRef)event
	controller:(JEPController *)controller;
- (void)waitForEventComplete;
- (void)performOnMainThread;

@end

@interface MyNSApplication : NSApplication
{
}

@end

@interface MyNSWindow : NSWindow
{
}

+ (WindowRef) existingCarbonWindowFor:(NSWindow *)cocoaWindow;

@end

@interface MyNSView : NSView
{
}

@end

/* Partial listing from output of class-dump on
 * /System/Library/Frameworks/AppKit.framework/AppKit on OS X 10.2.8
 * and 10.3.3.
 */
@interface NSCarbonWindow : NSWindow
{
@public
	struct OpaqueWindowPtr *_windowRef;
	char _windowRefIsOwned;
	void *_handleEventUPP;
	struct OpaqueEventHandlerRef *_eventHandler;
	char _passingCarbonWindowActivationEvents;
	char padding[512];	// In case future versions have more instance variables
}						// which aren't listed here

- (BOOL)reconcileToCarbonWindowBounds;
- (void)_postCarbonWindowActivateEvent:(BOOL)activate makeKeyWindow:(BOOL)makeKey;
- (void)handleCarbonBoundsChange;

@end

@interface NSInputContext : NSObject <NSTextInput>
{
    id _client;
    NSInputManager *_inputManager;
    id _keyBindingManager;
    id _keyBindingState;
    NSInputManager *_markingInputManager;
    struct {
        unsigned int _usingExtendedInputMethods:1;
        unsigned int _respondsToAcceptsMarkedText:1;
        unsigned int _disableEventProcessing:1;
        unsigned int _checkedClientKeyManager:1;
        unsigned int _reserved:28;
    } _icFlags;
}

+ (NSInputContext *)currentInputContext;
- (NSView *)client;
- (BOOL)_activate;
- (void)_deactivate;

@end

@interface NSTSMInputContext : NSInputContext
{
    void *_documentID;
    unsigned short _deadKeyChar;
    short _scriptID;
    struct {
        unsigned int _inPreProcess:1;
        unsigned int _acceptsGlyphInfo:1;
        unsigned int _respondsToString:1;
        unsigned int _respondsToTextStorage:1;
        unsigned int _hasLocking:1;
        unsigned int _unicodeInputEventReceived:1;
        unsigned int _lockCount:16;
        unsigned int _checkWindowLevel:1;
        unsigned int _respondsToLayoutManager:1;
        unsigned int _reserved:8;
    } _ticFlags;
	char padding[512];	// In case future versions have more instance variables
}						// which aren't listed here

@end

@interface MyNSTSMInputContext : NSTSMInputContext
{
}

@end

@interface NSViewAWTContext : NSObject
{
@public
    id client;
    BOOL alive;
    BOOL enabled;
    id cwindow;
    NSTimer *flusher;
    pthread_mutex_t lock;
    NSGraphicsContext *nsRef;
    struct CGContext *cgRef;
    BOOL dirty;
}

@end

@interface CPerformer : NSObject
{
@public
    id fTarget;
    SEL fSelector;
    id fArg;
}

@end

#ifdef __cplusplus
}
#endif
