//Copyright (c) 2005, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

package apple.awt;

import jep.*;
import java.awt.*;
import java.lang.ref.*;
import sun.java2d.SunGraphics2D;

// Prior to JEP 0.9.1 we clipped a Graphics object's devClip just once,
// in MyCAppletFrame's, MyCCanvas's and MyCPanel's getGraphics() methods.
// But some image objects appear to keep their Graphics objects for a long
// time, so we need to clip their devClips more often to keep their clipping
// accurate (examples that demonstrate this are http://www.niftyfiftys.com/
// and http://www.homestead.com/prosites-niyh/hom_eng.html.)

public class GraphicsClipper {
	protected static boolean showDebugInfo = false;
	protected WeakReference graphics = null;
	protected AppletFrame ancestor = null;
	protected ComponentModel peer = null;
	protected Object devClipLock = new Object();

	public static GraphicsClipper createNew(SunGraphics2D graphics,
		 AppletFrame ancestor, ComponentModel peer)
	{
		return new GraphicsClipper(graphics, ancestor, peer);
	}

	public GraphicsClipper(SunGraphics2D graphics, AppletFrame ancestor,
		ComponentModel peer)
	{
		this.graphics = new WeakReference(graphics);
		this.ancestor = ancestor;
		this.peer = peer;
		ancestor.addClipper(this);
	}

	public Rectangle clipToApplet(int x, int y, int width, int height) {
		synchronized (devClipLock) {
			if (ancestor.destroying || (peer.fTarget == null) ||
				(graphics.get() == null))
			{
				return new Rectangle();
			}
			if (!AppletHandlerFactory.
				isAppletViewDisplayable(ancestor.getCocoaParentViewLong()))
			{
				if (showDebugInfo)
					System.err.println("GraphicsClipper.clipToApplet(): Zero-clipping because applet invisible");
				return new Rectangle();
			}
			if (ancestor.updatesOff || ancestor.asyncUpdatesOff) {
				if (showDebugInfo)
					System.err.println("GraphicsClipper.clipToApplet(): Zero-clipping because updates off");
				return new Rectangle();
			}
			Rectangle appletClip = null;
			Rectangle bounds = new Rectangle(x, y, width, height);
			try {
				appletClip = ancestor.translateCurrentClip(peer.fTarget).
					intersection(bounds);
				if (appletClip.isEmpty())
					appletClip.x = appletClip.y = appletClip.width = appletClip.height = 0;
			} catch (Throwable e) {
				appletClip = new Rectangle();
			}
			if (showDebugInfo) {
				System.err.println("GraphicsClipper.clipToApplet(): Clipping from " +
					bounds.toString() + " to " + appletClip.toString());
			}
			return appletClip;
		}
	}

	public Rectangle clipToApplet(Rectangle r) {
		return clipToApplet(r.x, r.y, r.width, r.height);
	}

	public void validateDevClip() {
		synchronized(devClipLock) {
			if (ancestor.destroying || (peer.fTarget == null)) {
				zeroDevClip();
				return;
			}
			SunGraphics2D g = (SunGraphics2D) graphics.get();
			if (g == null)
				return;
			Rectangle bounds = null;
			try {
				bounds = g.getSurfaceData().getBounds();
			} catch (NullPointerException e) {
				zeroDevClip();
				return;
			}
			Rectangle newDevClip = clipToApplet(bounds);
			if (!AppletHandlerFactory.isJava142()) {
				g.setDevClip(newDevClip);
				return;
			}
			Rectangle oldDevClip = getDevClip();
			if ((oldDevClip.x != newDevClip.x) ||
				(oldDevClip.y != newDevClip.y) ||
				(oldDevClip.width != newDevClip.width) ||
				(oldDevClip.height != newDevClip.height))
			{
				g.setDevClip(newDevClip);
			}
		}
	}

	public void zeroDevClip() {
		synchronized(devClipLock) {
			SunGraphics2D g = (SunGraphics2D) graphics.get();
			if (g == null)
				return;
			g.setDevClip(new Rectangle());
		}
	}

	// Override for Java 1.4.2 and above.
	public Rectangle getDevClip() {
		return null;
	}

	public SunGraphics2D getGraphics() {
		return (SunGraphics2D) graphics.get();
	}

	public String toString() {
		return getClass().getName() + " (jep)";
	}

	static {
		if ("true".equalsIgnoreCase(System.getProperty("jep.debuginfo", "false"))) {
			showDebugInfo = true;
		}
	}

}

