// Copyright (c) 2006, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

package jep;

import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.security.*;
import java.lang.reflect.*;
import sun.awt.AppContext;
import sun.awt.SunToolkit;

// This class is needed to deal with a problem in Tiger (all Java versions):
// LiveConnect calls that trigger Java drawing can no longer take place on
// the main thread (otherwise you get hangs).  We resolve this by creating
// a special LiveConnect event dispatching thread and using it to process
// LiveConnect proxy calls, while waiting on the main thread for the results.
// Since we can't block the main thread, we must also process main thread
// events while we're waiting.  This process is _very_ costly (hundreds of
// times slower than just using the main thread), so it's fortunate that only
// some LiveConnect calls (the CallMethod ones) might trigger Java drawing,
// and so need this treatment.

// As of JEP 0.9.5+f, we process as many LiveConnect proxy calls as possible
// in the context of the applet on which these calls are made.  The rest still
// run on the LiveConnect event dispatching thread.

public class LiveConnect {
	protected static Object fakeSource = new Object();
	protected static ThreadGroup ourThreadGroup = null;
	protected static AppContext ourAppContext = null;

	public static void initialize() {
		if (ourThreadGroup == null)
			ourThreadGroup = new ThreadGroup("LiveConnectJEP-threadGroup");
		if (ourAppContext == null) {
			InitLiveConnect init = new InitLiveConnect(ourThreadGroup);
			synchronized(init.done) {
				init.start();
				try {
					init.done.wait();
				} catch (InterruptedException e) {
				}
				ourAppContext = init.appContext;
			}
			// Wake up (i.e. create) our event dispatcher thread, if need be.
			SunToolkit.postEvent(ourAppContext, new EmptyEvent());
		}
	}

	public static void destroy() {
		if (ourAppContext != null) {
			try {
				ourAppContext.dispose();
			} catch (Throwable e) {
			}
			ourAppContext = null;
			ourThreadGroup = null;
		} else if (ourThreadGroup != null) {
			try {
				ourThreadGroup.destroy();
			} catch (Throwable e) {
			}
			ourThreadGroup = null;
		}
	}

	// As of JEP 0.9.6.1 we use a "jobID" to identify each of the "jobs"
	// that's currently using processMainThreadEvents().  This allows us
	// re-enter doProxy() (and processMainThreadEvents()) with impunity --
	// Each "job" that interrupts previous one(s) will be processed to
	// completion; then the previous "job" will resume.  These "jobs" might
	// not run (on other threads) in exactly this order, but we can guarantee
	// that nested calls to doProxy() (and processMainThreadEvents()) will
	// return in the order expected, and only after their respective "jobs"
	// have completed (on other threads).
	private static void doProxy(Class proxyClass, Runnable runnable,
		Applet applet)
	{
		// If possible, process our proxy call in its applet's context.
		AppContext context = SunToolkit.targetToAppContext(applet);
		if (context == null)
			context = ourAppContext;
		Object source = ((applet != null) ? applet : fakeSource);
		long jobID = AppletHandlerFactory.beforeProcessing(runnable);
		SunToolkit.postEvent(context, new InvocationEvent(source,
			new DoProxy(proxyClass, runnable, jobID)));
		AppletHandlerFactory.processMainThreadEvents(0, jobID);
	}

	private static class DoProxy implements Runnable {
		private final Class proxyClass;
		private final Runnable runnable;
		private final long jobID;

		DoProxy(Class proxyClass, Runnable runnable, long jobID) {
			this.proxyClass = proxyClass;
			this.runnable = runnable;
			this.jobID = jobID;
		}

		public void run() {
			try {
				Class[] parmTypes = {Runnable.class};
				Object[] parms = {runnable};
				proxyClass.getMethod("run", parmTypes).invoke(null, parms);
			} catch (Throwable e) {
				System.err.println("LiveConnect event dispatcher caught exception: " +
					e.toString());
				e.printStackTrace();
			}
			AppletHandlerFactory.stopProcessing(jobID);
		}

	}

	private static class EmptyEvent extends AWTEvent implements ActiveEvent {

		public EmptyEvent() {
			super(fakeSource, 0);
		}

		public void dispatch() { }

	}

	private static class InitLiveConnect extends Thread {
		Object done = new Object();
		AppContext appContext = null;

		InitLiveConnect(ThreadGroup group) {
			super(group, "InitLiveConnect");
		}

		public void run() {
			synchronized(done) {
				appContext = SunToolkit.createNewAppContext();
				// We need to do this before our event queue's event dispatch
				// thread has started.  Otherwise that thread will die when
				// we call EventQueue.push(), and our ThreadGroup will also
				// die (since the event dispatch thread is its only thread),
				// which means EventQueue.initDispatchThread() will never be
				// able to create another event dispatch thread for our queue,
				// which means that there will be hangs when doProxy() can't
				// find a suitable applet AppContext and tries to post a job
				// to ourAppContext.
				EventQueue ourQueue = (EventQueue)
					appContext.get(AppContext.EVENT_QUEUE_KEY);
				MySimpleEventQueue newQueue = new MySimpleEventQueue();
				ourQueue.push(newQueue);
				appContext.put(AppContext.EVENT_QUEUE_KEY, newQueue);
				done.notifyAll();
			}
		}

	}

}

