//Copyright (c) 2004, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

package apple.awt;

import jep.*;
import java.util.*;
import java.awt.*;
import java.security.AccessControlException;

// This class times out calls to get or post AWT events after a fixed (and
// fairly lengthy) interval.  It does this by making the calls on a different
// thread, then "wait"ing until either the call has finished or the interval
// has elapsed.  To make sure that a caller experiences only its "own" delays,
// each calling thread gets its own "get-post" thread.  If the interval has
// elapsed without the call returning, the "get-post" thread is interrupted,
// then killed, and replaced by a new "get-post" thread.  As far as I can
// tell, doing this stops the (unfinished) call to our super-class's get or
// post method from having any side effects.  (I wish the same thing were
// possible in C++ in Mozilla's NSPR environment -- though it _is_ possible
// (with some contortions) to create an interruptible "get-post" thread, if
// you interrupt it there's no good way to clean up the event that's currently
// being processed.)

// I made these changes (new in JEP 0.8.3) after I noticed that, especially
// on OS X 10.3.X, Apple's JVM has a tendency to get stuck in PeerPaintEvent.
// paintFromCocoaDrawRect() on a call to waitForDispatch().  My changes
// seem to eliminate this particular hang, and reduce the frequency of UI
// hangs in general, but don't eliminate them (at least on OS X 10.3.X).
// (You can unfreeze a hang at waitForDispatch() by calling EventFactoryProxy.
// setCurrentEvent(null) on a different thread, as I do in mungeEvent()
// below.)

// As of Java 1.4.2 Update 1, this whole business now seems unnecessary.

public class MyEventQueue extends MyEventQueue2 {
	private static MyEventQueue topQueue = null;
	public static final int GETPOST_TIMEOUT = 100;		// In milliseconds
	public static final int POSTPAINT_TIMEOUT = 20;		// In milliseconds
	protected final HashMap mungerTable = new HashMap();
	protected LatestMunger latestMunger = null;
	// safeToLock is set externally (from AppletView.m's
	// Java_java_lang_Object_wait_redirect()).
	private boolean safeToLock = false;

	public MyEventQueue() {
	}

	// getMunger can sometimes throw a spurious AccessControlException when
	// called via LiveConnect (the security manager thinks that the MRJ
	// Plugin's LiveConnect proxy shouldn't be allowed to create a new
	// thread).  Methods that call this one should catch this error and
	// deal with it.
	protected EventMunger getMunger() {
		Thread current = Thread.currentThread();
		EventMunger munger =
			((latestMunger != null) && (latestMunger.key == current)) ?
			latestMunger.value : null;
		if (munger == null) {
			synchronized (mungerTable) {
				munger = (EventMunger) mungerTable.get(current);
				if (munger == null) {
					// As of JEP 0.9.5+h, the EventMunger class extends Thread
					// instead of implementing Runnable.  This is because (very
					// bizarrely) "new Thread(munger)" sometimes hangs, but
					// "new EventMunber()" and "munger.start()" never seem to.
					munger = new EventMunger();
					if (munger != null) {
						mungerTable.put(current, munger);
						munger.start();
					}
				}
				if (munger == null) {
					latestMunger = null;
				} else {
					latestMunger = new LatestMunger(current, munger);
				}
			}
		}
		return munger;
	}

	protected void removeCurrentMunger() {
		synchronized (mungerTable) {
			Thread current = Thread.currentThread();
			mungerTable.remove(current);
			if ((latestMunger != null) && (latestMunger.key == current))
				latestMunger = null;
			cleanMungerTable();
		}
	}

	protected void cleanMungerTable() {
		synchronized (mungerTable) {
			Object[] keys = mungerTable.keySet().toArray();
			for (int i = 1; i <= keys.length; ++i) {
				Thread key = (Thread) keys[i-1];
				if (!key.isAlive()) {
					EventMunger munger =
						(EventMunger) mungerTable.remove(key);
					if (munger != null)
						munger.killThread();
					if ((latestMunger != null) && (latestMunger.key == key))
						latestMunger = null;
				}
			}
		}
	}

	public AWTEvent getNextEvent() throws InterruptedException {
		AWTEvent retval = null;
		try {
			retval = getMunger().getNextEvent();
		} catch (AccessControlException e) {
			return new EmptyEvent();
		}
		if (retval != null)
			return retval;
		return new EmptyEvent();
	}

	public AWTEvent getNextEvent(int id) throws InterruptedException {
		AWTEvent retval = null;
		try {
			retval = getMunger().getNextEvent(id);
		} catch (AccessControlException e) {
			return new EmptyEvent();
		}
		if (retval != null)
			return retval;
		return new EmptyEvent();
	}

	public AWTEvent peekEvent() {
		AWTEvent retval = null;
		try {
			retval = getMunger().peekEvent();
		} catch (AccessControlException e) {
			return null;
		}
		return retval;
	}

	public AWTEvent peekEvent(int id) {
		AWTEvent retval = null;
		try {
			retval = getMunger().peekEvent(id);
		} catch (AccessControlException e) {
			return null;
		}
		return retval;
	}

	public void postEvent(AWTEvent theEvent) {
		try {
			getMunger().postEvent(theEvent);
		} catch (AccessControlException e) {
		}
	}

	protected AWTEvent superGetNextEvent() throws InterruptedException {
		return super.getNextEvent();
	}

	protected AWTEvent superGetNextEvent(int id) throws InterruptedException {
		return callGetNextEvent(id);
	}

	protected AWTEvent superPeekEvent() {
		return super.peekEvent();
	}

	protected AWTEvent superPeekEvent(int id) {
		return super.peekEvent(id);
	}

	protected void superPostEvent(AWTEvent theEvent) {
		super.postEvent(theEvent);
	}

	// Don't let an applet "pop()" back above us.
	public synchronized void push(EventQueue newEventQueue) {
		if (topQueue == null)
			topQueue = this;
		super.push(newEventQueue);
	}

	protected void pop() throws EmptyStackException {
		if ((topQueue == null) || (topQueue == this))
			throw new EmptyStackException();
		super.pop();
	}

	protected void dispatchEvent(AWTEvent event) {
		long pool = AppletHandlerFactory.allocAutoreleasePool();
		try {
			Object source = event.getSource();
			Component ancestor = (Component) source;
			for (;;) {
				Component next = (Component) ancestor.getParent();
				if (next == null)
					break;
				ancestor = next;
			}
			if (ancestor instanceof AppletFrame)
				AppletFrame.setCurrentAppletFrame((AppletFrame) ancestor);
		} catch (Throwable e) {
		}
		super.dispatchEvent(event);
		AppletFrame.setCurrentAppletFrame(null);
		AppletHandlerFactory.releaseAutoreleasePool(pool);
	}

	class EventMunger extends Thread {
		private long timeout = GETPOST_TIMEOUT;
		private Thread ourThread = null;
		private boolean dieNow = false;
		private int idIn = 0;
		private boolean useId = false;
		private boolean doPeek = false;
		private boolean doPost = false;
		private AWTEvent eventIn = null;
		private AWTEvent eventOut = null;
		private InterruptedException interruptedOut = null;
		private ThreadDeath threadDeathOut = null;
		private boolean succeeded = false;
		private boolean threadStarted = false;
		private boolean isReady = false;

		public EventMunger() {
		}

		private void perEventInitialization() {
			idIn = 0;
			useId = false;
			doPeek = false;
			doPost = false;
			eventIn = null;
			eventOut = null;
			interruptedOut = null;
			threadDeathOut = null;
			succeeded = false;
		}

		private boolean waitTilReady() throws InterruptedException {
			try {
				timeout = GETPOST_TIMEOUT;
				while (!threadStarted) {
					Thread.sleep(2);
					if ((timeout -= 2) <= 0) {
						hardReset();
						return false;
					}
				}
				while (!isReady) {
					Thread.sleep(5);
					if ((timeout -= 5) <= 0) {
						hardReset();
						return false;
					}
				}
				if (timeout > 0)
					return true;
				hardReset();
				return false;
			} catch (InterruptedException e) {
				hardReset();
				throw e;
			} catch (ThreadDeath e) {
				hardReset();
				throw e;
			}
		}

		public AWTEvent getNextEvent() throws InterruptedException {
			if (!waitTilReady())
				return null;
			try {
				perEventInitialization();
				return mungeEvent(true);
			} catch (InterruptedException e) {
				hardReset();
				throw e;
			} catch (ThreadDeath e) {
				hardReset();
				throw e;
			}
		}

		public AWTEvent getNextEvent(int id) throws InterruptedException {
			if (!waitTilReady())
				return null;
			try {
				perEventInitialization();
				idIn = id;
				useId = true;
				return mungeEvent(true);
			} catch (InterruptedException e) {
				hardReset();
				throw e;
			} catch (ThreadDeath e) {
				hardReset();
				throw e;
			}
		}

		public AWTEvent peekEvent() {
			try {
				if (!waitTilReady())
					return null;
			} catch (InterruptedException e) {
				throw new ThreadDeath();
			}
			try {
				perEventInitialization();
				doPeek = true;
				return mungeEvent(true);
			} catch (InterruptedException e) {
				hardReset();
				// We can't throw an InterruptedException here, so we're
				// forced to throw a ThreadDeath.
				throw new ThreadDeath();
			} catch (ThreadDeath e) {
				hardReset();
				throw e;
			}
		}

		public AWTEvent peekEvent(int id) {
			try {
				if (!waitTilReady())
					return null;
			} catch (InterruptedException e) {
				throw new ThreadDeath();
			}
			try {
				perEventInitialization();
				idIn = id;
				useId = true;
				doPeek = true;
				return mungeEvent(true);
			} catch (InterruptedException e) {
				hardReset();
				// We can't throw an InterruptedException here, so we're
				// forced to throw a ThreadDeath.
				throw new ThreadDeath();
			} catch (ThreadDeath e) {
				hardReset();
				throw e;
			}
		}

		public void postEvent(AWTEvent theEvent) {
			try {
				if (!waitTilReady())
					return;
			} catch (InterruptedException e) {
				throw new ThreadDeath();
			}
			try {
				perEventInitialization();
				eventIn = theEvent;
				doPost = true;
				if ((AppletHandlerFactory.PaintEventClass != null) &&
					(AppletHandlerFactory.PaintEventClass.isInstance(theEvent)))
					timeout = POSTPAINT_TIMEOUT;
				mungeEvent(true);
			} catch (InterruptedException e) {
				hardReset();
				// We can't throw an InterruptedException here, so we're
				// forced to throw a ThreadDeath.
				throw new ThreadDeath();
			} catch (ThreadDeath e) {
				hardReset();
				throw e;
			}
		}

		private void killThread() {
			dieNow = true;
			if ((ourThread != null) && ourThread.isAlive())
				ourThread.interrupt();
		}

		// "Reset" our thread by interrupting it and (later) replacing it
		// with a new one.
		private void hardReset() {
			killThread();
			try {
				Class[] parmTypes = {Object.class};
				Object[] parms = new Object[1];
				EventFactoryProxy.class.
					getMethod("setCurrentEvent", parmTypes).invoke(null, parms);
			} catch (Throwable e) {
			}
			MyEventQueue.this.removeCurrentMunger();
		}

		// "Reset" our thread by posting an empty event (if this can be done
		// without risking a hang).  Otherwise do a "hard reset".  A "soft
		// reset" is less disruptive, since (if it works) we can keep using
		// the same thread.  The value of safeToLock is determined externally
		// (in AppletView.m by Java_java_lang_Object_wait_redirect()).
		// safeToLock is true if and only if our queue (MyEventQueue.this) is
		// "wait"ing (in getNextEvent()) on our EventMunger thread.
		private void softReset() {
			if (safeToLock) {
				synchronized (MyEventQueue.this) {
					// We need to avoid calling SunToolkit.flushPendingEvents()
					// -- which is called by the public postEvent() method, and
					// can lead to deadlocks.
					MyEventQueue.this.callPostEventPrivate(new EmptyEvent());
				}
			} else {
				hardReset();
			}
		}

		private AWTEvent mungeEvent(boolean wait) throws InterruptedException {
			synchronized (this) {
				if (!isReady)
					return null;
				notifyAll();
				if (!wait)
					return null;
				wait(timeout);
			}
			if (succeeded) {
				return eventOut;
			} else if (interruptedOut != null) {
				hardReset();
				throw interruptedOut;
			} else if (threadDeathOut != null) {
				hardReset();
				throw threadDeathOut;
			}
			softReset();
			return null;
		}

		// Prior to JEP 0.9.6.1, almost all of the following was inside a
		// single "synchronized (this)" statement.  But this tended to cause
		// hangs while destroying applets (Thread.stop() would block trying
		// to get a lock on the EventMunger thread).  Now I'm using two
		// "synchronized (this)" blocks.  I also now handle the ThreadDeath
		// exception explicitly.  (ThreadDeath is used by Thread.stop(), which
		// can in turn (a bit surprisingly) be used by AppContext.dispose(),
		// which can be called from AppletClassLoader.release() (or
		// PluginClassLoader.release()), called from AppletPanel.release().)
		public void run() {
			ourThread = Thread.currentThread();
			while (!dieNow) {
				synchronized (this) {
					try {
						threadStarted = true;
						isReady = true;
						wait();
						isReady = false;
					} catch (InterruptedException e) {
						isReady = false;
						ourThread = null;
						interruptedOut = e;
						return;
					} catch (ThreadDeath e) {
						isReady = false;
						ourThread = null;
						threadDeathOut = e;
						return;
					} catch (Throwable e) {
						isReady = false;
						ourThread = null;
						System.out.println("EventMunger.run(): Unexpected error " + e.getMessage());
						e.printStackTrace();
						return;
					}
				}
				try {
					if (doPost) {
						MyEventQueue.this.superPostEvent(eventIn);
					} else if (doPeek) {
						if (useId) {
							eventOut = MyEventQueue.this.superPeekEvent(idIn);
						} else {
							eventOut = MyEventQueue.this.superPeekEvent();
						}
					} else {
						if (useId) {
							eventOut = MyEventQueue.this.superGetNextEvent(idIn);
						} else {
							eventOut = MyEventQueue.this.superGetNextEvent();
						}
					}
					succeeded = true;
				} catch (InterruptedException e) {
					ourThread = null;
					interruptedOut = e;
					return;
				} catch (ThreadDeath e) {
					ourThread = null;
					threadDeathOut = e;
					return;
				} catch (Throwable e) {
					ourThread = null;
					System.out.println("EventMunger.run(): Unexpected error " + e.getMessage());
					e.printStackTrace();
					return;
				}
				synchronized(this) {
					notifyAll();
				}
			}
			ourThread = null;
		}

	}

	class EmptyEvent extends AWTEvent implements ActiveEvent {

		public EmptyEvent() {
			super(MyEventQueue.this, 0);
		}

		public void dispatch() {}

	}

	class LatestMunger {
		Thread key;
		EventMunger value;

		LatestMunger(Thread key, EventMunger value) {
			this.key = key;
			this.value = value;
		}

	}

}

