//Copyright (c) 2007, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// This class exists to deal with bugs in Apple's JVMs on OS X 10.2.8 and
// 10.3.X -- these JVMs often fail to allocate autorelease pools when they
// should.  This sometimes causes Cocoa objects to be leaked.  But more
// often it just causes multiple copies of this annoying message to be
// displayed in the system console:
//
//   ** ATTENTION ***
//
//   Please Attach Debugger : No Autorelease pool exists
//   Press any key to continue...
//
// I work around the problem by overriding most Thread objects' run() methods
// with RunnableWrapper.run(), which allocates and releases autorelease pools
// when it should.
//
// For Thread objects created by passing a Runnable to the Thread constructor
// the workaround is quite straightforward -- I substitute a RunnableWrapper
// for the original Runnable, and call the original Runnable from
// RunnableWrapper.run().  The workaround for Thread object subclasses is more
// complicated:  I start a RunnableWrapper.WrapperThread in place of the
// original Thread subclass object, and call the original Thread's run()
// method from the RunnableWrapper associated with the WrapperThread object.
// This requires some low-level trickery in poseAsThread() and unposeThreads().
//
// RunnableWrapper and WrapperThread objects are created (and started) in
// AppletView.m's Java_java_lang_Thread_start_redirect().

package jep;

public class RunnableWrapper implements Runnable {
	private int timesRun = 0;
	private Runnable oldTarget = null;
	private Runnable runnable = null;
	private Thread newThread = null;

	private RunnableWrapper(Runnable runnable) {
		this.runnable = runnable;
	}

	private RunnableWrapper(Thread oldThread, Runnable oldTarget) {
		runnable = oldThread;
		this.oldTarget = oldTarget;
	}

	private static RunnableWrapper newWrapper(Runnable runnable) {
		return new RunnableWrapper(runnable);
	}

	private static RunnableWrapper newWrapper(Thread oldThread,
		Runnable oldTarget)
	{
		RunnableWrapper wrapper = new RunnableWrapper(oldThread, oldTarget);
		Thread newThread = new WrapperThread(oldThread.getThreadGroup(),
			wrapper, oldThread.getName());
		newThread.setDaemon(oldThread.isDaemon());
		newThread.setPriority(oldThread.getPriority());
		wrapper.newThread = newThread;
		return wrapper;
	}

	private Thread getOldThread() {
		try {
			return (Thread) runnable;
		} catch (ClassCastException e) {
			return null;
		}
	}

	private Thread getNewThread() {
		return newThread;
	}

	public void run() {
		if (runnable != null) {
			// If an applet-created Thread subclass calls super.run() (or the
			// equivalent) from its run() method, our own run() method will be
			// entered in an infinite loop (unless we do something about it).
			// We use the timesRun variable to stop this from happening.  This
			// change (new as of JEP 0.9.6.2) resolves bmo bug 377215.
			// It's not clear why RunnableWrapper gets treated as the
			// superclass of every applet-created subclass.  But WrapperThread
			// (the class of the newThread which is posing as an oldThread) is
			// a member class of RunnableWrapper, and RunnableWrapper is one
			// step above the Thread subclass's run() method in the execution
			// stack.
			if (++timesRun == 1) {
				long pool = AppletHandlerFactory.allocAutoreleasePool();
				try {
					runnable.run();
				} catch (ThreadDeath e) {
				} catch (Throwable e) {
					System.out.println("RunnableWrapper.run(): Unexpected error " + e.getMessage());
					e.printStackTrace();
				} finally {
					AppletHandlerFactory.releaseAutoreleasePool(pool);
					if (newThread != null) {
						Thread oldThreadCopy = getOldThread();
						Thread newThreadCopy = newThread;
						runnable = null;
						newThread = null;
						unposeThreads(newThreadCopy, oldThreadCopy);
					} else {
						runnable = null;
					}
				}
			} else if (oldTarget != null) {
				// If oldThread has a 'target', call it -- it may actually do
				// something.
				oldTarget.run();
			}
		}
	}

	private native static void poseAsThread(Thread poser, Thread target);
	private native static void unposeThreads(Thread poser, Thread target);

	public static class WrapperThread extends Thread {
		private int timesRun = 0;
		private RunnableWrapper wrapper = null;

		private WrapperThread(ThreadGroup group, RunnableWrapper wrapper,
			String name)
		{
			super(group, wrapper, name);
			this.wrapper = wrapper;
		}

		private RunnableWrapper getWrapper() {
			return wrapper;
		}

		public void run() {
			if (wrapper != null) {
				// It doesn't seem that we need to use timesRun here (as we do
				// in RunnableWrapper.run()).  But just in case ...
				if (++timesRun == 1) {
					synchronized (this) {
						poseAsThread(this, wrapper.getOldThread());
					}
				}
				wrapper.run();
				wrapper = null;
			}
		}

		public String toString() {
			return super.toString() + " (WrapperThread)";
		}

	}

}
