//Copyright (c) 2004, Steven Michaud, All Rights Reserved
//
//Permission is hereby granted, free of charge, to any person obtaining
//a copy of this software and associated documentation files (the
//"Software"), to deal in the Software without restriction, including
//without limitation the rights to use, copy, modify, merge, publish,
//distribute, sublicense, and/or sell copies of the Software, and to
//permit persons to whom the Software is furnished to do so, subject to
//the following conditions:
//
//The above copyright notice and this permission notice shall be
//included in all copies or substantial portions of the Software.
//
//THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
//EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
//MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
//CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
//TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
//SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// This class needs to be in the java.awt package so that we can have access
// to Component.valid (which has no access modifier).
package java.awt;

import jep.*;
import java.applet.Applet;
import sun.awt.AppContext;
import sun.awt.PeerEvent;
import sun.awt.SunToolkit;

// Container objects (Windows, Frames, Panels and so forth) always have a
// "layout manager", which can use very elaborate means to decide how to lay
// out their contents.  But sometimes you just want the damn thing laid out
// exactly as you specify -- which is generally speaking impossible with any
// of the layout managers built in to Java.  That's what this class is for.
// We use it to replace the original layout managers of our applet's parent
// (a Panel object) and grandparent (a Frame object), so that when we mess
// with our applet's bounds and position to do clipping, these objects' layout
// managers won't interfere with our efforts.

// As of version 0.8.2 we've made this class less "simple" (i.e. less brain-
// dead) than it was previously.  The addition of "layout" code here, plus
// the fact that AppletFrame.imposeClip() no longer clips the applet's right
// or bottom, makes our clipping smoother and more accurate.

public class SimpleLayout implements LayoutManager2 {
	protected static boolean showDebugInfo = false;
	protected AppletFrame host = null;
	protected int width = 0;
	protected int height = 0;

	public SimpleLayout() {
	}

	public SimpleLayout(int width, int height, AppletFrame host) {
		this.width = width;
		this.height = height;
		this.host = host;
	}

	// See resizeApplet() in AppletFrame.java.
	public void resetSize(int width, int height) {
		this.width = width;
		this.height = height;
	}

	public int getWidth() {
		return width;
	}

	public int getHeight() {
		return height;
	}

	public void addLayoutComponent(String name, Component comp) {
	}

	public void addLayoutComponent(Component comp, Object constraints) {
	}

	public void removeLayoutComponent(Component comp) {
	}

	public Dimension preferredLayoutSize(Container target) {
		Dimension retval;
		if ((target instanceof AppletFrame) ||
			(target instanceof AppletFramePanel) ||
			(target instanceof Applet))
			retval = new Dimension(width, height);
		else {
			retval = target.getSize();
		}
		if (showDebugInfo) {
			System.err.println("SimpleLayout preferredLayoutSize(): " +
				"Container " + ((target == null) ? "null" : target.toString()) +
				"returning " + retval.toString());
		}
		return retval;
	}

	public Dimension minimumLayoutSize(Container target) {
		return preferredLayoutSize(target);
	}

	public Dimension maximumLayoutSize(Container target) {
		return preferredLayoutSize(target);
	}

	public float getLayoutAlignmentX(Container target) {
		return 0;
	}

	public float getLayoutAlignmentY(Container target) {
		return 0;
	}

	// This method shouldn't be necessary, but at least one bizarrely written
	// (and very popular) applet seems to require it -- the one that powers
	// http://www.map24.com's "interactive" maps.  Without these calls to
	// setVisible(), every map becomes invisible the first time the page is
	// resized -- because all its elements are "hidden".
	// As of JEP 0.8.9 I had to tone down this fix somewhat, because the
	// original broke a _lot_ of other applets.  The map24.com interactive
	// maps still seem to work ... at least most of the time. The applets
	// broken by the previous map24.com fix include:
	// http://www.xrce.xerox.com/competencies/content-analysis/arabic/input/keyboard_input.html
	// http://www.boursorama.com/graphiques/graphique_histo.phtml?symbole=INTC
	// http://webbook.nist.gov/cgi/fluid.cgi?ID=C7732185&TUnit=K&PUnit=MPa&DUnit=mol%2Fl&HUnit=kJ%2Fmol&WUnit=m%2Fs&VisUnit=uPa*s&STUnit=N%2Fm&Type=IsoTherm&RefState=DEF&Action=Page
	// Many of the demo applets at http://www.javaview.de/
	// As of JEP 0.9.6 (0.9.5+d) I toned down this fix yet again.  It used to
	// enable and make visible all disabled components.  Now it enables only
	// the applet's top-level subcomponents (though it continues to make
	// visible other disabled components).  This fixes
	// http://sourceforge.net/forum/forum.php?thread_id=1448187&forum_id=371125.
	protected static void enableDisabledComponents(Component components[]) {
		for (int i = 1; i <= components.length; ++i) {
			if (!components[i-1].isEnabled()) {
				components[i-1].setEnabled(true);
				components[i-1].setVisible(true);
			}
			try {
				makeVisibleDisabledComponents(((Container)components[i-1]).getComponents());
			} catch (ClassCastException e) {
			}
		}
	}
	protected static void makeVisibleDisabledComponents(Component components[]) {
		for (int i = 1; i <= components.length; ++i) {
			if (!components[i-1].isEnabled()) {
				components[i-1].setVisible(true);
			}
			try {
				makeVisibleDisabledComponents(((Container)components[i-1]).getComponents());
			} catch (ClassCastException e) {
			}
		}
	}

	protected static void validateAllComponents(Component components[]) {
		for (int i = 1; i <= components.length; ++i) {
			components[i-1].valid = true;
			try {
				validateAllComponents(((Container)components[i-1]).getComponents());
			} catch (ClassCastException e) {
			}
		}
	}

	public void layoutContainer(Container target) {
		if (target instanceof AppletFrame) {
			target.setBounds(0, 0, width, height);
			target.valid = true;
		} else if (target instanceof AppletFramePanel) {
			host.setBounds(0, 0, width, height);
			if (AppletHandlerFactory.isJava142Update1()) {
				target.setBounds(0, 0, width, height);
			} else {
				target.setBounds(0, 22, width, height);
			}
			Applet applet = host.getApplet();
			if (applet != null) {
				applet.setBounds(0, 0, width, height);
				enableDisabledComponents(applet.getComponents());
				// We don't expect this to be called from anything but the
				// applet's context.  But if it is, mark all components as
				// valid so that Container.validate() doesn't call their
				// layout managers' layoutContainer() methods after we return.
				// (If a layout manager's methods were allowed to run on the
				// main thread, an applet-writer (who can change a component's
				// layout manager) would be able to run his/her own code on
				// the main thread.)
				AppContext targetAppContext =
					SunToolkit.targetToAppContext(applet);
				if (!AppContext.getAppContext().equals(targetAppContext)) {
					((Component)applet).valid = true;
					validateAllComponents(applet.getComponents());
					((Component)target).valid = true;
					((Component)host).valid = true;
					return;
				}
			}
			((Component)host).valid = true;
			target.valid = true;
		}
		if (showDebugInfo) {
			System.err.println("SimpleLayout.layoutContainer(): " +
				"Container " + ((target == null) ? "null" : target.toString()));
		}
	}

	public void invalidateLayout(Container target) {
	}

	public String toString() {
		return getClass().getName() + "[width=" + width + ",height=" + height + "]";
	}

	static {
		if ("true".equalsIgnoreCase(System.getProperty("jep.debuginfo", "false"))) {
			showDebugInfo = true;
		}
	}

}

