/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the MRJ Carbon OJI Plugin.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corp.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):  Patrick C. Beard <beard@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */

/*
    MRJPlugin.cpp
    
    XP COM Plugin Implementation.
    
    by Patrick C. Beard.
 */

#include "MRJPlugin.h"
#include "MRJSession.h"
#include "MRJContext.h"
#include "MRJConsole.h"
#include "EmbeddedFramePluginInstance.h"

#import <mach-o/dyld.h>
#import "JavaEmbeddingPlugin.h"
extern "C" Boolean JEPCheckBundle();
extern Boolean CheckJavaEmbeddingFramework();
extern Boolean debugPrinting;
extern nsPluginThreadPtr theJavaScriptThread;
extern OSStatus spec2path(const FSSpec& spec, char* path, UInt32 maxPathSize);

#if !TARGET_CARBON
#include "MRJFrame.h"
#endif

#include "nsIServiceManager.h"
#include "nsIServiceManagerObsolete.h"
#include "nsObsoleteModuleLoading.h"

#include "nsIMemory.h"
#include "nsIJVMManager.h"
#include "nsIJVMPluginTagInfo.h"
#include "nsIPluginManager2.h"
#include "nsIPluginInstancePeer.h"
#include "nsIPluginInstancePeer2.h"
#include "nsIWindowlessPlugInstPeer.h"
#include "LiveConnectNativeMethods.h"
#include "CSecureEnv.h"

#include <Resources.h>
#include <LaunchServices.h>
#include <string>
#include <CFBundle.h>

nsIPluginManager* thePluginManager = NULL;
nsIPluginManager2* thePluginManager2 = NULL;
nsIMemory* theMemoryAllocator = NULL;

FSSpec thePluginSpec;
short thePluginRefnum = -1;
MRJPlugin *theJVMPlugin = NULL;

// We assume (in LiveConnectNativeMethods.cpp) that there's exactly one
// JavaScript thread, that that it's the main thread.  When/if this becomes
// invalid, we'll need to change things.
static nsPluginThreadPtr getJavaScriptThread(MRJPluginInstance *pluginInstance)
{
    nsPluginThreadPtr threadID = 0;
    if (MRJPluginInstance::isValidInstance(pluginInstance)) {
        nsIPluginInstancePeer *peer;
        if (pluginInstance->GetPeer(&peer) == NS_OK) {
            nsIPluginInstancePeer2 *peer2 = NULL;
            if (peer->QueryInterface(NS_GET_IID(nsIPluginInstancePeer2), (void**) &peer2) == NS_OK) {
                if (peer2->GetJSThread((PRUint32 *) &threadID) != NS_OK)
                    threadID = 0;
                NS_RELEASE(peer2);
            }
            NS_RELEASE(peer);
        }
    }
    return threadID;
}

// Deal with breakage caused by the fix for bmo bug 326273.  See nsjvmFixes.h.
nsID nsIJVMThreadManagerIID = NS_IJVMTHREADMANAGER_IID;

// Common interface IDs.

static NS_DEFINE_IID(kPluginCID, NS_PLUGIN_CID);
static NS_DEFINE_IID(kPluginManagerCID, NS_PLUGINMANAGER_CID);
static NS_DEFINE_IID(kJVMManagerCID, NS_JVMMANAGER_CID);

static NS_DEFINE_IID(kIWindowlessPluginInstancePeerIID, NS_IWINDOWLESSPLUGININSTANCEPEER_IID);

/**
 * Bottleneck all uses of the service manager, and use the obsolete service manager
 * if the modern one is unavailable.
 */
static nsIServiceManager* theServiceManager = NULL;
static nsIServiceManagerObsolete* theServiceManagerObsolete = NULL;

nsresult MRJPlugin::GetService(const nsCID& aCID, const nsIID& aIID, void* *aService)
{
    if (theServiceManager)
        return theServiceManager->GetService(aCID, aIID, aService);
    if (theServiceManagerObsolete)
        return theServiceManagerObsolete->GetService(aCID, aIID, (nsISupports **)aService);
    return NS_ERROR_FAILURE;
}

nsresult MRJPlugin::GetService(const char* aContractID, const nsIID& aIID, void* *aService)
{
    if (theServiceManager)
        return theServiceManager->GetServiceByContractID(aContractID, aIID, aService);
    if (theServiceManagerObsolete)
        return theServiceManagerObsolete->GetService(aContractID, aIID, (nsISupports **)aService);
    return NS_ERROR_FAILURE;
}

struct QITableEntry;

typedef nsresult (*TABLEDRIVENQI) (void *, const QITableEntry*, REFNSIID, void **);

// As of the fix for bug 313309, the Mozilla.org trunk now has "table-driven
// QI", which means that all plugins/extensions that use the NS_IMPL_ISUPPORTS
// macros now need to (statically) link in or implement a NS_TableDrivenQI()
// method.  In principle we could link to the "standalone xpcomglue library".
// But this would be a royal pain (since we'd need a universal-binary version
// of libxpcomglue.a).  So (to avoid having to link it in) I'm going to
// implement NS_TableDrivenQI() right here.  I _think_ the OS's "two-level
// namespace" magic will ensure that the right copy of this symbol is used
// (i.e. that our implementation will never be called).  But in case our
// NS_TableDrivenQI() does get called, I've made it locate the "right"
// NS_TableDrivenQI method and call through to that.
nsresult NS_TableDrivenQI(void* aThis, const QITableEntry* entries, REFNSIID aIID, void **aInstancePtr)
{
	nsresult retval = NS_ERROR_FAILURE;
	static TABLEDRIVENQI NS_TableDrivenQI_Ptr = NULL;
	if (!NS_TableDrivenQI_Ptr) {
		unsigned long i, count;
		NSSymbol symbol = nil;
		const struct mach_header *image = nil;
		char our_name[4096];
		const char *image_name = nil;
		spec2path(thePluginSpec, our_name, sizeof(our_name));
		count = _dyld_image_count();
		// We assume that the "right" NS_TableDrivenQI symbol will be the
		// first one we find that isn't ours (i.e. that other plugins and
		// extensions that implement or statically link in this symbol will
		// have been loaded later than the module that contains the "right"
		// one).
		for (i = 1; i <= count; ++i) {
			image = (const struct mach_header *) _dyld_get_image_header(i - 1);
			image_name = _dyld_get_image_name(i - 1);
			if (strcmp(our_name, image_name) == 0)
				continue;
			symbol = NSLookupSymbolInImage(image,
				"__Z16NS_TableDrivenQIPvPK12QITableEntryRK4nsIDPS_",
				NSLOOKUPSYMBOLINIMAGE_OPTION_RETURN_ON_ERROR);
			if (symbol)
				NS_TableDrivenQI_Ptr = (TABLEDRIVENQI) NSAddressOfSymbol(symbol);
			if (NS_TableDrivenQI_Ptr)
				break;
		}
	}
	if (NS_TableDrivenQI_Ptr)
		retval = NS_TableDrivenQI_Ptr(aThis, entries, aIID, aInstancePtr);
    return retval;
}

#pragma export on

long getSystemVersion()
{
    static long version = 0;
    if (!version) {
        Gestalt(gestaltSystemVersion, &version);
        version &= 0xFFFF;
    }
    return version;
}

extern "C"
nsresult NSGetFactory(nsISupports* serviceManager, const nsCID &aClass, const char *aClassName, const char *aContractID, nsIFactory **aFactory)
{
    nsresult result = NS_OK;

    // First off, fail if not running on AT LEAST Mac OS X, v10.1.
    if (getSystemVersion() < 0x00001010)
        return NS_ERROR_FAILURE;

	// Also fail if neither the Java Embedding Plugin nor the JavaEmbedding
	// framework are present -- as happens if someone tries to use
	// MRJPlugin.plugin without JavaEmbeddingPlugin.bundle on Mac OS X for
	// Intel.
	if (!JEPCheckBundle() && !CheckJavaEmbeddingFramework())
		return NS_ERROR_FAILURE;

    if (theServiceManager == NULL && theServiceManagerObsolete == NULL) {
        if (NS_FAILED(serviceManager->QueryInterface(NS_GET_IID(nsIServiceManager), (void**)&theServiceManager)))
            if (NS_FAILED(serviceManager->QueryInterface(NS_GET_IID(nsIServiceManagerObsolete), (void**)&theServiceManagerObsolete)))
                return NS_ERROR_FAILURE;

        // Our global operator new wants to use nsIMalloc to do all of its allocation.
        // This should be available from the Service Manager.
        if (NS_FAILED(MRJPlugin::GetService("@mozilla.org/xpcom/memory-service;1", NS_GET_IID(nsIMemory), (void **)&theMemoryAllocator)))
            return NS_ERROR_FAILURE;
    }

    if (aClass.Equals(kPluginCID)) {
        MRJPlugin* pluginFactory = new MRJPlugin();
        pluginFactory->AddRef();
        *aFactory = pluginFactory;
        return NS_OK;
    }
    return NS_NOINTERFACE;
}

#include "nsICookieStorage.h"

extern "C"
unsigned MRJGetCookie(const char *inCookieURL, void *inOutCookieBuffer, unsigned *inOutCookieSize)
{
	nsICookieStorage *cookieMonster = NULL;
	nsresult rv = MRJPlugin::GetService(kPluginManagerCID, NS_GET_IID(nsICookieStorage), (void **) &cookieMonster);
	if (NS_SUCCEEDED(rv) && (cookieMonster != NULL)) {
		rv = cookieMonster->GetCookie(inCookieURL, inOutCookieBuffer, *inOutCookieSize);
		cookieMonster->Release();
	}
	if (NS_SUCCEEDED(rv))
		rv = 0;
	return rv;
}

extern "C"
unsigned MRJSetCookie(const char *inCookieURL, const void *inCookieBuffer, unsigned inCookieSize)
{
	nsICookieStorage *cookieMonster = NULL;
	nsresult rv = MRJPlugin::GetService(kPluginManagerCID, NS_GET_IID(nsICookieStorage), (void **) &cookieMonster);
	if (NS_SUCCEEDED(rv) && (cookieMonster != NULL)) {
		rv = cookieMonster->SetCookie(inCookieURL, inCookieBuffer, inCookieSize);
		cookieMonster->Release();
	}
	if (NS_SUCCEEDED(rv))
		rv = 0;
	return rv;
}

#pragma export off

#if TARGET_RT_MAC_CFM

extern "C" {

pascal OSErr __initialize(const CFragInitBlock *initBlock);
pascal void __terminate(void);

#if defined(MRJPLUGIN_GC)
pascal OSErr __NSInitialize(const CFragInitBlock* initBlock);
pascal void __NSTerminate(void);
#define __initialize __NSInitialize
#define __terminate __NSTerminate
#endif

pascal OSErr MRJPlugin__initialize(const CFragInitBlock *initBlock);
pascal void MRJPlugin__terminate(void);

}

pascal OSErr MRJPlugin__initialize(const CFragInitBlock *initBlock)
{
    OSErr err = __initialize(initBlock);
    if (err != noErr) return err;

    if (initBlock->fragLocator.where == kDataForkCFragLocator) {
        thePluginSpec = *initBlock->fragLocator.u.onDisk.fileSpec;
    
        // Open plugin's resource fork for read-only access.
        thePluginRefnum = ::FSpOpenResFile(&thePluginSpec, fsRdPerm);
    }
    
    return noErr;
}

pascal void MRJPlugin__terminate()
{
#if !TARGET_CARBON
    // last ditch release of the memory allocator.
    if (theMemoryAllocator != NULL) {
        theMemoryAllocator->Release();
        theMemoryAllocator = NULL;
    }
#endif

    // Close plugin's resource fork.
    // If we don't, Mac OS X 10.1 crashes.
    if (thePluginRefnum != -1) {
        ::CloseResFile(thePluginRefnum);
        thePluginRefnum = -1;
    }

    __terminate();
}

#endif /* TARGET_RT_MAC_CFM */

#if TARGET_RT_MAC_MACHO

extern "C" {
CF_EXPORT Boolean CFURLGetFSRef(CFURLRef url, FSRef *fsRef);
}

/* getPluginBundle()'s now being called from getJEPBundle() in
 * JavaEmbeddingPlugin.cpp.
 */
CFBundleRef getPluginBundle()
{
    printf("### MRJPlugin:  getPluginBundle() here. ###\n");
    CFBundleRef bundle = NULL;
    CFArrayRef allBundles = CFBundleGetAllBundles();
    int count = CFArrayGetCount(allBundles);
    for (int i = 1; i <= count; ++i) {
        CFBundleRef aBundle = (CFBundleRef) CFArrayGetValueAtIndex(allBundles, i-1);
        CFStringRef anIdentifier = CFBundleGetIdentifier(aBundle);
        if (!anIdentifier)
            continue;
        if (CFStringCompare(anIdentifier, CFSTR("com.netscape.MRJPlugin"), 0) == kCFCompareEqualTo) {
            bundle = aBundle;
            break;
        }
    }
    if (bundle) {
        printf("### MRJPlugin:  CFBundleGetBundleWithIdentifier() succeeded. ###\n");
        // initialize thePluginSpec for later use. open our resource fork as well?
        CFURLRef url = CFBundleCopyExecutableURL(bundle);
        if (url) {
            FSRef ref;
            if (CFURLGetFSRef(url, &ref)) {
                printf("### MRJPlugin:  CFURLGetFSRef() succeeded. ###\n");
                FSGetCatalogInfo(&ref, kFSCatInfoNone, NULL, NULL, &thePluginSpec, NULL);

                // Open plugin's resource fork for read-only access. is this really necessary?
                thePluginRefnum = ::CFBundleOpenBundleResourceMap(bundle);
            }
            CFRelease(url);
        }
    }
    return bundle;
}

CFBundleRef thePluginBundle;

#endif /* TARGET_RT_MAC_MACHO */

//
// The MEAT of the plugin.
//

#pragma mark *** MRJPlugin ***

MRJPlugin::MRJPlugin()
    :   mManager(NULL), mThreadManager(NULL), mSession(NULL), mConsole(NULL), mPluginThreadID(NULL), mIsEnabled(false)
{
}

MRJPlugin::~MRJPlugin()
{
    // Release the console.
    NS_IF_RELEASE(mConsole);

    // tear down the MRJ session, if it exists.
    if (mSession != NULL) {
        delete mSession;
        mSession = NULL;
    }

    // Release the managers.
    NS_IF_RELEASE(mManager);
    NS_IF_RELEASE(mThreadManager);
}

NS_IMPL_ADDREF(MRJPlugin)
NS_IMPL_RELEASE(MRJPlugin)

NS_METHOD MRJPlugin::QueryInterface(const nsIID& aIID, void** aInstancePtr)
{
    if (aInstancePtr == NULL) {
        return NS_ERROR_NULL_POINTER;
    }
    *aInstancePtr = NULL;
    if (aIID.Equals(NS_GET_IID(nsIPlugin))) {
        *aInstancePtr = (void *) (nsIPlugin *) this;
        AddRef();
        return NS_OK;
    }
    if (aIID.Equals(NS_GET_IID(nsIJVMPlugin))) {
        *aInstancePtr = (void *) (nsIJVMPlugin *) this;
        AddRef();
        return NS_OK;
    }
    if (aIID.Equals(NS_GET_IID(nsIRunnable))) {
        *aInstancePtr = (void *) (nsIRunnable *) this;
        AddRef();
        return NS_OK;
    }
/*
 * MRJPlugin aggregates MRJConsole, so that it can be QI'd to be an nsIJVMConsole.
 */
#if USE_SYSTEM_CONSOLE
    if (aIID.Equals(NS_GET_IID(nsIJVMConsole))) {
        *aInstancePtr = (void *) (nsIJVMConsole *) this;
        AddRef();
        return NS_OK;
    }
#endif
    if (aIID.Equals(NS_GET_IID(nsISupports))) {
        *aInstancePtr = (void *) this;
        AddRef();
        return NS_OK;
    }
#if USE_SYSTEM_CONSOLE
    return mConsole->QueryInterface(aIID, aInstancePtr);
#endif
    return NS_NOINTERFACE;
}

NS_METHOD MRJPlugin::CreateInstance(nsISupports *aOuter, const nsIID& aIID, void **aResult)
{
    nsresult result = StartupJVM();
    if (result == NS_OK) {
        MRJPluginInstance* instance = new MRJPluginInstance(this);
        if (instance == nsnull)
            return NS_ERROR_OUT_OF_MEMORY;
        result = instance->QueryInterface(aIID, aResult);
        if (result != NS_OK)
            delete instance;
    }
    return result;
}

#define NS_APPLET_MIME_TYPE "application/x-java-applet"

NS_METHOD MRJPlugin::CreatePluginInstance(nsISupports *aOuter, REFNSIID aIID, const char* aPluginMIMEType, void **aResult)
{
    nsresult result = NS_NOINTERFACE;

    if (::strcmp(aPluginMIMEType, "application/x-java-frame") == 0) {
#if !TARGET_CARBON
        // create a special plugin instance that manages an embedded frame.
        EmbeddedFramePluginInstance* instance = new EmbeddedFramePluginInstance();
        nsresult result = instance->QueryInterface(aIID, aResult);
        if (result != NS_OK)
            delete instance;
#endif
    } else {
        // assume it's some kind of an applet.
        result = CreateInstance(aOuter, aIID, aResult);
    }
    return result;
}

NS_METHOD MRJPlugin::Initialize()
{
    nsresult result = NS_OK;

    // try to get a plugin manager.
    if (thePluginManager == NULL) {
        result = MRJPlugin::GetService(kPluginManagerCID, NS_GET_IID(nsIPluginManager), (void**)&thePluginManager);
        if (result != NS_OK || thePluginManager == NULL)
            return NS_ERROR_FAILURE;
    }

    // see if the enhanced plugin manager exists.
    if (thePluginManager2 == NULL) {
        if (thePluginManager->QueryInterface(NS_GET_IID(nsIPluginManager2), (void**)&thePluginManager2) != NS_OK)
            thePluginManager2 = NULL;
    }

    // try to get a JVM manager. we have to be able to run without one.
    if (MRJPlugin::GetService(kJVMManagerCID, NS_GET_IID(nsIJVMManager), (void**)&mManager) != NS_OK)
        mManager = NULL;
    
    // try to get a Thread manager.
    if (mManager != NULL) {
        if (mManager->QueryInterface(nsIJVMThreadManagerIID, (void**)&mThreadManager) != NS_OK)
            mThreadManager = NULL;

        if (mThreadManager != NULL)
            mThreadManager->GetCurrentThread(&mPluginThreadID);
    }

#if !USE_SYSTEM_CONSOLE
    // create a console, only if there's user interface for it.
    if (thePluginManager2 != NULL) {
        mConsole = new MRJConsole(this);
        mConsole->AddRef();
    }
#endif

    return result;
}

NS_METHOD MRJPlugin::Shutdown()
{
    // shutdown LiveConnect.
    ShutdownLiveConnectSupport();

    // release our reference to the plugin manager(s).
    NS_IF_RELEASE(thePluginManager2);
    NS_IF_RELEASE(thePluginManager);

    // release our reference(s) to the service manager.
    NS_IF_RELEASE(theServiceManager);
    NS_IF_RELEASE(theServiceManagerObsolete);
    
    return NS_OK;
}

NS_METHOD MRJPlugin::GetMIMEDescription(const char* *result)
{
    *result = NS_JVM_MIME_TYPE;
    return NS_OK;
}

NS_METHOD MRJPlugin::GetValue(nsPluginVariable variable, void *value)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

MRJSession* MRJPlugin::getSession()
{
    StartupJVM();
    return mSession;
}

nsIJVMManager* MRJPlugin::getManager()
{
    return mManager;
}

nsIJVMThreadManager* MRJPlugin::getThreadManager()
{
    return mThreadManager;
}

char* getJavaConsolePath(char* path, UInt32 maxPathSize)
{
    path[0] = '\0';
    FSRef ref;
    OSErr err = FSFindFolder(kUserDomain, kDomainLibraryFolderType, true, &ref);
    if (err == noErr) {
        err = FSRefMakePath(&ref, (UInt8*) path, maxPathSize);
        const char kJavaConsoleLog[] = { "/Logs/Java Console.log" };
        int len = strlen(path);
        if (err == noErr && (len + sizeof(kJavaConsoleLog)) <= maxPathSize) {
            strcat(path + len, kJavaConsoleLog);
        }
    }
    return path;
}

nsresult MRJPlugin::StartupJVM()
{
    if (mSession == NULL) {
        // start a session with MRJ.
        mSession = new MRJSession(this);

        // Add "MRJPlugin.jar" to the class path.
        FSSpec jarFileSpec = { thePluginSpec.vRefNum, thePluginSpec.parID, "\pMRJPlugin.jar" };
        mSession->addToClassPath(jarFileSpec);

        // Compute location of ~/Library/Logs/JavaConsole.log
        char consolePath[512];
        mSession->open(getJavaConsolePath(consolePath, sizeof(consolePath)));

        if (mSession->getStatus() != noErr) {
            // how can we signal an error?
            delete mSession;
            mSession = NULL;
            ShutdownLiveConnectSupport();
            return NS_ERROR_FAILURE;
        }

#if 0
        // start our idle thread.
        if (mThreadManager != NULL) {
            PRUint32 threadID;
            mThreadManager->CreateThread(&threadID, this);
        }
#endif

        mIsEnabled = true;
    }
    return NS_OK;
}

NS_METHOD MRJPlugin::AddToClassPath(const char* dirPath)
{
    if (mSession != NULL) {
        mSession->addToClassPath(dirPath);
        return NS_OK;
    }
    return NS_ERROR_FAILURE;
}

NS_METHOD MRJPlugin::GetClassPath(const char* *result)
{
    char* classPath = mSession->getProperty("java.class.path");
    *result = classPath;
    return (classPath != NULL ? NS_OK : NS_ERROR_FAILURE);
}

NS_METHOD MRJPlugin::GetJavaWrapper(JNIEnv* env, jint jsobj, jobject *jobj)
{
    // use jsobj as key into a table.
    // if not in the table, then create a new netscape.javascript.JSObject that references this.
    *jobj = Wrap_JSObject(env, jsobj);
    return NS_OK;
}

NS_METHOD MRJPlugin::UnwrapJavaWrapper(JNIEnv* env, jobject jobj, jint* jsobj)
{
    *jsobj = Unwrap_JSObject(env, jobj);
    return NS_OK;
}

NS_METHOD MRJPlugin::CreateSecureEnv(JNIEnv* proxyEnv, nsISecureEnv* *outSecureEnv)
{
    *outSecureEnv = NULL;
    nsresult rv = StartupJVM();
    if (rv == NS_OK) {
        // Need to spawn a new JVM communication thread here.
        // rv = CSecureEnv::Create(this, proxyEnv, NS_GET_IID(nsISecureEnv), (void**)outSecureEnv);
        static NS_DEFINE_IID(kISecureEnvIID, NS_ISECUREENV_IID);
        rv = CSecureEnv::Create(this, proxyEnv, kISecureEnvIID, (void**)outSecureEnv);
    }
    return rv;
}

NS_METHOD MRJPlugin::SpendTime(PRUint32 timeMillis)
{
    nsresult result = NS_OK;
    // Only do this if there aren't any plugin instances.
    if (MRJPluginInstance::getInstances() == NULL) {
        if (mSession == NULL)
            result = StartupJVM();
        if (mSession != NULL)
            mSession->idle(timeMillis);
    }
    return result;
}

/**
 * Poor Man's Mac OS X Java Console implementation. MRJSession::open()
 * opens ~/Library/Logs/JavaConsole.log, and redirects all output
 * from System.out and System.err to this file. Rather than
 * implementing a Java based console, just take advantage of
 * Mac OS X's /Applications/Utilities/Console.app capability
 * of showing log files. The following simply uses Launch Services
 * to open whatever application normally opens ".log" files.
 * A more sophisticated implementation would launch Console.app
 * with the log file, but this works just fine for now.
 */
NS_METHOD MRJPlugin::Show()
{
    // Launch the system console application.
    char consolePath[512];
    getJavaConsolePath(consolePath, sizeof(consolePath));
    CFURLRef consoleURL = CFURLCreateFromFileSystemRepresentation(kCFAllocatorDefault, (UInt8*)consolePath, strlen(consolePath), false);
    if (consoleURL) {
        LSOpenCFURLRef(consoleURL, NULL);
        CFRelease(consoleURL);
    }
    return NS_OK;
}

NS_METHOD MRJPlugin::Hide()
{
    // Hide system console?
    return NS_OK;
}

NS_METHOD MRJPlugin::IsVisible(PRBool *result)
{
    // *result = IsConsoleRunning();
    *result = PR_FALSE;
    return NS_OK;
}
    
NS_METHOD MRJPlugin::Print(const char* msg, const char* /* encodingName */)
{
    // printf onto the console.
    return NS_OK;
}

NS_METHOD MRJPlugin::Run()
{
    while (mSession != NULL) {
        mSession->idle();
        mThreadManager->Sleep();
    }
    return NS_OK;
}

MRJPluginInstance* MRJPlugin::getPluginInstance(jobject applet)
{
    JNIEnv* env = mSession->getCurrentEnv();
    MRJPluginInstance* instance = MRJPluginInstance::getInstances();
    while (instance && !instance->isDestroying()) {
        jobject object = NULL;
        instance->GetJavaObject(&object);
        if (object && env->IsSameObject(applet, object)) {
            instance->AddRef();
            return instance;
        }
        instance = instance->getNextInstance();
    }
    return NULL;
}

MRJPluginInstance* MRJPlugin::getPluginInstance(JNIEnv* jenv)
{
#if !TARGET_CARBON
    // Apple will provide an API that maps a JNIEnv to an JMAWTContextRef. We can map this to the MRJContext/Applet/Instance.
    MRJPluginInstance* instance = MRJPluginInstance::getInstances();
    if (&::JMJNIToAWTContext != NULL) {
        JMAWTContextRef contextRef = ::JMJNIToAWTContext(mSession->getSessionRef(), jenv);
        if (contextRef != NULL) {
            while (instance && !instance->isDestroying()) {
                MRJContext context = instance->getContext();
                if (!context)
                    continue;
                if (context->getContextRef() == contextRef) {
                    instance->AddRef();
                    return instance;
                }
                instance = instance->getNextInstance();
            }
        }
    } else {
        if (MRJPluginInstance::isValidInstance(instance)) {
            instance->AddRef();
            return instance;
        }
    }
#endif
    return NULL;
}

Boolean MRJPlugin::inPluginThread()
{
    Boolean result = false;
    nsPluginThread *currentThreadID = NULL;
    
    if (mThreadManager != NULL)
        mThreadManager->GetCurrentThread(&currentThreadID);
    if ((NULL != currentThreadID) && (NULL != mPluginThreadID)) {
        if (currentThreadID == mPluginThreadID) {
            result = true;
        }
    }
    
    return result;
}

#pragma mark *** MRJPluginInstance ***

MRJPluginInstance::MRJPluginInstance(MRJPlugin* plugin)
    :   mPeer(NULL), mWindowlessPeer(NULL),
        mPlugin(plugin), mSession(plugin->getSession()),
        mContext(NULL), mApplet(NULL), mPluginWindow(NULL), mDestroying(false),
        mNext(NULL)
{
    // add this instance to the instance list.
    pushInstance();

    // Tell the plugin we are retaining a reference.
    mPlugin->AddRef();
}

/* In recent Mozilla.org browsers (at least) this never gets called.  So it
 * appears all MRJPluginInstances are leaked, and we need to do all our
 * tearing down in Destroy().  (Tested with Camino 1.0 and Firefox 1.5.0.2.)
 */
MRJPluginInstance::~MRJPluginInstance()
{
    // Remove this instance from the global list.
    //popInstance();

#if 0
    if (mContext != NULL) {
        delete mContext;
        mContext = NULL;
    }

    if (mPlugin != NULL) {
        mPlugin->Release();
        mPlugin = NULL;
    }

    if (mWindowlessPeer != NULL) {
        mWindowlessPeer->Release();
        mWindowlessPeer = NULL;
    }

    if (mPeer != NULL) {
        mPeer->Release();
        mPeer = NULL;
    }

    if (mApplet != NULL) {
        JNIEnv* env = mSession->getCurrentEnv();
        env->DeleteGlobalRef(mApplet);
        mApplet = NULL;
    }
#endif
}

NS_IMPL_ADDREF(MRJPluginInstance)
NS_IMPL_RELEASE(MRJPluginInstance)

NS_METHOD MRJPluginInstance::QueryInterface(const nsIID& aIID, void** aInstancePtr)
{
    if (aInstancePtr == NULL) {
        return NS_ERROR_NULL_POINTER;
    }
    *aInstancePtr = NULL;
    if (aIID.Equals(NS_GET_IID(nsIPluginInstance))) {
        *aInstancePtr = (void *) (nsIPluginInstance *) this;
        AddRef();
        return NS_OK;
    }
    if (aIID.Equals(NS_GET_IID(nsIJVMPluginInstance))) {
        *aInstancePtr = (void *) (nsIJVMPluginInstance *) this;
        AddRef();
        return NS_OK;
    }
    if (aIID.Equals(NS_GET_IID(nsIEventHandler))) {
        *aInstancePtr = (void *) (nsIEventHandler *) this;
        AddRef();
        return NS_OK;
    }
    if (aIID.Equals(NS_GET_IID(nsISupports))) {
        *aInstancePtr = (void *) this;
        AddRef();
        return NS_OK;
    }
    return NS_NOINTERFACE;
}

static const char* kGetCodeBaseScriptURL = "javascript:var href = window.location.href; href.substring(0, href.lastIndexOf('/') + 1)";
static const char* kGetDocumentBaseScriptURL = "javascript:window.location";

static bool hasTagInfo(nsISupports* supports)
{
    nsIJVMPluginTagInfo* tagInfo;
    if (supports->QueryInterface(NS_GET_IID(nsIJVMPluginTagInfo), (void**)&tagInfo) == NS_OK) {
        NS_RELEASE(tagInfo);
        return true;
    }
    return false;
}

NS_METHOD MRJPluginInstance::Initialize(nsIPluginInstancePeer* peer)
{
    // Tell the peer we are retaining a reference.
    mPeer = peer;
    mPeer->AddRef();

    // See if we have a windowless peer.
    nsresult result = mPeer->QueryInterface(kIWindowlessPluginInstancePeerIID, (void**)&mWindowlessPeer);
    if (result != NS_OK) mWindowlessPeer = NULL;

    // create a context for the applet we will run.
    mContext = new MRJContext(mSession, this);

    if (hasTagInfo(mPeer)) {
        mContext->processAppletTag();
        mContext->createContext();
    } else {
        // we'll be using JavaScript to create windows.
        // fire up a JavaScript URL to get the current document's location.
        nsIPluginInstance* pluginInstance = this;
        nsIPluginStreamListener* listener = this;
        result = thePluginManager->GetURL(pluginInstance, kGetDocumentBaseScriptURL, NULL, listener);
    }

    if (!theJavaScriptThread)
        theJavaScriptThread = getJavaScriptThread(this);

    return NS_OK;
}

NS_METHOD MRJPluginInstance::OnDataAvailable(nsIPluginStreamInfo* pluginInfo, nsIInputStream* input, PRUint32 length)
{
    // hopefully all our data is available.
    char* documentBase = new char[length + 1];
    if (documentBase && !mDestroying) {
        if (input->Read(documentBase, length, &length) == NS_OK) {
            // We've delayed processing the applet tag, because we
            // don't know the location of the current document yet.
            documentBase[length] = '\0';
            
            // set up the default document location, which can be used to compute relative CODEBASE, etc.
            mContext->setDocumentBase(documentBase);
            delete[] documentBase;
            
            mContext->processAppletTag();
            mContext->createContext();
            
            // SetWindow may be called at an inopportune time.
            if (mPluginWindow != NULL)
                mContext->setWindow(mPluginWindow);
        }
    }
    return NS_OK;
}

NS_METHOD MRJPluginInstance::GetPeer(nsIPluginInstancePeer* *result)
{
    NS_IF_ADDREF(mPeer);
    *result = mPeer;
    return NS_OK;
}

NS_METHOD MRJPluginInstance::Start()
{
    if (mDestroying)
        return NS_OK;

    // Take this moment to show the applet's frames (if any).
    mContext->showFrames();
    
    mContext->resumeApplet();
    
    return NS_OK;
}

NS_METHOD MRJPluginInstance::Stop()
{
    if (mDestroying)
        return NS_OK;

    // Take this moment to hide the applet's frames.
    mContext->hideFrames();

    mContext->suspendApplet();

    return NS_OK;
}

NS_METHOD MRJPluginInstance::Destroy()
{
    mDestroying = true;

	popInstance();

    // Use this opportunity to break any cycles that might exist, and reduce
    // reference counts to their minimum values.
    if (mContext != NULL)
        mContext->startDestroying();

    if (mPlugin != NULL) {
        mPlugin->Release();
        mPlugin = NULL;
    }

    if (mWindowlessPeer != NULL) {
        mWindowlessPeer->Release();
        mWindowlessPeer = NULL;
    }

    if (mPeer != NULL) {
        mPeer->Release();
        mPeer = NULL;
    }

    if (mApplet != NULL) {
        JNIEnv* env = mSession->getCurrentEnv();
        env->DeleteGlobalRef(mApplet);
        mApplet = NULL;
    }

    if (mContext != NULL) {
        delete mContext;
        mContext = NULL;
    }

    return NS_OK;
}

/** FIXME:  Need an intelligent way to track changes to the NPPluginWindow. */

NS_METHOD MRJPluginInstance::SetWindow(nsPluginWindow* pluginWindow)
{
    if (mDestroying)
        return NS_OK;

    mPluginWindow = pluginWindow;

    mContext->setWindow(pluginWindow);

    return NS_OK;
}

NS_METHOD MRJPluginInstance::HandleEvent(nsPluginEvent* pluginEvent, PRBool* eventHandled)
{
    *eventHandled = PR_TRUE;
    Boolean isUpdate;

    if (pluginEvent && !mDestroying) {
        EventRecord* event = pluginEvent->event;

        // Check for coordinate/clipping changes.
        // Too bad nsPluginEventType_ClippingChangedEvent events aren't implemented.
        isUpdate = (event->what == updateEvt);
        inspectInstance(isUpdate);
    
        if (event->what == nullEvent) {
            // Give Java another quantum of time. We don't need this on Mac OS X.
            mSession->idle(0x00000400); // now SpendTime does this.
        } else {
#if TARGET_CARBON
            *eventHandled = mContext->handleEvent(event);
#else
            MRJFrame* frame = mContext->findFrame(pluginEvent->window);
            if (frame != NULL) {
                switch (event->what) {
                case nsPluginEventType_GetFocusEvent:
                    frame->focusEvent(true);
                    break;
                
                case nsPluginEventType_LoseFocusEvent:
                    frame->focusEvent(false);
                    break;

                case nsPluginEventType_AdjustCursorEvent:
                    frame->idle(event->modifiers);
                    break;
                
                case nsPluginEventType_MenuCommandEvent:
                    frame->menuSelected(event->message, event->modifiers);
                    break;
                    
                default:
                    *eventHandled = frame->handleEvent(event);
                    break;
                }
            } else {
                if (event->what == updateEvt) {
                    mContext->drawApplet();
                }
            }
#endif
        }
    }
    
    return NS_OK;
}

NS_METHOD MRJPluginInstance::Print(nsPluginPrint* platformPrint)
{
    if (debugPrinting)
        JEPLogJavaMessage(CFSTR("MRJPluginInstance::Print(): Calling"));
    if (mDestroying)
        return NS_ERROR_FAILURE;
    if (platformPrint->mode == nsPluginMode_Embedded) {
        mContext->printApplet(&platformPrint->print.embedPrint.window);
        return NS_OK;
    }
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_METHOD MRJPluginInstance::GetValue(nsPluginInstanceVariable variable, void *value)
{
    switch (variable) {
    case nsPluginInstanceVariable_WindowlessBool:
        *(PRBool*)value = PR_FALSE;
        break;
    case nsPluginInstanceVariable_TransparentBool:
        *(PRBool*)value = PR_FALSE;
        break;
    case nsPluginInstanceVariable_DoCacheBool:
        *(PRBool*)value = PR_FALSE;
        break;
    default:
        break;
    }
    return NS_OK;
}

// Doesn't "addref" the object it returns.
NS_METHOD MRJPluginInstance::GetJavaObject(jobject *result)
{
    if (mDestroying) {
        *result = NULL;
        return NS_ERROR_FAILURE;
    }
    if (mApplet == NULL) {
        jobject applet = mContext->getApplet();
        JNIEnv* env = mSession->getCurrentEnv();
        mApplet = env->NewGlobalRef(applet);
    }
    //printf("MRJPluginInstance::GetJavaObject(): returning mApplet == 0x%.8X, thread 0x%.8X\n", mApplet, pthread_self());
    *result = mApplet;
    return NS_OK;
}

// Accessing the list of instances.

static MRJPluginInstance* theInstances = NULL;

void MRJPluginInstance::pushInstance()
{
    mNext = theInstances;
    theInstances = this;
}

void MRJPluginInstance::popInstance()
{
    MRJPluginInstance** link = &theInstances;
    MRJPluginInstance* instance  = *link;
    while (instance != NULL) {
        if (instance == this) {
            *link = mNext;
            mNext = NULL;
            break;
        }
        link = &instance->mNext;
        instance = *link;
    }
}

MRJPluginInstance* MRJPluginInstance::getInstances()
{
    return theInstances;
}

MRJPluginInstance* MRJPluginInstance::getNextInstance()
{
    return mNext;
}

Boolean MRJPluginInstance::isValidInstance(MRJPluginInstance *instance)
{
    if (!instance)
        return false;
    MRJPluginInstance *pluginInstance = MRJPluginInstance::getInstances();
    while (pluginInstance && !pluginInstance->isDestroying()) {
        if (instance == pluginInstance)
            return true;
        pluginInstance = pluginInstance->getNextInstance();
    }
    return false;
}

MRJContext* MRJPluginInstance::getContext()
{
    if (mDestroying)
        return NULL;
    return mContext;
}

MRJSession* MRJPluginInstance::getSession()
{
    return mSession;
}

Boolean MRJPluginInstance::isDestroying()
{
	return mDestroying;
}

void MRJPluginInstance::inspectInstance(Boolean isUpdateEvt)
{
    if (!mDestroying && (mContext != NULL) && mContext->inspectWindow() && !isUpdateEvt && (mWindowlessPeer != NULL))
        mWindowlessPeer->ForceRedraw();
}
