/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the MRJ Carbon OJI Plugin.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corp.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):  Patrick C. Beard <beard@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */

/*
	MRJSession.cpp

	Encapsulates a session with the MacOS Runtime for Java.
	
	by Patrick C. Beard.
 */

#include "CSecureEnv.h"
#include "MRJSession.h"
#include "MRJPlugin.h"
#include "MRJContext.h"
#include "MRJConsole.h"
#include "MRJMonitor.h"
#include "TimedMessage.h"
#include "LiveConnectNativeMethods.h"
#import "JavaEmbeddingPlugin.h"
extern "C" Boolean JEPCheckBundle();

#include <ControlDefinitions.h>
#include <string.h>
#include <Memory.h>
#include <Files.h>
#include <Dialogs.h>
#include <Appearance.h>
#include <Resources.h>
#include <Gestalt.h>
#include <Folders.h>
#include <Script.h>

#include <JavaControl.h>
#include <CFString.h>

#include <string>

extern MRJConsole* theConsole;
extern short thePluginRefnum;
extern FSSpec thePluginSpec;

static MRJSession* theSession;
Boolean debugUpdates = false;
Boolean debugProxyJNI = false;
Boolean debugPrinting = false;

#if REDIRECT_VFPRINTF

/**
 * As of JDK 1.2, there is a standard way to redirect *LOW LEVEL* messages from the JVM by
 * providing a vfprintf hook. Since this plugin is built using CFM compilers, this means
 * the function pointer has to be wrapped in glue code which can be called from Mach-O.
 * To get low level messages to go to the same location as Java's System.out/err, we can
 * try various techniques. One way would be to have all output go through a common
 * native method, which ends up calling java_vfprintf. Another way, is to send output
 * to Java from java_vfprintf. It turns out that you can't always do this safely,
 * so you have to send the messages later at a known safe time. The current implementation
 * uses a Carbon Events timer to delay sending the message to Java. This has the unfortunate
 * side effect of reordering messages sent through java_vfprintf with respect to those
 * sent through System.out/err.
 */
 
//
//	This function allocates a block of CFM glue code which contains the instructions to call CFM routines
//
static void* NewMachOFunctionPointer(void* cfmfp)
{
#if TARGET_RT_MAC_CFM
    static UInt32 CFM_glue[6] = {0x3D800000, 0x618C0000, 0x800C0000, 0x804C0004, 0x7C0903A6, 0x4E800420};
    UInt32	*mfp = (UInt32*) NewPtr(sizeof(CFM_glue));		//	Must later dispose of allocated memory
    if (mfp) {
	    BlockMoveData(CFM_glue, mfp, sizeof(CFM_glue));
	    mfp[0] |= ((UInt32)cfmfp >> 16);
	    mfp[1] |= ((UInt32)cfmfp & 0xFFFF);
	    MakeDataExecutable(mfp, sizeof(CFM_glue));
	}
	return mfp;
#elif TARGET_RT_MAC_MACHO
    return cfmfp;
#endif
}

inline jobject ToGlobalRef(JNIEnv* env, jobject localRef)
{
    jobject globalRef = env->NewGlobalRef(localRef);
    env->DeleteLocalRef(localRef);
    return globalRef;
}

static jobject Get_System_out(JNIEnv* env)
{
    jclass java_lang_System = env->FindClass("java/lang/System");
    if (java_lang_System) {
        jfieldID outID = env->GetStaticFieldID(java_lang_System, "out", "Ljava/io/PrintStream;");
        jobject out = (outID ? env->GetStaticObjectField(java_lang_System, outID) : NULL);
        env->DeleteLocalRef(java_lang_System);
        return ToGlobalRef(env, out);
    }
    return NULL;
}

static jmethodID GetObjectMethodID(JNIEnv* env, jobject object, const char* name, const char* sig)
{
    jclass clazz = env->GetObjectClass(object);
    if (clazz) {
        jmethodID result = env->GetMethodID(clazz, name, sig);
        env->DeleteLocalRef(clazz);
        return result;
    }
    return NULL;
}

static void System_out_print(JNIEnv* env, const jchar* chars, jsize len)
{
    static jobject java_lang_System_out = Get_System_out(env);
    static jmethodID java_io_PrintStream_print = GetObjectMethodID(env, java_lang_System_out, "print", "([C)V");
    jcharArray array = env->NewCharArray(len);
    if (array) {
        env->SetCharArrayRegion(array, 0, len, (jchar*) chars);
        env->CallVoidMethod(java_lang_System_out, java_io_PrintStream_print, array);
        env->DeleteLocalRef(array);
    }
}

/**
 * Sends a message to the Java Console from a Carbon
 * Events timer. This is done asynchronously so that
 * messages from deep within the JavaVM can be processed
 * using the JavaVM.
 */
class ConsoleMessage : public TimedMessage {
    CFStringRef mMessage;

public:
    ConsoleMessage(CFStringRef message)
        :   mMessage(message)
    {
        ::CFRetain(mMessage);
    }
    
    ~ConsoleMessage()
    {
        ::CFRelease(mMessage);
    }
    
    virtual void execute();
};

void ConsoleMessage::execute()
{
    if (theSession) {
        jsize len = ::CFStringGetLength(mMessage);
        jchar* buffer = new jchar[len];
        CFRange range = { 0, len };
        CFStringGetCharacters(mMessage, range, buffer);
        System_out_print(theSession->getCurrentEnv(), buffer, len);
        delete[] buffer;
    }
}

static jint JNICALL java_vfprintf(FILE *fp, const char *format, va_list args)
{
    jint result = 0;
    CFStringRef formatRef = CFStringCreateWithCString(NULL, format, kCFStringEncodingASCII);
    if (formatRef) {
        CFStringRef text = CFStringCreateWithFormatAndArguments(NULL, NULL, formatRef, args);
        CFRelease(formatRef);
        if (text) {
            ConsoleMessage* message = new ConsoleMessage(text);
            if (message) {
                if (message->send() != noErr)
                    delete message;
            }
            result = ::CFStringGetLength(text);
            ::CFRelease(text);
        }
    }
    return result;
}

#endif /* REDIRECT_VFPRINTF */

static OSStatus OpenSessionJNI(JNIEnv *env, jclass *session, const char* consolePath);
static OSStatus InitSecurityJNI(JNIEnv *env, jclass session);

#if TARGET_RT_MAC_MACHO

#import <mach-o/dyld.h>

static jclass netscape_oji_JNIUtils = NULL;
static jmethodID netscape_oji_JNIUtils_checkClassAccessID = NULL;
static jmethodID netscape_oji_JNIUtils_checkClassAccess2ID = NULL;
static jmethodID netscape_oji_JNIUtils_fixProtectionDomainsID = NULL;

typedef jclass (*CLASS_FORNAME0) (JNIEnv *, jclass, jstring, jboolean, jobject);
// Java 1.4.X only (start)
typedef jobjectArray (*CLASS_GETDECLAREDFIELDS0) (JNIEnv *, jobject, jboolean);
typedef jobjectArray (*CLASS_GETDECLAREDMETHODS0) (JNIEnv *, jobject, jboolean);
typedef jobjectArray (*CLASS_GETDECLAREDCONSTRUCTORS0) (JNIEnv *, jobject, jboolean);
// Java 1.4.X only (end)
// Java 1.3.1 only (start)
typedef jobjectArray (*CLASS_GETFIELDS0) (JNIEnv *, jobject, jint);
typedef jobjectArray (*CLASS_GETMETHODS0) (JNIEnv *, jobject, jint);
typedef jobjectArray (*CLASS_GETCONSTRUCTORS0) (JNIEnv *, jobject, jint);
typedef jobject (*CLASS_GETFIELD0) (JNIEnv *, jobject, jstring, jint);
typedef jobject (*CLASS_GETMETHOD0) (JNIEnv *, jobject, jstring, jobjectArray, jint);
typedef jobject (*CLASS_GETCONSTRUCTOR0) (JNIEnv *, jobject, jobjectArray, jint);
// Java 1.3.1 only (end)
typedef jobjectArray (*CLASS_GETDECLAREDCLASSES0) (JNIEnv *, jobject);
typedef jobject (*ACCESSCONTROLLER_GETSTACK) (JNIEnv *, jclass);

// We use the following routine to "hook" a Java native method --
// Class.forName0().  This is part of a fix for a security hole in Java
// versions prior to 1.4.2_06 (though the earliest one we're actually
// concerned about is 1.3.1).  See the comment to JNIUtils.checkClassAccess().
// That method will throw a SecurityException when appropriate.
JNIEXPORT jclass JNICALL
	Java_java_lang_Class_forName0_redirect(JNIEnv *env, jclass clazz,
	jstring name, jboolean initialize, jobject loader)
{
	jclass retval = NULL;
	NSSymbol symbol = NULL;
	static CLASS_FORNAME0 forName0 = NULL;
	if (!forName0) {
		/* We assume the library containing the symbol we're looking for
		 * has already been loaded -- if it hasn't the following call will
		 * cause a failed-assertion-style crash.  We might be running in
		 * either Java 1.4.X or Java 1.3.1, so we don't want to explicitly
		 * load (or link to) a particular library.
		 */
		if (symbol = NSLookupAndBindSymbol("_Java_java_lang_Class_forName0"))
			forName0 = (CLASS_FORNAME0) NSAddressOfSymbol(symbol);
	}
	if (forName0)
		retval = forName0(env, clazz, name, initialize, loader);
	static unsigned entered = 0;
	if (++entered > 1) {
		--entered;
		return retval;
	}
	env->CallStaticVoidMethod(netscape_oji_JNIUtils,
		netscape_oji_JNIUtils_checkClassAccessID, name);
	if (env->ExceptionCheck())
		retval = NULL;
	--entered;
	return retval;
}

/* The following ten routines (from Java_java_lang_Class_getDeclaredFields0_redirect()
 * to Java_java_lang_Class_getDeclaredClasses0_redirect()) constitute a fix
 * for a (relatively minor) security hole in Sun's and Apple's JVMs (all
 * versions including the latest, 1.5.0_02).  Without this fix, it's possible
 * to call an applet's "EmbeddedFrame"'s methods via JavaScript-to-Java
 * LiveConnect even when access to this class has been denied using the
 * "package access" setting in jre/lib/security/java.security.  Create the
 * following method in your applet:
 *
 *  public Frame getAppletFrame() {
 *      Component c = this;
 *      while (!(c instanceof Frame))
 *          c = (Component) c.getParent();
 *      return (Frame) c;
 *  }
 *
 * Then call it from JavaScript as follows:
 *
 *  document.[appletName].getAppletFrame().[appletFrameMethod]();
 *
 */
JNIEXPORT jobjectArray JNICALL
	Java_java_lang_Class_getDeclaredFields0_redirect(JNIEnv *env, jobject self,
	jboolean publicOnly)
{
	jobjectArray retval = NULL;
	NSSymbol symbol = NULL;
	static CLASS_GETDECLAREDFIELDS0 getFields = NULL;
	if (!getFields) {
		if (symbol = NSLookupAndBindSymbol("_JVM_GetClassDeclaredFields"))
			getFields = (CLASS_GETDECLAREDFIELDS0) NSAddressOfSymbol(symbol);
	}
	if (getFields)
		retval = getFields(env, self, publicOnly);
	static unsigned entered = 0;
	if (++entered > 1) {
		--entered;
		return retval;
	}
	env->CallStaticVoidMethod(netscape_oji_JNIUtils,
		netscape_oji_JNIUtils_checkClassAccess2ID, self);
	if (env->ExceptionCheck())
		retval = NULL;
	--entered;
	return retval;
}

JNIEXPORT jobjectArray JNICALL
	Java_java_lang_Class_getDeclaredMethods0_redirect(JNIEnv *env, jobject self,
	jboolean publicOnly)
{
	jobjectArray retval = NULL;
	NSSymbol symbol = NULL;
	static CLASS_GETDECLAREDMETHODS0 getMethods = NULL;
	if (!getMethods) {
		if (symbol = NSLookupAndBindSymbol("_JVM_GetClassDeclaredMethods"))
			getMethods = (CLASS_GETDECLAREDMETHODS0) NSAddressOfSymbol(symbol);
	}
	if (getMethods)
		retval = getMethods(env, self, publicOnly);
	static unsigned entered = 0;
	if (++entered > 1) {
		--entered;
		return retval;
	}
	env->CallStaticVoidMethod(netscape_oji_JNIUtils,
		netscape_oji_JNIUtils_checkClassAccess2ID, self);
	if (env->ExceptionCheck())
		retval = NULL;
	--entered;
	return retval;
}

JNIEXPORT jobjectArray JNICALL
	Java_java_lang_Class_getDeclaredConstructors0_redirect(JNIEnv *env, jobject self,
	jboolean publicOnly)
{
	jobjectArray retval = NULL;
	NSSymbol symbol = NULL;
	static CLASS_GETDECLAREDCONSTRUCTORS0 getConstructors = NULL;
	if (!getConstructors) {
		if (symbol = NSLookupAndBindSymbol("_JVM_GetClassDeclaredConstructors"))
			getConstructors = (CLASS_GETDECLAREDCONSTRUCTORS0) NSAddressOfSymbol(symbol);
	}
	if (getConstructors)
		retval = getConstructors(env, self, publicOnly);
	static unsigned entered = 0;
	if (++entered > 1) {
		--entered;
		return retval;
	}
	env->CallStaticVoidMethod(netscape_oji_JNIUtils,
		netscape_oji_JNIUtils_checkClassAccess2ID, self);
	if (env->ExceptionCheck())
		retval = NULL;
	--entered;
	return retval;
}

JNIEXPORT jobjectArray JNICALL
	Java_java_lang_Class_getFields0_redirect(JNIEnv *env, jobject self, jint which)
{
	jobjectArray retval = NULL;
	NSSymbol symbol = NULL;
	static CLASS_GETFIELDS0 getFields = NULL;
	if (!getFields) {
		if (symbol = NSLookupAndBindSymbol("_JVM_GetClassFields"))
			getFields = (CLASS_GETFIELDS0) NSAddressOfSymbol(symbol);
	}
	if (getFields)
		retval = getFields(env, self, which);
	static unsigned entered = 0;
	if (++entered > 1) {
		--entered;
		return retval;
	}
	env->CallStaticVoidMethod(netscape_oji_JNIUtils,
		netscape_oji_JNIUtils_checkClassAccess2ID, self);
	if (env->ExceptionCheck())
		retval = NULL;
	--entered;
	return retval;
}

JNIEXPORT jobjectArray JNICALL
	Java_java_lang_Class_getMethods0_redirect(JNIEnv *env, jobject self, jint which)
{
	jobjectArray retval = NULL;
	NSSymbol symbol = NULL;
	static CLASS_GETMETHODS0 getMethods = NULL;
	if (!getMethods) {
		if (symbol = NSLookupAndBindSymbol("_JVM_GetClassMethods"))
			getMethods = (CLASS_GETMETHODS0) NSAddressOfSymbol(symbol);
	}
	if (getMethods)
		retval = getMethods(env, self, which);
	static unsigned entered = 0;
	if (++entered > 1) {
		--entered;
		return retval;
	}
	env->CallStaticVoidMethod(netscape_oji_JNIUtils,
		netscape_oji_JNIUtils_checkClassAccess2ID, self);
	if (env->ExceptionCheck())
		retval = NULL;
	--entered;
	return retval;
}

JNIEXPORT jobjectArray JNICALL
	Java_java_lang_Class_getConstructors0_redirect(JNIEnv *env, jobject self, jint which)
{
	jobjectArray retval = NULL;
	NSSymbol symbol = NULL;
	static CLASS_GETCONSTRUCTORS0 getConstructors = NULL;
	if (!getConstructors) {
		if (symbol = NSLookupAndBindSymbol("_JVM_GetClassConstructors"))
			getConstructors = (CLASS_GETCONSTRUCTORS0) NSAddressOfSymbol(symbol);
	}
	if (getConstructors)
		retval = getConstructors(env, self, which);
	static unsigned entered = 0;
	if (++entered > 1) {
		--entered;
		return retval;
	}
	env->CallStaticVoidMethod(netscape_oji_JNIUtils,
		netscape_oji_JNIUtils_checkClassAccess2ID, self);
	if (env->ExceptionCheck())
		retval = NULL;
	--entered;
	return retval;
}

JNIEXPORT jobject JNICALL
	Java_java_lang_Class_getField0_redirect(JNIEnv *env, jobject self,
	jstring name, jint which)
{
	jobject retval = NULL;
	NSSymbol symbol = NULL;
	static CLASS_GETFIELD0 getField = NULL;
	if (!getField) {
		if (symbol = NSLookupAndBindSymbol("_JVM_GetClassField"))
			getField = (CLASS_GETFIELD0) NSAddressOfSymbol(symbol);
	}
	if (getField)
		retval = getField(env, self, name, which);
	static unsigned entered = 0;
	if (++entered > 1) {
		--entered;
		return retval;
	}
	env->CallStaticVoidMethod(netscape_oji_JNIUtils,
		netscape_oji_JNIUtils_checkClassAccess2ID, self);
	if (env->ExceptionCheck())
		retval = NULL;
	--entered;
	return retval;
}

JNIEXPORT jobject JNICALL
	Java_java_lang_Class_getMethod0_redirect(JNIEnv *env, jobject self,
	jstring name, jobjectArray parameterTypes, jint which)
{
	jobject retval = NULL;
	NSSymbol symbol = NULL;
	static CLASS_GETMETHOD0 getMethod = NULL;
	if (!getMethod) {
		if (symbol = NSLookupAndBindSymbol("_JVM_GetClassMethod"))
			getMethod = (CLASS_GETMETHOD0) NSAddressOfSymbol(symbol);
	}
	if (getMethod)
		retval = getMethod(env, self, name, parameterTypes, which);
	static unsigned entered = 0;
	if (++entered > 1) {
		--entered;
		return retval;
	}
	env->CallStaticVoidMethod(netscape_oji_JNIUtils,
		netscape_oji_JNIUtils_checkClassAccess2ID, self);
	if (env->ExceptionCheck())
		retval = NULL;
	--entered;
	return retval;
}

JNIEXPORT jobject JNICALL
	Java_java_lang_Class_getConstructor0_redirect(JNIEnv *env, jobject self,
	jobjectArray parameterTypes, jint which)
{
	jobject retval = NULL;
	NSSymbol symbol = NULL;
	static CLASS_GETCONSTRUCTOR0 getConstructor = NULL;
	if (!getConstructor) {
		if (symbol = NSLookupAndBindSymbol("_JVM_GetClassConstructor"))
			getConstructor = (CLASS_GETCONSTRUCTOR0) NSAddressOfSymbol(symbol);
	}
	if (getConstructor)
		retval = getConstructor(env, self, parameterTypes, which);
	static unsigned entered = 0;
	if (++entered > 1) {
		--entered;
		return retval;
	}
	env->CallStaticVoidMethod(netscape_oji_JNIUtils,
		netscape_oji_JNIUtils_checkClassAccess2ID, self);
	if (env->ExceptionCheck())
		retval = NULL;
	--entered;
	return retval;
}

JNIEXPORT jobjectArray JNICALL
	Java_java_lang_Class_getDeclaredClasses0_redirect(JNIEnv *env, jobject self)
{
	jobjectArray retval = NULL;
	NSSymbol symbol = NULL;
	static CLASS_GETDECLAREDCLASSES0 getClasses = NULL;
	if (!getClasses) {
		/* We assume the library containing the symbol we're looking for
		 * has already been loaded -- if it hasn't the following call will
		 * cause a failed-assertion-style crash.  We might be running in
		 * either Java 1.4.X or Java 1.3.1, so we don't want to explicitly
		 * load (or link to) a particular library.
		 */
		if (symbol = NSLookupAndBindSymbol("_JVM_GetDeclaredClasses"))
			getClasses = (CLASS_GETDECLAREDCLASSES0) NSAddressOfSymbol(symbol);
	}
	if (getClasses)
		retval = getClasses(env, self);
	static unsigned entered = 0;
	if (++entered > 1) {
		--entered;
		return retval;
	}
	env->CallStaticVoidMethod(netscape_oji_JNIUtils,
		netscape_oji_JNIUtils_checkClassAccess2ID, self);
	if (env->ExceptionCheck())
		retval = NULL;
	--entered;
	return retval;
}

/* The following routine is another hook for a Java native method.  It's part
 * of a fix for a design flaw in the LiveConnect implentation of the original
 * MRJ Plugin -- it sometimes failed to elevate an applet's privileges when
 * the applet's code was called via LiveConnect.  See JNIUtils.
 * fixProtectionDomains() for more information.
 */
JNIEXPORT jobject JNICALL
	Java_java_security_AccessController_getStackAccessControlContext_redirect(JNIEnv *env,
	jclass self)
{
	jobject retval = NULL;
	NSSymbol symbol = NULL;
	static ACCESSCONTROLLER_GETSTACK getStack = NULL;
	if (!getStack) {
		/* We assume the library containing the symbol we're looking for
		 * has already been loaded -- if it hasn't the following call will
		 * cause a failed-assertion-style crash.  We might be running in
		 * either Java 1.4.X or Java 1.3.1, so we don't want to explicitly
		 * load (or link to) a particular library.
		 */
		if (symbol = NSLookupAndBindSymbol("_Java_java_security_AccessController_getStackAccessControlContext"))
			getStack = (ACCESSCONTROLLER_GETSTACK) NSAddressOfSymbol(symbol);
	}
	if (!getStack)
		return NULL;
	static unsigned entered = 0;
	if (++entered == 1) {
		env->CallStaticVoidMethod(netscape_oji_JNIUtils,
			netscape_oji_JNIUtils_fixProtectionDomainsID);
		if (env->ExceptionCheck()) {
			--entered;
			return NULL;
		}
	}
	retval = getStack(env, self);
	--entered;
	return retval;
}

#endif	//#if TARGET_RT_MAC_MACHO

typedef JNIEXPORT void JNICALL (*JVM_SETPROTECTIONDOMAIN) (JNIEnv *, jclass, jobject);
typedef JNIEXPORT jobject JNICALL (*JVM_GETPROTECTIONDOMAIN) (JNIEnv *, jclass);
typedef JNIEXPORT jobject JNICALL (*JVM_GETCLASSLOADER) (JNIEnv *, jclass);

/* The following two methods (called from JNIUtils.fixProtectionDomains())
 * allow us to bypass standard Java permission checks when getting and setting
 * a class's protection domain.
 */
JNIEXPORT void JNICALL
	Java_netscape_oji_JNIUtils_setClassProtectionDomain(JNIEnv *env, jclass self,
	jclass clazz, jobject newProtectionDomain)
{
#if TARGET_RT_MAC_MACHO
	NSSymbol symbol = NULL;
	static JVM_SETPROTECTIONDOMAIN JVM_SetProtectionDomain = NULL;
	if (!JVM_SetProtectionDomain) {
		/* We assume the library containing the symbol we're looking for
		 * has already been loaded -- if it hasn't the following call will
		 * cause a failed-assertion-style crash.  We might be running in
		 * either Java 1.4.X or Java 1.3.1, so we don't want to explicitly
		 * load (or link to) a particular library.
		 */
		if (symbol = NSLookupAndBindSymbol("_JVM_SetProtectionDomain"))
			JVM_SetProtectionDomain = (JVM_SETPROTECTIONDOMAIN) NSAddressOfSymbol(symbol);
	}
	if (JVM_SetProtectionDomain && clazz && newProtectionDomain)
		JVM_SetProtectionDomain(env, clazz, newProtectionDomain);
#endif
}

JNIEXPORT jobject JNICALL
	Java_netscape_oji_JNIUtils_getClassProtectionDomain(JNIEnv *env, jclass self,
	jclass clazz)
{
#if TARGET_RT_MAC_MACHO
	jobject retval = NULL;
	NSSymbol symbol = NULL;
	static JVM_GETPROTECTIONDOMAIN JVM_GetProtectionDomain = NULL;
	if (!JVM_GetProtectionDomain) {
		/* We assume the library containing the symbol we're looking for
		 * has already been loaded -- if it hasn't the following call will
		 * cause a failed-assertion-style crash.  We might be running in
		 * either Java 1.4.X or Java 1.3.1, so we don't want to explicitly
		 * load (or link to) a particular library.
		 */
		if (symbol = NSLookupAndBindSymbol("_JVM_GetProtectionDomain"))
			JVM_GetProtectionDomain = (JVM_GETPROTECTIONDOMAIN) NSAddressOfSymbol(symbol);
	}
	if (JVM_GetProtectionDomain && clazz)
		retval = JVM_GetProtectionDomain(env, clazz);
	return retval;
#endif
}

/* The following two methods are needed by ProxyClassLoader in
 * ProxyClassLoaderFactory.java.
 */
JNIEXPORT void JNICALL
	Java_netscape_oji_ProxyClassLoaderFactory_holdSecureEnv(JNIEnv *env,
	jclass self, jint secureEnvAddr)
{
	if (secureEnvAddr == 0)
		return;
	CSecureEnv *secureEnv = (CSecureEnv *) secureEnvAddr;
	NS_ADDREF(secureEnv);
}

JNIEXPORT void JNICALL
	Java_netscape_oji_ProxyClassLoaderFactory_releaseSecureEnv(JNIEnv *env,
	jclass self, jint secureEnvAddr)
{
	if (secureEnvAddr == 0)
		return;
	CSecureEnv *secureEnv = (CSecureEnv *) secureEnvAddr;
	NS_RELEASE(secureEnv);
}

/* This must be called just after the JVM was started and before the AWT
 * libraries are loaded -- so that the "poor man's console" is initialized
 * early enough.  If the "poor man's console" is initialized later,
 * Sun/Apple's Java Console won't work properly.
 */
static OSStatus OpenSessionJNI(JNIEnv *env, jclass *session, const char* consolePath)
{
	if (session)
		*session = NULL;
	if (!env)
		return EINVAL;
	jclass sessionLocal = env->FindClass("netscape/oji/MRJSession");
	if (!sessionLocal)
		return ENOENT;
	if (session)
		*session = (jclass) env->NewGlobalRef(sessionLocal);
	jmethodID openMethod = env->GetStaticMethodID(sessionLocal, "open",
		"(Ljava/lang/String;)V");
	if (!openMethod)
		return ENOENT;
	jstring path = env->NewStringUTF(consolePath);
	if (!path)
		return ENOMEM;
	env->CallStaticVoidMethod(sessionLocal, openMethod, path);
	if (env->ExceptionCheck())
		return ENOENT;
	env->DeleteLocalRef(sessionLocal);
	return noErr;
}

/* This must be called after JavaEmbeddingPlugin.bundle's security has been
 * initialized (i.e. after the AWT libraries are loaded).
 */
static OSStatus InitSecurityJNI(JNIEnv *env, jclass session)
{
	if (!env || !session)
		return EINVAL;
	jmethodID initSecurityMethod = env->GetStaticMethodID(session, "initSecurity",
		"()V");
	if (!initSecurityMethod)
		return ENOENT;
	env->CallStaticVoidMethod(session, initSecurityMethod);
	if (env->ExceptionCheck())
		return ENOENT;
	jclass jniUtilsClass = env->FindClass("netscape/oji/JNIUtils");
	jclass loaderFactoryClass = env->FindClass("netscape/oji/ProxyClassLoaderFactory");
	if (!jniUtilsClass || !loaderFactoryClass)
		return ENOENT;
#if TARGET_RT_MAC_MACHO
	netscape_oji_JNIUtils = (jclass) env->NewGlobalRef(jniUtilsClass);
#endif
	JNINativeMethod JNIUtilsMethods[] = {
		{"setClassProtectionDomain",
			"(Ljava/lang/Class;Ljava/security/ProtectionDomain;)V",
			(void *) &Java_netscape_oji_JNIUtils_setClassProtectionDomain},
		{"getClassProtectionDomain",
			"(Ljava/lang/Class;)Ljava/security/ProtectionDomain;",
			(void *) &Java_netscape_oji_JNIUtils_getClassProtectionDomain}
		};
	env->RegisterNatives(jniUtilsClass, JNIUtilsMethods,
		sizeof(JNIUtilsMethods)/sizeof(JNINativeMethod));
	env->DeleteLocalRef(jniUtilsClass);
	JNINativeMethod LoaderFactoryMethods[] = {
		{"holdSecureEnv", "(I)V",
			(void *) &Java_netscape_oji_ProxyClassLoaderFactory_holdSecureEnv},
		{"releaseSecureEnv", "(I)V",
			(void *) &Java_netscape_oji_ProxyClassLoaderFactory_releaseSecureEnv}
		};
	env->RegisterNatives(loaderFactoryClass, LoaderFactoryMethods,
		sizeof(LoaderFactoryMethods)/sizeof(JNINativeMethod));
	env->DeleteLocalRef(loaderFactoryClass);
	if (env->ExceptionCheck())
		return ENOENT;
#if TARGET_RT_MAC_MACHO
	jclass classClass = env->FindClass("java/lang/Class");
	jclass accessControllerClass = env->FindClass("java/security/AccessController");
	if (!classClass || !accessControllerClass)
		return ENOENT;
	netscape_oji_JNIUtils_checkClassAccessID =
		env->GetStaticMethodID(netscape_oji_JNIUtils, "checkClassAccess",
		"(Ljava/lang/String;)V");
	netscape_oji_JNIUtils_checkClassAccess2ID =
		env->GetStaticMethodID(netscape_oji_JNIUtils, "checkClassAccess",
		"(Ljava/lang/Class;)V");
	netscape_oji_JNIUtils_fixProtectionDomainsID =
		env->GetStaticMethodID(netscape_oji_JNIUtils, "fixProtectionDomains",
		"()V");
	if (env->ExceptionCheck())
		return ENOENT;
	JNINativeMethod ClassMethods14[] = {
		{"forName0",
			"(Ljava/lang/String;ZLjava/lang/ClassLoader;)Ljava/lang/Class;",
			(void *) &Java_java_lang_Class_forName0_redirect},
		{"getDeclaredFields0",
			"(Z)[Ljava/lang/reflect/Field;",
			(void *) &Java_java_lang_Class_getDeclaredFields0_redirect},
		{"getDeclaredMethods0",
			"(Z)[Ljava/lang/reflect/Method;",
			(void *) &Java_java_lang_Class_getDeclaredMethods0_redirect},
		{"getDeclaredConstructors0",
			"(Z)[Ljava/lang/reflect/Constructor;",
			(void *) &Java_java_lang_Class_getDeclaredConstructors0_redirect},
		{"getDeclaredClasses0",
			"()[Ljava/lang/Class;",
			(void *) &Java_java_lang_Class_getDeclaredClasses0_redirect}
		};
	JNINativeMethod ClassMethods131[] = {
		{"forName0",
			"(Ljava/lang/String;ZLjava/lang/ClassLoader;)Ljava/lang/Class;",
			(void *) &Java_java_lang_Class_forName0_redirect},
		{"getFields0",
			"(I)[Ljava/lang/reflect/Field;",
			(void *) &Java_java_lang_Class_getFields0_redirect},
		{"getMethods0",
			"(I)[Ljava/lang/reflect/Method;",
			(void *) &Java_java_lang_Class_getMethods0_redirect},
		{"getConstructors0",
			"(I)[Ljava/lang/reflect/Constructor;",
			(void *) &Java_java_lang_Class_getConstructors0_redirect},
		{"getField0",
			"(Ljava/lang/String;I)Ljava/lang/reflect/Field;",
			(void *) &Java_java_lang_Class_getField0_redirect},
		{"getMethod0",
			"(Ljava/lang/String;[Ljava/lang/Class;I)Ljava/lang/reflect/Method;",
			(void *) &Java_java_lang_Class_getMethod0_redirect},
		{"getConstructor0",
			"([Ljava/lang/Class;I)Ljava/lang/reflect/Constructor;",
			(void *) &Java_java_lang_Class_getConstructor0_redirect},
		{"getDeclaredClasses0",
			"()[Ljava/lang/Class;",
			(void *) &Java_java_lang_Class_getDeclaredClasses0_redirect}
		};
	env->RegisterNatives(classClass, ClassMethods14,
		sizeof(ClassMethods14)/sizeof(JNINativeMethod));
	if (env->ExceptionCheck()) {
		env->ExceptionClear();
		env->RegisterNatives(classClass, ClassMethods131,
			sizeof(ClassMethods131)/sizeof(JNINativeMethod));
		if (env->ExceptionCheck())
			return ENOENT;
	}
	env->DeleteLocalRef(classClass);
	JNINativeMethod AccessControllerMethods[] = {
		{"getStackAccessControlContext",
			"()Ljava/security/AccessControlContext;",
			(void *) &Java_java_security_AccessController_getStackAccessControlContext_redirect}
		};
	env->RegisterNatives(accessControllerClass, AccessControllerMethods,
		sizeof(AccessControllerMethods)/sizeof(JNINativeMethod));
	env->DeleteLocalRef(accessControllerClass);
	if (env->ExceptionCheck())
		return ENOENT;
#endif
	return noErr;
}

MRJSession::MRJSession(MRJPlugin *plugin)
    :   mStatus(noErr), mMainEnv(NULL), mJavaVM(NULL), mSession(NULL),
        mPlugin(plugin), mFirst(NULL), mLast(NULL), mMessageMonitor(NULL),
        mLockCount(0)
{
}

MRJSession::~MRJSession()
{
    close();
}

/* Called on a secondary thread in AppletView loadAWTLibraries, just after the
 * AWT libraries have finished loading.  We do this to make sure applets (and
 * the browser) don't start using Java until it has been fully initialized
 * (including LiveConnect).  This is required in order for the JVM's proxy
 * support to work properly:  The FindProxyForURL() JavaScript method is
 * invoked (in AutoProxyHandler.getProxyInfo()) using Java-to-JavaScript
 * LiveConnect), just after an MRJPluginInstance (including the very first
 * one) has been created.  And (of course) this needs to happen (for the first
 * time) before the first Java applet has even been loaded.
 */
void afterInitAWT(void *info)
{
    if (!info)
        return;
    MRJSession *session = (MRJSession *) info;
    if (session->getStatus() != noErr)
        return;
    JNIEnv *env = session->getCurrentEnv();
    OSStatus retval = InitSecurityJNI(env, session->getClass());
    if (retval == noErr)
        retval = PostInitLiveConnectSupport(session->getPlugin());
    if (retval != noErr)
        session->setStatus(retval);
}

OSStatus MRJSession::open(const char* consolePath)
{
    // Use vanilla JNI invocation API to fire up a fresh JVM.
    string classPath = getClassPath();
    string pluginHome = getPluginHome();
    JavaVMOption theOptions[] = {
    	{ (char*) classPath.c_str() },
    	{ (char*) pluginHome.c_str() },
#if REDIRECT_VFPRINTF
    	{ "vfprintf", NewMachOFunctionPointer(&java_vfprintf) }
#endif
    };

    JavaVMInitArgs theInitArgs = {
    	JNI_VERSION_1_2,
    	sizeof(theOptions) / sizeof(JavaVMOption),
    	theOptions,
    	JNI_TRUE
    };

	/* We need to be very careful how we launch the JVM when working with the
	 * Java Embedding Plugin, because in Apple's 1.4.X and 1.5.X JVMs,
	 * JNI_CreateJavaVM() does its work mostly asynchronously (i.e. it returns
	 * well before the JVM has finished initializing).  JEPCreateJavaVMAndInitAWT()
	 * doesn't return until the AWT libraries have loaded (which allows us to
	 * avoid hangs in libawt.jnilib's JNI_OnLoad() when we start using code
	 * from it).  But we need to initialize the MRJ Plugin's "poor man's
	 * console" just after we've created the JVM, and before the AWT libraries
	 * have started loading.  This is why, in what follows, we first call
	 * JEPCreateJavaVM(), then call OpenSessionJNI() (which creates the "poor
	 * man's console"), then call JEPCreateJavaVMAndInitAWT() (which, because
	 * we've called it after JEPCreateJavaVM(), just loads the AWT libraries
	 * (and returns after they've finished loading)).
	 */
    if (JEPCheckBundle()) {
        mStatus = JEPCreateJavaVM(theOptions, sizeof(theOptions)/sizeof(JavaVMOption),
            &mJavaVM, (void **) &mMainEnv);
        if (mStatus == EALREADY)
            mStatus = noErr;
    } else {
        mStatus = ::JNI_CreateJavaVM(&mJavaVM, (void**) &mMainEnv, &theInitArgs);
    }

    if (mStatus != noErr)
        return mStatus;

    // create a monitor for the message queue to unblock Java threads.
    mMessageMonitor = new MRJMonitor(this);

    JNIEnv* env = mMainEnv;
    if ((mStatus = OpenSessionJNI(env, &mSession, consolePath)) != noErr)
        return mStatus;
    if ((mStatus = InitLiveConnectSupport(getPlugin())) != noErr)
        return mStatus;
    if (JEPCheckBundle()) {
        mStatus = JEPCreateJavaVMAndInitAWT(theOptions,
            sizeof(theOptions)/sizeof(JavaVMOption), NULL, NULL, afterInitAWT, this);
        if (mStatus == EALREADY)
            mStatus = noErr;
        if (mStatus != noErr)
            return mStatus;
    } else {
        if ((mStatus = InitSecurityJNI(env, mSession)) != noErr)
            return mStatus;
        if ((mStatus = PostInitLiveConnectSupport(getPlugin())) != noErr)
            return mStatus;
    }

    theSession = this;
    debugUpdates =
        JEPGetJavaRuntimeParameterBoolean(CFSTR(JEP_DEBUG_UPDATES_OPTION));
	debugProxyJNI =
        JEPGetJavaRuntimeParameterBoolean(CFSTR(JEP_DEBUG_PROXYJNI_OPTION));
	debugPrinting =
		JEPGetJavaRuntimeParameterBoolean(CFSTR(JEP_DEBUG_PRINTING_OPTION));

#if REDIRECT_VFPRINTF
    // XXX test the vfprintf function.
    jclass notThere = env->FindClass("class/not/Found");
    if (env->ExceptionCheck()) {
        env->ExceptionDescribe();
        env->ExceptionClear();
    }
#endif

    return mStatus;
}

OSStatus MRJSession::close()
{
    theSession = NULL;

    if (mJavaVM) {
    	if (mMessageMonitor != NULL) {
    		mMessageMonitor->notifyAll();
    		delete mMessageMonitor;
    		mMessageMonitor = NULL;
    	}
    	
    	if (mSession) {
    	    jclass session = mSession;
    	    JNIEnv* env = mMainEnv;
            jmethodID closeMethod = env->GetStaticMethodID(session, "close", "()V");
            if (closeMethod)
                env->CallStaticVoidMethod(session, closeMethod);
            else
                env->ExceptionClear();
    	    env->DeleteGlobalRef(session);
    	    mSession = NULL;
        }

#if !TARGET_RT_MAC_MACHO // XXX right now, this hangs Chimera...
        mJavaVM->DestroyJavaVM();
#endif
        mJavaVM = NULL;
    }
    return noErr;
}

JNIEnv* MRJSession::getCurrentEnv()
{
	jint retval;
	JNIEnv* env;
	int jvmVersion = JNI_VERSION_1_2;
	/* We don't actually use any 1.4.X-specific JNI calls or features, and
	 * we should (presumably) preserve compatibility between the "native"
	 * JNIEnv structure and the Mozilla family's ProxyJNIEnv class (see
	 * modules/oji/src/ProxyJNI.cpp).
	 */
	//if (JEPCheckBundle()) {
	//	jvmVersion = JNI_VERSION_1_4;
	//}

	if (!mJavaVM)
		return NULL;

	retval = mJavaVM->GetEnv((void**)&env, jvmVersion);
	if (retval == JNI_EDETACHED) {
		if (!mJavaVM->AttachCurrentThread((void **)&env, nil))
			retval = JNI_OK;
	}
	if (retval == JNI_OK)
		return env;

    return NULL;
}

JNIEnv* MRJSession::getMainEnv()
{
	return mMainEnv;
}

JavaVM* MRJSession::getJavaVM()
{
    return mJavaVM;
}

Boolean MRJSession::onMainThread()
{
	return (getCurrentEnv() == mMainEnv);
}

inline StringPtr c2p(const char* cstr, StringPtr pstr)
{
	pstr[0] = (unsigned char)strlen(cstr);
	::BlockMoveData(cstr, pstr + 1, pstr[0]);
	return pstr;
}

Boolean MRJSession::addToClassPath(const FSSpec& fileSpec)
{
    // if the Java VM has started already, it's too late to do this (for now).
    if (mJavaVM)
        return false;

    // keep accumulating paths.
    FSRef ref;
    OSStatus status = FSpMakeFSRef(&fileSpec, &ref);
    if (status == noErr)
        mClassPath.push_back(ref);

    return true;
}

Boolean MRJSession::addToClassPath(const char* dirPath)
{
	// Need to convert the path into an FSSpec, and add it MRJ's class path.
	Str255 path;
	FSSpec pathSpec;
	OSStatus status = ::FSMakeFSSpec(0, 0, c2p(dirPath, path), &pathSpec);
	if (status == noErr)
		return addToClassPath(pathSpec);
	return false;
}

Boolean MRJSession::addURLToClassPath(const char* fileURL)
{
    Boolean success = false;
    // Use CFURL, FSRef and FSSpec?
    CFURLRef fileURLRef = CFURLCreateWithBytes(NULL, (UInt8*)fileURL, strlen(fileURL),
                                               kCFStringEncodingUTF8, NULL);
    if (fileURLRef) {
        FSRef fsRef;
        if (CFURLGetFSRef(fileURLRef, &fsRef)) {
            mClassPath.push_back(fsRef);
            success = true;
        }
        CFRelease(fileURLRef);
    }
    return success;
}

char* MRJSession::getProperty(const char* propertyName)
{
	char* result = NULL;
#if !TARGET_CARBON
	OSStatus status = noErr;
	JMTextRef nameRef = NULL, valueRef = NULL;
	status = ::JMNewTextRef(mSession, &nameRef, kTextEncodingMacRoman, propertyName, strlen(propertyName));
	if (status == noErr) {
		status = ::JMGetSessionProperty(mSession, nameRef, &valueRef);
		::JMDisposeTextRef(nameRef);
		if (status == noErr && valueRef != NULL) {
			UInt32 valueLength = 0;
			status = ::JMGetTextLengthInBytes(valueRef, kTextEncodingMacRoman, &valueLength);
			if (status == noErr) {
				result = new char[valueLength + 1];
				if (result != NULL) {
					UInt32 actualLength;
					status = ::JMGetTextBytes(valueRef, kTextEncodingMacRoman, result, valueLength, &actualLength);
					result[valueLength] = '\0';
				}
				::JMDisposeTextRef(valueRef);
			}
		}
	}
#endif
	return result;
}

void MRJSession::setStatus(OSStatus status)
{
	mStatus = status;
}

OSStatus MRJSession::getStatus()
{
	return mStatus;
}

void MRJSession::idle(UInt32 milliseconds)
{
	// Each call to idle processes a single message.
	dispatchMessage();

#if !TARGET_CARBON
	// Guard against entering the VM multiple times.
	if (mLockCount == 0) {
		lock();
		mStatus = ::JMIdle(mSession, milliseconds);
		unlock();
	}
#endif
}

void MRJSession::sendMessage(NativeMessage* message, Boolean async)
{
	// can't block the main env, otherwise messages will never be processed!
	if (onMainThread()) {
		message->execute();
	} else {
		postMessage(message);
		if (!async)
			mMessageMonitor->wait();
	}
}

/**
 * Put a message on the queue.
 */
void MRJSession::postMessage(NativeMessage* message)
{
	if (mFirst == NULL) {
		mFirst = mLast = message;
	} else {
		mLast->setNext(message);
		mLast = message;
	}
	message->setNext(NULL);
}

void MRJSession::dispatchMessage()
{
	if (mFirst != NULL) {
		NativeMessage* message = mFirst;
		mFirst = message->getNext();
		if (mFirst == NULL) mLast = NULL;
		
		message->setNext(NULL);
		message->execute();
		mMessageMonitor->notify();
	}
}

void MRJSession::lock()
{
	++mLockCount;
}

void MRJSession::unlock()
{
	--mLockCount;
}

static OSStatus ref2path(const FSRef& ref, char* path, UInt32 maxPathSize)
{
    return FSRefMakePath(&ref, (UInt8*)path, maxPathSize);
}

OSStatus spec2path(const FSSpec& spec, char* path, UInt32 maxPathSize)
{
    FSRef ref;
    OSStatus status = FSpMakeFSRef(&spec, &ref);
    if (status == noErr)
        status = ref2path(ref, path, maxPathSize);
    return status;
}

string MRJSession::getClassPath()
{
    // to work around problem in Mac OS X 10.2 (Jaguar) (bugzilla #164712),
    // put our classes in the boot class path.
    string classPath("-Xbootclasspath/p:");
    
    // keep appending paths make from FSSpecs.
    MRJClassPath::const_iterator i = mClassPath.begin();
    if (i != mClassPath.end()) {
        char path[1024];
        if (ref2path(*i, path, sizeof(path)) == noErr)
            classPath += path;
        ++i;
        while (i != mClassPath.end()) {
            if (ref2path(*i, path, sizeof(path)) == noErr) {
                classPath += ":";
                classPath += path;
            }    
            ++i;
        }
    }

    return classPath;
}

string MRJSession::getPluginHome()
{
    string pluginHome("-Dnetscape.oji.plugin.home=");

    char path[1024];
    if (spec2path(thePluginSpec, path, sizeof(path)) == noErr) {
        char* lastSlash = strrchr(path, '/');
        if (lastSlash) {
            *lastSlash = '\0';
            pluginHome += path;
        }
    }

    return pluginHome;
}
