/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the MRJ Carbon OJI Plugin.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corp.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):  Patrick C. Beard <beard@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */

package netscape.oji;

import java.io.*;
import java.util.Date;
import java.util.Properties;
import java.security.Security;

/**
 * Performs startup actions on behalf of the MRJ plugin.
 * 1. Sends System.out/err to a specified disk file.
 * 2. Installs an appropriate security manager for
 *    integrating with the Netscape 6 security system.
 */
public class MRJSession {
    // Save primordial System streams.
    private static PrintStream out = null;
    private static PrintStream err = null;
    private static PrintStream console = null;
    private static Properties props = null;
    private static String pluginHome = "";

    private static Properties loadProperties() {
        Properties props = new Properties();
        try {
            InputStream propsStream = new FileInputStream(pluginHome + "/MRJPlugin.properties");
            props.load(propsStream);
            propsStream.close();
        } catch (IOException ex) {
            System.out.println("MRJSession.loadProperties(): Exception " +
                ex.getMessage());
            ex.printStackTrace();
            throw new InternalError();
        }
        return props;
    }

	// This code was "borrowed" from AppletHolderFactory.java of the Java
	// Embedding Plugin.  The previous code simply replaced other policy files
	// with the one that comes with the MRJ Plugin Carbon.  This needs either
	// to be called on the main thread (as it is here) or before a security
	// manager is created.  Original comment:  Make sure that the classes in
	// MRJPlugin.jar are granted all permissions.  See p. 117 of "Inside Java
	// 2 Platform Security" for more information.
	private static void giveOurselvesAllPermissions() {
		try {
			int n = 1;
			while (Security.getProperty("policy.url." + n) != null)
				++n;
			String ourPolicyProperty = "policy.url." + n;
			Security.setProperty(ourPolicyProperty,
				"file:" + pluginHome + "/MRJPlugin.policy");
		} catch (Exception ex) {
			System.out.println("MRJSession.giveOurselvesAllPermissions(): Exception " +
				ex.getMessage());
			ex.printStackTrace();
			throw new InternalError();
		}
	}

	// We should restrict access to the "netscape.oji" package, to avoid
	// possible security holes.  This code was "borrowed" from
	// AppletHolderFactory.java of the Java Embedding Plugin.  For some
	// reason this doesn't work in Java 1.3.1 -- the "package.access"
	// property gets changed to the correct value, but SecurityManager.
	// checkPackageAccess() ignores the change.
	private static void restrictPackageAccess() {
		try {
			String packageAccess = Security.getProperty("package.access");
			packageAccess += ",netscape.oji.";
			Security.setProperty("package.access", packageAccess);
		} catch (Exception ex) {
			System.out.println("MRJSession.restrictPackageAccess(): Exception " +
				ex.getMessage());
			ex.printStackTrace();
			throw new InternalError();
		}
	}

    public static void open(String consolePath) throws IOException {
        pluginHome = System.getProperty("netscape.oji.plugin.home");
        props = loadProperties();
        boolean append = Boolean.valueOf(props.getProperty("netscape.oji.plugin.console.append")).booleanValue();
    
        // Make sure the parent directories exist.
        File consoleFile = new File(consolePath);
        File parentFile = consoleFile.getParentFile();
        if (!parentFile.exists()) {
            parentFile.mkdirs();
        }
    
        // redirect console I/O to specified file.
        MRJSession.out = System.out;
        MRJSession.err = System.err;
        console = new PrintStream(new FileOutputStream(consolePath, append));
        System.setOut(console);
        System.setErr(console);
    }

    public static void initSecurity() {
        Date date = new Date();
        String version = props.getProperty("netscape.oji.plugin.version");
        System.out.println("MRJ Plugin for Mac OS X v" + version);
        System.out.println("[starting up Java Applet Security @ " + date + "]");

        giveOurselvesAllPermissions();
        restrictPackageAccess();

        // What follows is only necessary when not using JavaEmbeddingPlugin.bundle.
        if (System.getProperty("java.version").compareTo("1.4") >= 0)
            return;

        // bring up MRJ Applet Security.
        // In versions of the MRJ Plugin JEP before 0.8.8, we failed to create
        // and set a security manager here when doing Java 1.3.1 -- because the
        // "netscape.oji.plugin.security" property used to specify a path that
        // didn't exist in Java 1.3.1.  But Java 1.3.1's applet launcher saved
        // the day because it creates (another) security manager later on -- so
        // that we _weren't_ running with no applet security in Java 1.3.1 in
        // previous versions of the JEP.  Whew!  (The old code didn't have a
        // problem running in Java 1.4.X.)  So the following code isn't,
        // strictly speaking, actually necessary.  But I'm keeping it anyway,
        // if only for redundancy.
        if (System.getSecurityManager() == null) {
            try {
                String name = props.getProperty("netscape.oji.plugin.security");
                SecurityManager securityManager = (SecurityManager) Class.forName(name).newInstance();
                System.setSecurityManager(securityManager);
            } catch (Exception ex) {
                // Part of the reason I missed the earlier problems with this
                // code is that no error messages were displayed when it
                // failed to create and install a security manager.  Now we
                // do.
                System.err.println("MRJ Plugin: Exception while setting Java 1.3.1 security manager: " +
                    ex.getMessage());
                ex.printStackTrace();
                throw new InternalError();
            }
        }
    }

    public static void close() throws IOException {
        System.setOut(MRJSession.out);
        System.setErr(MRJSession.err);
        console.close();
    }
}
