/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the MRJ Carbon OJI Plugin.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corp.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Patrick C. Beard <beard@netscape.com>
 *   Xiaobin Lu <Xiaobin.Lu@eng.Sun.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */

package netscape.oji;

import java.net.URLClassLoader;
import java.net.URL;
import java.net.MalformedURLException;

import java.io.*;
import java.util.*;
import java.util.zip.*;
import java.lang.ref.WeakReference;

import java.security.CodeSource;
import java.security.cert.Certificate;

public abstract class ProxyClassLoaderFactory {
    static void debug(String message) {
        System.out.println("<<< " + message + " >>>");
    }

    /**
     * Loads specified class file as an array of bytes from ${netscape.oji.plugin.home}/MRJPlugin.jar.
     */
    private static byte[] getMRJPluginClassFile(String path) {
        try {
    		String homeDir = System.getProperty("netscape.oji.plugin.home");
    		ZipFile jarFile = new ZipFile(new File(homeDir, "MRJPlugin.jar"));
    		ZipEntry classEntry = jarFile.getEntry(path);
    		int size = (int) classEntry.getSize();
    		if (size > 0) {
    			byte[] data = new byte[size];
    			DataInputStream input = new DataInputStream(jarFile.getInputStream(classEntry));
    			input.readFully(data);
    			input.close();
    			jarFile.close();
    			return data;
    		}
        } catch (IOException ioe) {
        }
		return null;
    }

    /**
     * Trivial subclass of URLClassLoader that predefines the class netscape.oji.LiveConnectProxy
     * as if it were loaded from the specified codebase URL passed to the constructor.
     * @see        netscape.oji.LiveConnectProxy
     */
    private static class ProxyClassLoader extends URLClassLoader {
        private int secureEnv = 0;
        private String documentURL = null;
        private static byte[] data = getMRJPluginClassFile("netscape/oji/LiveConnectProxy.class");
        ProxyClassLoader(String documentURL, int secureEnv)
            throws MalformedURLException
        {
            super(new URL[] {new URL(documentURL)});
            this.documentURL = documentURL;
            this.secureEnv = secureEnv;
            holdSecureEnv(secureEnv);
            if (data != null) {
                Class proxyClass = defineClass("netscape.oji.LiveConnectProxy",
                                               data, 0, data.length,
                                               new CodeSource(new URL(documentURL),
                                               (Certificate[]) null));
                debug("ProxyClassLoader: defined LiveConnectProxy class.");
                debug("Here're the permissions you've got:");
                debug(proxyClass.getProtectionDomain().getPermissions().toString());
            } else {
                debug("ProxyClassLoader: failed to define LiveConnectProxy class.");
            }
        }
        protected void finalize() throws Throwable {
            releaseSecureEnv(secureEnv);
			mClassLoaders.remove(documentURL);
            super.finalize();
        }
    }

	private static native void holdSecureEnv(int secureEnv);
	private static native void releaseSecureEnv(int secureEnv);

    private static HashMap mClassLoaders = new HashMap();

    // Previous versions (before JEP 0.9.1) cached proxy classloaders in
    // mClassLoaders and re-used ones with the same codebase.  But there
    // isn't necessarily a one-to-one correspondence between documentURL
    // (the JavaScript principal's "origin") and an applet's codebase.
    // Moreover, we don't want to re-use a classloader whose protection
    // domain we've changed in JNIUtils.fixProtectionDomains().  The cost
    // of not caching the proxy classloaders is very minor.
    // As of JEP 0.9.6 (0.9.5+e) we're caching proxy classloaders again
    // (just the last one we created) but for a different reason -- see
    // below.
    public static ClassLoader createClassLoader(final String documentURL, int secureEnv)
        throws MalformedURLException
    {
        ClassLoader loader = null;
        try {
            loader = new ProxyClassLoader(documentURL, secureEnv);
            // We only cache the last one we created.
            mClassLoaders.clear();
            mClassLoaders.put(documentURL, new WeakReference(loader));
        } catch (MalformedURLException e) {
        }
        return loader;
    }

    // Mozilla.org browsers actually call this method (in getScriptClassLoader()
    // in modules/oji/src/ProxyClassLoader.cpp), so we need to have a version
    // that uses the "old" API.  Moreover, these calls are always made after
    // we've already created a ClassLoader for documentURL using the "new" API,
    // and when we don't have permissions to create a new ClassLoader.  So we
    // cache the last one created and return that instead (if it matches
    // documentURL).
    // But getScriptClassLoader() serves no useful purpose, so there's no point
    // in making this method available, after all.
    //public static ClassLoader createClassLoader(final String documentURL)
    //    throws MalformedURLException
    //{
    //    ClassLoader loader = null;
    //    WeakReference classLoaderRef = (WeakReference)
    //        mClassLoaders.get(documentURL);
    //    if (classLoaderRef != null)
    //        loader = (ClassLoader) classLoaderRef.get();
    //    return loader;
    //}
}
