/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the MRJ Carbon OJI Plugin.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corp.
 * Portions created by the Initial Developer are Copyright (C) 2001
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):  Patrick C. Beard <beard@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */

/*
	JNIUtils.java
 */

package netscape.oji;

import java.lang.reflect.*;
import java.security.CodeSource;
import java.security.ProtectionDomain;
import java.security.cert.Certificate;

import java.util.Stack;
import java.util.EmptyStackException;

public class JNIUtils {
	/**
	 * Returns a local reference for a passed-in global reference.
	 */
	public static Object NewLocalRef(Object object) {
		return object;
	}
	
	/**
	 * Returns the currently running thread.
	 */
	public static Object GetCurrentThread() {
		return Thread.currentThread();
	}

	public static void checkClassAccess(Class clazz) {
		//System.out.println("JNIUtils.checkClassAccess(Class): class is " + clazz.getName());
		checkClassAccess(clazz.getName());
	}

	// At least in Java versions prior to 1.4.2_06 (of which there isn't yet
	// any Apple release), Class.forName() doesn't always work properly when
	// doing JavaScript-to-Java LiveConnect.  I'm not sure why this is the
	// case, but it seems that calls to SecurityManager.checkPackageAccess()
	// and possibly also ClassLoader.loadClass() are sometimes being missed.
	// I work around this problem by hooking calls to Class.forName0() (in
	// MRJSession.cpp), calling this method, then calling checkPackageAccess()
	// if we're doing JavaScript-to-Java LiveConnect (which we are if the top
	// class of the current class context is netscape.oji.LiveConnectProxy).
	// See the following URLs for more info and a proof of concept:
	// http://securityfocus.com/archive/1/382072/2004-11-23/2004-11-29/0
	// http://securityfocus.com/archive/1/381940/2004-11-16/2004-11-22/0
	public static void checkClassAccess(String name) {
		Class[] classContext = stubManager.getCurrentClassContext();
		//String topClassName =
		//	classContext[classContext.length-1].getName().replace('/', '.');
		//System.out.println("JNIUtils.checkClassAccess(String): name is " + name +
		//	" topClassName is " + topClassName);
		if (!inLiveConnect(classContext))
			return;
		// The following is based on code found in URLClassLoader.loadClass()
		// and AppletClassLoader.loadClass(), among other places.  I'm not
		// sure it's all really necessary.  But I've noticed that class names
		// can sometimes have wierd formats (e.g. "[A" and "[B"), so I'm
		// keeping it just in case.
		String pkg = name.replace('/', '.');
		if (pkg.startsWith("[")) {
			int b = pkg.lastIndexOf('[') + 2;
			if (b > 1 && b < pkg.length()) {
				pkg = pkg.substring(b);
			}
		}
		int i = pkg.lastIndexOf('.');
		if (i != -1) {
			pkg = pkg.substring(0, i);
			System.getSecurityManager().checkPackageAccess(pkg);
		}
	}

	// This is called from our hook for AccessController.
	// getStackAccessControlContext() in MRJSession.cpp.  I use it to fix a
	// design flaw in the LiveConnect implementation of the original MRJ
	// Plugin -- it sometimes failed to elevate a signed applet's privileges
	// when the applet's code was called via LiveConnect.  My fix is to reset
	// the LiveConnectProxy class's protection domain, when appropriate, just
	// before a call to getStackAccessControlContext() (which takes place
	// before every permission check).  I only do this during JavaScript-to-
	// Java LiveConnect.  The effect is to either increase or decrease our
	// privileges.  If the protection domain of the "top" class
	// (netscape.oji.LiveConnectProxy) currently has ordinary privileges but
	// the protection domain of one of the other classes in the execution
	// stack has certificates, I take this to mean that a signed applet (with
	// elevated privileges) has been loaded.  If no other class has a cert-
	// less protection domain (which would cancel our elevated privileges), I
	// replace the "top" class's protection domain with the "highest" one that
	// has certificates (which gives the currently running code elevated
	// privileges).  On the other hand, if the top class currently has
	// elevated privileges but at least one class has a cert-less protection
	// domain, I take this to mean that we're currently running code from an
	// unsigned applet (with no special privileges).  In this case I replace
	// the top class's protection domain with the "highest" cert-less
	// protection domain.  I try to make these changes only as needed, for the
	// sake of efficiency.  It's probably not necessary to downgrade the top
	// class's protection domain -- the presence of a cert-less protection
	// domain in the "stack" should stop privileged actions from taking place.
	// I'm doing it on general principles (just in case) -- I'm worried about
	// the possibility that an unsigned applet loaded from the same page as a
	// signed applet might run with the signed applet's privileges (if both
	// were called via LiveConnect and the unsigned applet was called after
	// the signed one).
	public static void fixProtectionDomains() {
		Class[] classContext = stubManager.getCurrentClassContext();
		if (!inLiveConnect(classContext))
			return;
		int topPDClassOffset = getTopPDClassOffset(classContext);
		if (topPDClassOffset == -1)
			return;

		Class topPDClass = classContext[topPDClassOffset];
		//String topPDClassName = topPDClass.getName().replace('/', '.');
		//System.out.println("JNIUtils.fixProtectionDomains(): " +
		//	" topPDClassName is " + topPDClassName);
		ProtectionDomain topPD = getClassProtectionDomain(topPDClass);
		boolean topPDHasCerts = false;
		CodeSource cs = topPD.getCodeSource();
		Certificate[] certs;
		if (cs != null) {
			certs = cs.getCertificates();
			if (certs != null)
				topPDHasCerts = true;
		}
		ProtectionDomain nextPDWithCerts = null;
		ProtectionDomain nextPDWithoutCerts = null;
		for (int i = 1; i <= topPDClassOffset; ++i) {
			ProtectionDomain pd = getClassProtectionDomain(classContext[i-1]);
			if (pd == null)
				continue;
			boolean hasCerts = false;
			cs = pd.getCodeSource();
			if (cs != null) {
				certs = cs.getCertificates();
				if (certs != null)
					hasCerts = true;
			}
			if (hasCerts) {
				nextPDWithCerts = pd;
			} else {
				nextPDWithoutCerts = pd;
			}
		}

		//String topPDClassName = topPDClass.getName().replace('/', '.');
		if (topPDHasCerts) {
			if (nextPDWithoutCerts != null) {
				//System.out.println("JNIUtils.fixProtectionDomains(): Setting topPDClass " +
				//	topPDClassName + "'s pd to one without certs");
				setClassProtectionDomain(topPDClass, nextPDWithoutCerts);
				topPD = nextPDWithoutCerts;
			}
		} else if ((nextPDWithCerts != null) && (nextPDWithoutCerts == null)) {
			//System.out.println("JNIUtils.fixProtectionDomains(): Setting topPDClass " +
			//	topPDClassName + "'s pd to one with certs");
			setClassProtectionDomain(topPDClass, nextPDWithCerts);
			topPD = nextPDWithCerts;
		}
	}

	// Check if we're doing LiveConnect either directly (in which case the
	// top class is always netscape.oji.LiveConnectProxy) or via
	// JEPRunOnLiveConnectThread() (in which case the top class is
	// java.awt.EventDispatchThread and netscape.oji.LiveConnectProxy is
	// a few classes below it).
	protected static boolean inLiveConnect(Class[] classContext) {
		int elements = classContext.length;
		String topClassName = classContext[elements-1].getName().replace('/', '.');
		if (topClassName.equals("netscape.oji.LiveConnectProxy"))
			return true;
		if (!topClassName.equals("java.awt.EventDispatchThread"))
			return false;
		for (int i = 1; i <= elements; ++i) {
			String className = classContext[i-1].getName().replace('/', '.');
			if (className.equals("netscape.oji.LiveConnectProxy"))
				return true;
		}
		return false;
	}

	// Returns the (zero-based) offset of the topmost class that has a
	// protection domain, or -1 if none of them do.
	protected static int getTopPDClassOffset(Class[] classContext) {
		for (int i = classContext.length - 1; i >= 0; --i) {
			ProtectionDomain pd = getClassProtectionDomain(classContext[i]);
			if (pd == null)
				continue;
			return i;
		}
		return -1;
	}

	private static native ProtectionDomain getClassProtectionDomain(Class clazz);
	private static native void setClassProtectionDomain(Class clazz,
		ProtectionDomain newProtectionDomain);

	/**
	 * Stub SecurityManager class, to expose access to class loaders.
	 */
	static class StubSecurityManager extends SecurityManager {
		// We generally call this from a thread with AllPermission, so
		// currentClassLoader() will generally no longer work here.
		public ClassLoader getCurrentClassLoader() {
			ClassLoader retval = null;
			ClassLoader systemLoader = ClassLoader.getSystemClassLoader();
			Class[] classContext = getClassContext();
			for (int i = 1; i <= classContext.length; ++i) {
				ClassLoader loader = null;
				try {
					loader = classContext[i-1].getClassLoader();
				} catch (SecurityException ex) {
					loader = currentClassLoader();
				}
				if ((loader != null) && !systemLoader.equals(loader)) {
					retval = loader;
					break;
				}
			}
			return retval;
		}
		public Class[] getCurrentClassContext() {
			return getClassContext();
		}
	}
	
	private static StubSecurityManager stubManager = new StubSecurityManager();
	
	/**
	 * Returns the current class loader.
	 */
	public static Object GetCurrentClassLoader() {
		return stubManager.getCurrentClassLoader();
	}
	
	public static Object GetObjectClassLoader(Object object) {
		return object.getClass().getClassLoader();
	}

	// PushLCInfo(), PopLCInfo() and friends are needed to get around a design
	// flaw in how the OJI interface creates JSObject objects -- it uses an
	// nsIPlugin method (GetJavaWrapper()) to create them (instead of an
	// nsIPluginInstance method), and therefore provides no easy way to
	// associate them with a particular plugin instance.  This is important
	// for a couple of reasons:  1) When a JSObject method is called from a
	// Java applet's instance method, we should run the method in the security
	// context of the plugin instance correponding to the applet.  2) When a
	// Java method is called on a non-applet Java object during JavaScript-to-
	// Java LiveConnect, and this call was initiated (however indirectly) from
	// a Java applet instance, the method should run in the AppContext of the
	// Java applet instance.  But neither Java nor JavaScript gives us any way
	// of knowing which applet we're running under.  (These problems don't
	// arise when using an applet's static methods.)  We get around these
	// problems by assuming that only one plugin instance will be doing
	// JavaScript-to-Java LiveConnect at any one time, and that only one
	// JavaScript-to-Java or Java-to-JavaScript call will be running at any
	// one time (since they all get processed on the main thread).  So we can
	// keep a stack of "current" plugin instances, and "push" and "pop" these
	// as appropriate (before and after every JavaScript-to-Java or Java-to-
	// JavaScript call, on the main thread).
	private static Object lcInfoLock = new Object();
	private static Stack lcInfo = new Stack();

	public static void PushLCInfo(int pluginInstance)
	{
		synchronized (lcInfoLock) {
			int newInfo[] = {0};
			try {
				newInfo = (int[]) lcInfo.peek();
			} catch (EmptyStackException e) {
			}
			if (pluginInstance != 0)
				newInfo[0] = pluginInstance;
			lcInfo.push(newInfo);
		}
	}

	public static void PopLCInfo() {
		synchronized (lcInfoLock) {
			try {
				lcInfo.pop();
			} catch (EmptyStackException e) {
			}
		}
	}

	public static int CurrentLCPluginInstance() {
		synchronized (lcInfoLock) {
			int currentInfo[] = {0};
			try {
				currentInfo = (int[]) lcInfo.peek();
			} catch (EmptyStackException e) {
			}
			return currentInfo[0];
		}
	}

}
